/* Copyright 2001, Todd Lewis <tlewis@mindspring.com>
 * Released under GNU GPL v2. */

/* group_action.c: creates event groups*/

#include "hank.h"
#include "event.h"
#include "packet.h"
#include "paengine.h"
#include "protocol_engine.h"
#include "protocol_lookup.h"
#include "support.h"
#include <stdio.h>
#include <string.h>

static char rcsid[] = "$Id: group_action.c,v 1.2 2001/11/10 01:45:33 tlewis Exp $";

static event_id create_event(void);
static handler_id register_handler(event_id incoming, event_id target, xmlNodePtr m);
static int unregister_handler(handler_id handler);
static int set_handler_target(event_id incoming, handler_id handler, event_id target);
static void handle_event(struct packet *p);

struct event_handler group_action  = {
	{
		"event_handler",
		"group_action ",
		EHMAJ,
		EHMIN,
		&group_action 
	},
	&create_event,
	&register_handler,
	&unregister_handler,
	&set_handler_target,
	&handle_event
};

struct eg {
	event_id id;
	struct eg *next;
};

struct *eg[1024] groups;
event_id group=1024;

static event_id
create_event(void)
{
	event_id e = --group;

	if(e>0){
		eg[e]=NULL;
		return(e);
	}

	return(0);
}

static handler_id
register_handler(event_id incoming, event_id target, xmlNodePtr m)
{
	xmlNodePtr x;
	struct eg *e;

	/* We don't have an xmlNode associated with this, just incoming
	 * and target. */

	if((e=malloc(sizeof(*e)))==NULL)
		exit(-1);
	e->id=target;
	e->next=eg[incoming];
	eg[incoming]=e;

	return(0);
}

static int
unregister_handler(handler_id handler)
{
	return(-1);
}

static int
set_handler_target(event_id incoming, handler_id handler, event_id target)
{
	return(-1);
}

static void
handle_event(struct packet *p)
{
	struct event *f, *g;
	struct eg *e;

   for(e=p->q;e;e=e->next){
      if((e->id>>24)==EH_GROUP){
			for(e=eg[e->id&0xffffff];e;e=e->next;){
				f=p->c->eng->alloc(p->c, sizeof(*f));
            f->id=e->event;
            f->consumed=0;
            f->next=NULL;
            f->payload=NULL;
            for(g=p->q;g->next;g=g->next)
            	;
				if(g)
					g->next=f;
				else
					p->q=f;
			}
		}
	}
}

