/* This file is compiled twice to create caseful and nocase data types.
 * The dominant macro is ACBM_CASEFUL in hdt_blob.c. */

static struct acbm_node*
TYPE(_rebuild_acbm_node)(struct acbm_node *n, struct memcon_context *mcc)
{
	struct acbm_node *a, *b, *c;
	unsigned int i, j, k;

	/* allocate node */
	a=mcc->eng->alloc(mcc, sizeof(*a));

	/* duplicate structure */
	memcpy(a, n, sizeof(*n));
	a->next=NULL;

	/* duplicate label */
	a->label=mcc->eng->alloc(mcc, n->len);
	memcpy(a->label, n->label, n->len);

	/* figure hash size */
	i=0;
	for(b=n->child[0];b;b=b->next)
		i++;
	PRINTF("Using raw mask %u, ", i);
	a->cmask = 0;
	if(i){
		for(k=(8*sizeof(i));k>0;k--){
			if ((i>>(k-1))&1){
				i = 1<<(k-1);
				break;
			}
		}
		a->cmask = i - 1;
	}

	/* set mask */
	PRINTF("%u\n", i);

	/* create table */
	j=(i*sizeof(a->child[0]));
	a->child=mcc->eng->alloc(mcc, j);
	memset(a->child, 0, j);

	/* dup & insert kids */
	for(b=n->child[0];b;b=b->next){
		c=TYPE(_rebuild_acbm_node)(b, mcc);
		c->next=a->child[((c->label[0])&(a->cmask))];
		a->child[ ( (c->label[0]) & (a->cmask) ) ] = c;
		PRINTF("Inserted child into slot (%i & %i) %i\n", 
			(c->label[0]), (a->cmask),
			((c->label[0]) & (a->cmask) ));
	}

	/* return node */
	return(a);
}

static void
TYPE(_rebuild_acbm_tree)(struct hdt_blob_set *s)
{
	struct memcon_context *c;
	struct acbm_node *n;

	/* create a new context */
	if((c=s->c->eng->create(NULL, "hdt_blob/acbm"))==NULL){
		fprintf(stderr, "AIEE!  Can't create memcon context.  Aborting.\n");
		exit(-1);
	}

	/* dup the tree */
	n=TYPE(_rebuild_acbm_node)(s->tree, c);

	/* free the old tree */
	s->c->eng->destroy(s->c);

	/* set set->tree = newtree */
	s->c=c;
	s->tree=n;
}


static void
TYPE (_encode) (char *pattern, prim_id prid, struct blob_match * r,
		struct memcon_context * mmc)
{
  char *a, *b;
  unsigned int bmode = 0, i;
  char c;

  PRINTF ("Encoding pattern.\n");
  /* This can be too big, but fuck it. */
  if ((r->pattern = mmc->eng->alloc (mmc, strlen (pattern))) == NULL)
    exit (-1);

  r->prid = prid;
  r->depth = r->offset = r->len = 0;

  /* syntax is "offset,depth,pattern".  depth and/or offset can
   * be empty.  pattern is either plain, '|'-encased binary, or
   * escaped (as in the case of '|', '\' or ':'). */

  a = pattern;
  if ((b = strchr (pattern, ',')))
    *b = 0;
  else
    {
      FPRINTF (stderr, "AIEE!  Poorly-formatted pattern (%s).  Exiting.\n",
	       pattern);
      exit (-1);
    }
  b++;
  if (strlen (a))
    r->offset = atoi (a);
  a = b;

  PRINTF ("Offset: %i; ", r->offset);

  b = strchr (a, ',');
  *b = 0;
  b++;
  if (strlen (a))
    r->depth = atoi (a);
  a = b;

  PRINTF ("Depth: %i; ", r->depth);

  while (*a)
    {
      if (a[0] == '\\')
	{
	  r->pattern[r->len++] = CCONV (a[1]);
	  a = a + 2;
	}
      else if (a[0] == '|')
	{
	  a++;
	  if (bmode)
	    {
	      bmode = 0;
	    }
	  else
	    {
	      bmode = 1;
	      while (isspace (a[0]))
		a++;
	    }
	}
      else
	{
	  if (bmode)
	    {
	      c = CCONV (a[2]);
	      a[2] = 0;
	      if (sscanf (a, "%2x", &i))
		{
		  r->pattern[r->len++] = CCONV ((unsigned char) i);
		}
	      else
		{
		  FPRINTF (stderr,
			   "AIEE!  Could not convert binary: %s.  Exiting.\n",
			   a);
		  exit (-1);
		}
	      a[2] = c;
	      a = a + 2;
	      while (isspace (a[0]))
		a++;
	    }
	  else
	    {
	      r->pattern[r->len++] = CCONV (a[0]);
	      a++;
	    }
	}
    }

  PRINTF ("Pattern length: %i; ", r->len);
}

static void TYPE (_build_acbm) (struct hdt_blob_set * h)
{
  unsigned char *pattern;
  unsigned char y;
  unsigned int i, j, k;
  SIZE_T s, x;
  struct acbm_node *a, *b, *c, *d;

	if(!(h->kids)) /* There are no patterns here */
		return;

  PRINTF("Rebuilding acbm\n");

  x = h->pmax;

  PRINTF ("Creating tree: npatterns %i, total size %i\n",
	  h->pattern_cursor, h->ptot);
  /* PRIMO: Set up bad character shift:
   * "The preprocessing needed to implement the bad character rule
   * is simple and is left to the reader."(p160)  Huh huh, I'm, like,
   * not a good computer scientist, and stuff. */

  /* First dimension is character; second dimension is suffix-offset */

  /* PRINTF("Creating tree; test memory now.  (Waiting 15 seconds...)"); */
  /* sleep(15); */
  PRINTF ("creating.\n");

  DUMP_ACBM_TREE (h, "j",
		  ACBM_REPORT_GST | ACBM_REPORT_ZV | ACBM_REPORT_DPRIMEV);

  PRINTF ("Creating bad char table (256 x %i)\n", x);
  /* First, allocate bad_char table */
  for (i = 0; i < 256; i++)
    {
      if ((h->bad_char[i] = malloc ((x * sizeof (SIZE_T)))) == NULL)
	exit (-1);
		memset(h->bad_char[i], 0, (x * sizeof (SIZE_T)));
    }

  /* second, fill out the table */
  for (i = 0; h->patterns[i].pattern; i++)
    {
      PRINTF ("Pattern %i (%s):\n", i, h->patterns[i].pattern);
      pattern = h->patterns[i].pattern;
      /* s = strlen ((char *) pattern); Fuck me! */
		s = h->patterns[i].len;
      for (j = 0; j < s; j++)
	{
	  y = CCONV (pattern[j]);
	  PRINTF (" Adding %c at [%3i][%i] as (%i - %i)\n", y, y, s - j - 1,
		  s, j);
	  h->bad_char[y][s - j - 1] = s - j;
	}
    }

  PRINTF ("TABLE BEFORE PULL-UP:\n");
  dump_table (h);
  PRINTF ("(That was it.)\n");

  /* Third, pull up the values */
  for (i = 0; i < 256; i++)
    {				/* for each character */
      s = 0;
      for (j = 1; j <= x; j++)
	{
	  if (h->bad_char[i][x - j] == 0)
	    {
	      /* carry old value in here */
	      h->bad_char[i][x - j] = s;
	    }
	  else
	    {
	      /* replace old value */
	      s = h->bad_char[i][x - j];
	    }
	}
    }
  PRINTF ("\nTABLE AFTER PULL-UP:\n");
  dump_table (h);

  /* SECUNDO: set up good suffix shift tree, lambda */

  h->tree = create_acbm_node (0, NULL, 0, 0, h->c, 1);

  /* First: Build generalized suffix tree for good suffix shifting */
  for (i = 0; h->patterns[i].pattern; i++)
    {
      unsigned char *rbuf;	/* reverse suffix buffer */

      PRINTF ("Doing pattern %i\n", i);

      pattern = h->patterns[i].pattern;
      if ((rbuf = h->c->eng->alloc (h->c, (s = h->patterns[i].len))) == NULL)
	exit (-1);
      for (j = 0; j < s; j++)
	{
	  OFF_T rbufoff, bpos;

	  PRINTF ("	Doing suffix %i: ", j);

	  /* if(s==0){
	     PRINTF("Skipping empty pattern, %i.\n", i);
	     break;
	     } */

	  /* create the reverse suffix buffer */
	  for (k = 0; k <= j; k++)
	    {
	      PRINTF (" (%c: %i = %i) ", pattern[j - k], k, j - k);
	      rbuf[k] = pattern[j - k];
	    }

	  /* insert suffix pattern[s-j..j] having pattern=i, suffixpos=j */
	  /* set pointer to root */
	  PRINTF ("\nStarting loop: ");
	  rbufoff = 0;
	  a = h->tree;
	  while (a)
	    {
	      (int)bpos = -1;
	      PRINTF ("(step) ");
	      fflush (stdout);

	      if (j < rbufoff)
		{
		  /* we're a null node, so go to tail of class */
		  PRINTF (" null_child ");
		  c = create_acbm_node (0, NULL, i, s - j, h->c, 1);
		  if (!a->child[0])
		    {
		      a->child[0] = c;
		    }
		  else
		    {
		      a = a->child[0];
		      while (a->next)
			a = a->next;
		      a->next = c;
		      a = NULL;
		    }
		  break;
		}

	      if (a->len && a->label[0] == rbuf[rbufoff])
		{
		  /* we're good to see if (full|partial) match */
		  if (MEMCMP
		      (&rbuf[rbufoff], a->label,
		       MIN ((a->len), (j - rbufoff))))
		    {
		      /* we set the branch position to be the mismatch point */
		      PRINTF (" branch ");
		      for (k = 0; k <= MIN ((a->len), (j - rbufoff)); k++)
			{
			  if (rbuf[rbufoff + k] != a->label[k])
			    {
			      PRINTF (" (branch position %i) ", k);
			      bpos = k;
			      break;
			    }
			}
		    }
		  else if ((j - rbufoff + 1) < (a->len))
		    {
		      /* we're a leaf off of this edge */
		      bpos = j - rbufoff + 1;
		      PRINTF (" leaf at %i (less than %i) ", bpos, a->len);
		    }
		  else
		    {
		      /* we match this node. */
		      PRINTF (" match ");
		      rbufoff += a->len;
		    }
		}
	      if ((int)bpos != -1)
		{
		  /* we need to branch */
		  PRINTF (" create_branch of \"%s\" at %i ", a->label, bpos);
		  /* b is the new, second half of a */
		  b = create_acbm_node ((a->len - bpos), &(a->label[bpos]),
					a->pattern, a->z, h->c, 1);

		  /* c is the new, second half of our pattern */
		  PRINTF (" (Second half at %i) ", bpos);
		  c =
		    create_acbm_node ((j - rbufoff - bpos + 1),
				      &rbuf[rbufoff + bpos], i, s - j, h->c, 1);

		  a->len = bpos;
		  b->child[0] = a->child[0];

		  /* arrange the two kids in order */
		  if (c->len == 0 || c->label[0] < b->label[0])
		    {
		      a->child[0] = c;
		      c->next = b;
		    }
		  else
		    {
		      a->child[0] = b;
		      b->next = c;
		    }
		  a = NULL;
		  break;

		}
	      else
		{
		  /* Ok, we should be a child of this node */
		  PRINTF (" (real_child: no bpos: %i) ", bpos);
		  c = NULL;

		  /* If we're appending to a terminal node, we need to add
		   * a new, empty child. */
		  if (a->child[0] == NULL)
		    {
		      a->child[0] =
			create_acbm_node (0, NULL, a->pattern, a->z, h->c, 1);
		    }

		  for (b = a->child[0]; b; b = b->next)
		    {
		      if (b->len)
			{
			  if (b->label[0] > rbuf[rbufoff])
			    {
			      /* we passed it; insert in between c and b */
			      PRINTF (" nomatch_insert at %i ", rbufoff);
			      d =
				create_acbm_node ((j - rbufoff + 1),
						  &rbuf[rbufoff], i, s - j,
						  h->c, 1);
			      if (c)
				c->next = d;
			      else
				a->child[0] = d;
			      d->next = b;
			      a = NULL;
			      break;
			    }
			  else if (b->label[0] == rbuf[rbufoff])
			    {
			      PRINTF (" child_match ");
			      /* match! */
			      a = b;
			      break;
			    }
			  else
			    {
			      /* keep going */
			    }
			}
		      c = b;
		    }
		  if (b == NULL)
		    {
		      /* we are past all of the present nodes */

		      PRINTF (" append_child at %i ", rbufoff);
		      d =
			create_acbm_node ((j - rbufoff + 1), &rbuf[rbufoff],
					  i, s - j, h->c, 1);
		      if (c)
			c->next = d;
		      else
			a->child[0] = d;
		      a = NULL;
		      break;
		    }
		}
	    }
	}
      PRINTF (" %i (done)\n", i);
      DUMP_ACBM_TREE (h, "g", ACBM_REPORT_GST);
    }

  /* N.b., since all of these are implemented recursively, they are put in separate
   * functions, above.  I think I am starting to understand the
   * pascal notion of nested functions by virtue of not having them
   * here. */

  /* Second, mark every node that's an ancestor of a leaf numbered by suffix pos one */
  PRINTF ("Second:\n");
  second_phase (h->tree);

  /* Third, for each marked node, set zv to smallest suffix pos z, z>1, of any leaf in subtree */
  PRINTF ("Third:\n");
  third_phase (h->tree);

  /* Fourth, label terminal-leaf parents */
  /* Fifth, set dv to min(anccestors:du) */
  /* (Consolidated) */
  DUMP_ACBM_TREE (h, "h",
		  ACBM_REPORT_GST | ACBM_REPORT_ZV | ACBM_REPORT_DPRIMEV |
		  ACBM_REPORT_MARK | ACBM_REPORT_DU);
  PRINTF ("Fourth:\n");
  fourth_phase (h->tree);

  PRINTF ("Fifth:\n");
  fifth_phase (h->tree, 0);
  DUMP_ACBM_TREE (h, "i",
		  ACBM_REPORT_GST | ACBM_REPORT_ZV | ACBM_REPORT_DPRIMEV |
		  ACBM_REPORT_MARK | ACBM_REPORT_DU);

  /* Sixth, consolidate tree */
  PRINTF ("Sixth:\n");
  sixth_phase (h->tree);
  DUMP_ACBM_TREE (h, "j",
		  ACBM_REPORT_GST | ACBM_REPORT_ZV | ACBM_REPORT_DPRIMEV);
  /* TYPE(_rebuild_acbm_tree) (h); */

  /*
  PRINTF("Freeing tree; test memory now.  (Waiting 15 seconds...)");  fflush(stdout);
  sleep(15);
  PRINTF(" freed (sleeping 15 more seconds)...");  fflush(stdout);
  sleep(15);
  PRINTF("done.\n");
	*/

  DUMP_ACBM_TREE (h, "k",
		  ACBM_REPORT_GST | ACBM_REPORT_ZV | ACBM_REPORT_DPRIMEV);


  h->built = 1;
  PRINTF ("Done.\n");
}

static hdt_prim TYPE (_add) (hdt_set set, prim_id prid, char *arg)
{
  struct hdt_blob_set *h = sets[set];
  struct blob_match *b = &(h->patterns[h->pattern_cursor++]);

  PRINTF ("Encoding arg %s.\n", arg);
  TYPE (_encode) (arg, prid, b, h->c);
	if(b->len<1){
		fprintf(stderr, "AIEE!  Illegal pattern (too short) submitted.  Returning error.\n");
		return(-1);
	}

  if (h->pmin > b->len || h->pmin == 0){
    h->pmin = b->len;
  } else {
	PRINTF("Not setting pmin; length %i not less than pmin %i.\n", b->len, h->pmin);
  }

  if (h->pmax < b->len)
    h->pmax = b->len;

  h->ptot += b->len;

  h->built = 0;
  h->kids++;
  return (0);
}

static int
TYPE (_match) (hdt_set set, hdt_prim * prims, struct blob * val,
	       unsigned char *rbuf, char newbits)
{
  struct hdt_blob_set *x = sets[set];
  struct acbm_node *f, *g, *h = x->tree;
  u_int32_t i, ione = 0, itwo, ithree, offset, iold;
	char *casename;
	unsigned long start, end, fs, fe, ls, le;

	rdtscl(fs);
#ifdef ACBM_CASELESS
	casename = "caseless";
#else
	casename = "casefull";
#endif
	PRINTF ("Entering %s match.\n", casename);

  if (prims)
    {
      /* we don't support partial set matching yet */
		fprintf(stderr, "AIEE!  Received prims request at %s:%i.  Unsupported.  Aborting.\n", __FILE__, __LINE__);
      return (-1);
    }

  if (!x->pmin){
	fprintf(stderr, "AIEE!  x->pmin not set at %s:%i.  Aborting.\n", __FILE__, __LINE__);
    return (-1);
  }

  if (x->built == 0){
	PRINTF("AIEE!  Building tree.\n");
    TYPE (_build_acbm) (x);
  }

	rdtscl(fe);
	PRINTF("First took %i cycles.\n", (fe-fs));

  /* Now that that's out of the way, let's match! */

  PRINTF ("MATCHING for %i.  \n", val->len);
  for (i = (x->pmin - 1); i < val->len;)
    {
		iold=i;
		rdtscl(start);
      PRINTF ("\nBIGSTEP: (%i)\n", i);
      itwo = ithree = 0;
      h = x->tree;
      offset = 0;		/* the reverse offset of the cursor from i */
      while (h)
	{
	  f = h;
	  h = NULL;
	  PRINTF ("lstep (offset %i, pos %i) ", offset, i);
	  /* coming in, we already match h */

	  if (!f->child[CCONV((val->blob[i - offset]) & f->cmask)])
	    {
	      /* terminal node; confirm match, mark and move on */
	      struct blob_match *bb = &x->patterns[f->pattern];
	      PRINTF (" terminal ");
	      if (!MEMCMP (&val->blob[i - offset + 1], bb->pattern, bb->len))
		{
		  if (bb->offset)
		    {
		      if ((i - offset + 1) < bb->offset)
			{
			  PRINTF
			    ("\nSkipping match %i; i - offset + 1 (%i) < bb->offset (%i)\n",
			     f->pattern, i - offset + 1, bb->offset);
			  goto skip1;
			}
		    }
		  if (bb->depth)
		    {
		      if ((i - offset + 1 + bb->len) > bb->depth + bb->offset)
			{
			  PRINTF
			    ("\nSkipping match %i; ( i (%i) - offset (%i) + 1 + bb->len (%i) ) (%i) > ( bb->depth(%i) + bb->offset(%i) \n",
				  f->pattern, i, offset, bb->len, i - offset + 1 + bb->len, bb->depth, bb->offset);
			  goto skip1;
			}

		    }
		  {
		    char *s;
#ifdef ACBM_CASELESS
		    s = "caseless";
#else
		    s = "casefull";
#endif
		    PRINTF ("*** MATCH(%s) at %i to pattern %i ***\n",
			    s, i - offset + 1, f->pattern);
		  }
		  rbuf[bb->prid] |= newbits;
		skip1:
				itwo = i + 1;
		}
	      else
		{
		  PRINTF
		    ("FAILED(1) Match at %i (%i - %i) (%c, %s) to pattern %i (%c, %s)\n",
		     i - offset + 1, i, offset - 1,
		     val->blob[i - offset], val->blob, bb->pattern, bb->pattern[0], bb->pattern);
		}
	    }

	  PRINTF ("Looping on list at %c (%i & %i).\n", val->blob[i - offset],
		  CCONV(val->blob[i - offset]), f->cmask);
		  PRINTF("Jumping into child at char %i, mask %i, value %i\n",
	  		CCONV(val->blob[i - offset]) , f->cmask,
	  		(CCONV(val->blob[i - offset]) & f->cmask));

	  for (g = f->child[(CCONV(val->blob[i - offset]) & f->cmask)]; g;
	       g = g->next)
	    {
	      /* we have children in here; check them */
	      PRINTF ("kid, ");
	      if (!g->len)
		{
		  /* empty terminal node; check for match.  Yes, it sucks
		   * that this is done twice, and if I continued to polish
		   * this, then I probably could roll this and the other
		   * memcmp into one piece of code, but fuck it, it works. */
		  struct blob_match *bb = &x->patterns[g->pattern];
		  PRINTF ("(terminal kid, skipcheck) ");
		  if (!MEMCMP
		      (&val->blob[i - offset +1], bb->pattern, bb->len))
		    {
		      if (bb->offset)
			{
			  if ((i - offset + 1) < bb->offset)
			    {
			      PRINTF
				("\nSkipping match %i; i - offset + 1 (%i) < bb->offset (%i)\n",
				 g->pattern, i - offset + 1, bb->offset);
			      goto skip2;
			    }
			}
		      if (bb->depth)
			{
		      if ((i - offset + 1 + bb->len) > bb->depth + bb->offset)
			    {
			  PRINTF
			    ("\nSkipping match %i; ( i (%i) - offset (%i) + 1 + bb->len (%i) ) (%i) > ( bb->depth(%i) + bb->offset(%i) \n",
				  f->pattern, i, offset, bb->len, i - offset + 1 + bb->len, bb->depth, bb->offset);
			      goto skip2;
			    }
			}
		      {
			char *s;
#ifdef ACBM_CASELESS
			s = "caseless";
#else
			s = "casefull";
#endif
			PRINTF ("*** MATCH(%s) at %i to pattern %i ***\n",
				s, i - offset + 1, g->pattern);
		      }
		      rbuf[bb->prid] |= newbits;

		    skip2:
				itwo = i + 1;
		    }
		  else
		    {
		      PRINTF
			("FAILED(2) Match at %i (%i - %i) (%c, %s) to pattern %i (%c, %s)\n",
			 i - offset + 1, i, offset - 1,
			 val->blob[i - offset + 1], val->blob, g->pattern,
			 bb->pattern[0], bb->pattern);
		    }

		}
	      else if (g->label[0] == CCONV (val->blob[i - offset]))
		{
		  unsigned int to;

		  /* at least partial match. */
		  PRINTF ("Reverse checking match (len=%i): ", g->len);
		  for (to = 0; to < g->len; to++)
		    {
		      if (CCONV (val->blob[i - offset - to]) != g->label[to])
			{
			  PRINTF ("(%c != %c), ",
				  val->blob[i - offset - to], g->label[to]);
			  break;
			}
		      else
			{
			  PRINTF ("(%c == %c), ",
				  val->blob[i - offset - to], g->label[to]);
			}
		    }

		  offset += to;

		  if (to >= g->len)
		    {
		      /* full match; set h & loop */
		      PRINTF ("full (%s) ", g->label);
		      h = g;
		    }
		  else
		    {
		      /* partial match */
		      PRINTF ("partial ");
		      h = NULL;
		    }

		  break;
		}
	      else
		{
		  PRINTF ("%c != %c ", g->label[0], val->blob[i - offset]);
		}
	    }

	  if (!h)
	    {
	      /* if present, none of the kids fully match. */
	      PRINTF ("bad char for \"%c\"\n", CCONV (val->blob[i - offset]));
	      if (!(x->bad_char[CCONV (val->blob[i - offset])][offset])){
		PRINTF("No bad char: Setting ione to i(%i) - offset(%i) + x->pmin(%i)\n",
			i , offset , x->pmin);
		ione = i - offset + x->pmin;
	      } else {
		PRINTF("Bad char: Setting ione to i(%i) - offset(%i) + x->bad_char[CCONV(%i/%c)][offset(%i)] (%i)\n",
			i , offset , CCONV (val->blob[i - offset]), CCONV (val->blob[i - offset]), offset,
		  	(x->bad_char[CCONV (val->blob[i - offset])][offset] - 1));
		ione = i - offset + (x->bad_char[CCONV (val->blob[i - offset])][offset] - 1);
	      }
	      if (f->dprimev)
		ithree = i + f->dprimev - 2;
	      if (g)
		f = g;
	  if (f->zv)
		itwo = i + f->zv - 1;
	    }
	}

     PRINTF 
	("MOVE: %i, chosing among pmin: %i, i1: %i, i2: %i, i3: %i.  ",
	 i, i + x->pmin, ione, itwo, ithree);

		if(ione==0 || ione > (i+x->pmin)){
			ione=i;
		}

		if(itwo==0 || itwo > (i+x->pmin)){
			itwo=i;
		}

		if(ithree==0 || ithree > (i+x->pmin)){
			ithree=i;
		}

     PRINTF 
	("post fix-up: pmin: %i, i1: %i, i2: %i, i3: %i.  ",
	 i + x->pmin, ione, itwo, ithree);

      itwo = MIN (itwo, ithree);

      ione = MAX (ione, itwo);

      if (1)
	{
	  ione = MIN ((i + x->pmin), ione);
	  if (ione <= i)
	    {
	      PRINTF
		("Aiee!  Supposed to move to %i, which is _left_ of %i!\n",
		 ione, i);
	      i++;
	    }
	  else
	    {
	      i = ione;
	    }
	}
      else
	{
	  i++;
	}

      PRINTF ("Chose %i.\n", i);
		 rdtscl(end);
		 PRINTF("Loop: %i bytes in %u cycles, %i microcycles per byte\n", i-iold, end-start, (1000*(end-start))/(i-iold));
    }
	 rdtscl(le);
	 PRINTF("acbm_match (%s) loop took %i cycles.\n", 
#ifdef ACBM_CASELESS
		"caseless",
#else
		"caseful",
#endif
	 	(le-fs));

  PRINTF ("Returning at offset %i (greater than %i)\n", i, val->len);
  return (0);
}
