#include "hdt_int.h"
#include "hank.h"
#include "memcon.h"
#include "support.h"
#include <string.h>

static char rcsid[] = "$Id: hdt_int.c,v 1.5 2001/11/10 01:45:33 tlewis Exp $";

extern struct memcon *mc;

/* This needs to contain a single registered op. */
/* needs to be ref-counted. */
struct hdt_int_prim {
	/* add a ref count */
	u_int32_t val;
	prim_id prid;
	struct hdt_int_prim *next;
};

struct hdt_int_two_prim {
	/* add a ref count */
	u_int32_t val1;
	u_int32_t val2;
	prim_id prid;
	struct hdt_int_two_prim *next;
};

/* This needs to contain all of the registered ops in the set. */
/* Has one list for each op. */
struct hdt_int_set {
	struct memcon_context *c;
	struct hdt_int_prim *equals;
	struct hdt_int_prim *lt;
	struct hdt_int_prim *gt;
	struct hdt_int_prim *and;
	struct hdt_int_two_prim *range;
	struct hdt_int_two_prim *modulo;
	/* add a list of prims */
};

static struct hdt_int_set *sets[16384];
static int hdt_int_sets_offset=0;

hdt_set 
hdt_int_create(void)
{
	hdt_set o = hdt_int_sets_offset++;
	struct memcon_context *c;
	struct hdt_int_set *s;

	if((unsigned int)o >= sizeof(sets)) /* out of room! */
		return(-1);

	/* create a new memcon */
	c=mc->create(0, "hdt_int");
	if(c==NULL)
		exit(-1);

	/* allocate a new set */
	if((s=c->eng->alloc(c, sizeof(*s)))==NULL)
		exit(-1);

	/* populate the set */
	memset(s, 0, sizeof(*s));
	s->c=c;

	/* insert the set */
	sets[o]=s;

	/* return the offset */
	return(o);
}

void 
hdt_int_destroy(hdt_set set)
{
	struct hdt_int_set *s = sets[set];

	/* free all of the plist's */
	s->c->eng->destroy(s->c);
	sets[set]=NULL;
}

hdt_prim 
hdt_int_add(hdt_set set, prim_id prid, char *op, char *arg)
{
	struct hdt_int_set *s = sets[set];
	struct hdt_int_prim *p;
	struct hdt_int_two_prim *r;
	char *a[2];

	/* TODO: detect duplicates */
	if(!strcmp(op, "=")){
		/* equals */
		if((p=s->c->eng->alloc(s->c, sizeof(*p)))==NULL)
			exit(-1);
		p->prid=prid;
		p->val=atoi(arg);
		p->next=s->equals;
		s->equals=p;

	} else if(!strcmp(op, "<")){
		/* lt */
		if((p=s->c->eng->alloc(s->c, sizeof(*p)))==NULL)
			exit(-1);
		p->prid=prid;
		p->val=atoi(arg);
		p->next=s->lt;
		s->lt=p;

	} else if(!strcmp(op, ">")){
		/* gt */
		if((p=s->c->eng->alloc(s->c, sizeof(*p)))==NULL)
			exit(-1);
		p->prid=prid;
		p->val=atoi(arg);
		p->next=s->gt;
		s->gt=p;

	} else if(!strcmp(op, "&")){
		/* and*/
		if((p=s->c->eng->alloc(s->c, sizeof(*p)))==NULL)
			exit(-1);
		p->prid=prid;
		p->val=atoi(arg);
		p->next=s->and;
		s->and=p;

	} else if(!strcmp(op, ":")){
		/* range */
		if((r=s->c->eng->alloc(s->c, sizeof(*r)))==NULL)
			exit(-1);
		r->prid=prid;
		split_kv(arg, a, '-');
		r->val1=atoi(a[0]);
		r->val2=atoi(a[1]);
		r->next=s->range;
		s->range=r;

	} else if(!strcmp(op, "%")){
		/* modulo*/
		if((r=s->c->eng->alloc(s->c, sizeof(*r)))==NULL)
			exit(-1);
		r->prid=prid;
		split_kv(arg, a, '=');
		r->val1=atoi(a[0]);
		r->val2=atoi(a[1]);
		r->next=s->modulo;
		s->modulo=r;

	} else {
		return(-1);
	}
	/* need to return list offset */
	return(0);
}

void 
hdt_int_del(hdt_set set, hdt_prim prim)
{
	/* need a list of prims in hdt_int_set */
	/* not implemented */
}

#define PRINTF(x...)

int
hdt_int_match(hdt_set set, hdt_prim *prims, u_int32_t arg, unsigned char *rbuf)
{
	struct hdt_int_prim *i;
	struct hdt_int_two_prim *r;
	struct hdt_int_set *s = sets[set];

	/* equals */
	for(i=s->equals; i; i=i->next){
		if(i->val==arg){
			PRINTF("%i: %i = %i\n", i->prid, i->val, arg);
			rbuf[i->prid]=1;
		} else {
			PRINTF("%i: %i != %i\n", i->prid, i->val, arg);
		}
	}

	/* and */
	for(i=s->and; i; i=i->next){
		if((i->val&arg) == arg){
			PRINTF("%i: %i & %i\n", i->prid, i->val, arg);
			rbuf[i->prid]=1;
		}
	}

	/* lt */
	/* FIXME: If this were an ordered list, the cost of this would
	 * approach O(log(n)), instead of O(n) as it is now. */
	for(i=s->lt; i; i=i->next){
		if(i->val<arg){
			PRINTF("%i: %i < %i\n", i->prid, i->val, arg);
			rbuf[i->prid]=1;
		} else {
			PRINTF("%i: %i !< %i\n", i->prid, i->val, arg);
		}
	}

	/* gt */
	/* FIXME: If this were an ordered list, the cost of this would
	 * approach O(log(n)), instead of O(n) as it is now. */
	for(i=s->gt; i; i=i->next){
		if(i->val>arg){
			PRINTF("%i: %i > %i\n", i->prid, i->val, arg);
			rbuf[i->prid]=1;
		} else {
			PRINTF("%i: %i !> %i\n", i->prid, i->val, arg);
		}
	}

	/* range */
	/* FIXME: If this list were ordered by min, then you could easily
	 * cut the work, on average, in half. */
	for(r=s->range; r; r=r->next){
		if( (r->val1 <= arg) && (r->val2 >= arg) ){
			PRINTF("%i: %i < %i < %i\n", i->prid, r->val1, arg, r->val2);
			rbuf[i->prid]=1;
		}
	}

	/* modulo*/
	for(r=s->modulo; r; r=r->next){
		if( (arg % r->val1) == r->val2 ) {
			PRINTF("%i: %i %% %i = %i\n", i->prid, r->val1, arg, r->val2);
			rbuf[i->prid]=1;
		}
	}

	return(0);
}

