#include "hdt_int32.h"
#include "hank.h"
#include "memcon.h"
#include "support.h"
#include <string.h>

static char rcsid[] = "$Id: hdt_int32.c,v 1.12 2002/01/03 22:02:28 tlewis Exp $";

extern struct memcon *mc;

/* This needs to contain a single registered op. */
/* needs to be ref-counted. */
struct hdt_int32_prim {
	/* add a ref count */
	u_int32_t val;
	prim_id prid;
	struct hdt_int32_prim *next;
};

struct int32_twoval {
	u_int32_t fill;
	u_int32_t **value1s;
	u_int32_t **value2s;
	u_int32_t **prids;
};

struct hdt_int32_two_prim {
	/* add a ref count */
	u_int32_t val1;
	u_int32_t val2;
	prim_id prid;
	struct hdt_int32_two_prim *next;
};

/* This needs to contain all of the registered ops in the set. */
/* Has one list for each op. */
struct hdt_int32_set {
	struct memcon_context *c;
	struct hdt_int32_prim *equals;
	struct hdt_int32_prim *lt;
	struct hdt_int32_prim *gt;
	struct hdt_int32_prim *mask;
	struct hdt_int32_two_prim *mask_all;
	struct hdt_int32_two_prim *range;
	struct hdt_int32_two_prim *modulo;
	struct hdt_int32_two_prim *me;
	/* add a list of prims */
};

static struct hdt_int32_set *sets[16384];
static int hdt_int32_sets_offset=1;

hdt_set 
hdt_int32_create(void)
{
	hdt_set o = hdt_int32_sets_offset++;
	struct memcon_context *c;
	struct hdt_int32_set *s;

	if((unsigned int)o >= sizeof(sets)) /* out of room! */
		return(-1);

	/* create a new memcon */
	c=mc->create(0, "hdt_int32");
	if(c==NULL)
		exit(-1);

	/* allocate a new set */
	if((s=c->eng->alloc(c, sizeof(*s)))==NULL)
		exit(-1);

	/* populate the set */
	memset(s, 0, sizeof(*s));
	s->c=c;

	/* insert the set */
	sets[o]=s;

	/* return the offset */
	return(o);
}

void 
hdt_int32_destroy(hdt_set set)
{
	struct hdt_int32_set *s = sets[set];

	/* free all of the plist's */
	s->c->eng->destroy(s->c);
	sets[set]=NULL;
}

hdt_prim 
hdt_int32_add(hdt_set set, prim_id prid, char *op, char *arg)
{
	struct hdt_int32_set *s = sets[set];
	struct hdt_int32_prim *p;
	struct hdt_int32_two_prim *r;
	char *a[2];
	u_int32_t u;

	/* TODO: detect duplicates */
	if(!strcmp(op, "=")){
		/* equals */
		if((p=s->c->eng->alloc(s->c, sizeof(*p)))==NULL)
			exit(-1);
		p->prid=prid;
		p->val=atoi(arg);
		p->next=s->equals;
		s->equals=p;

	} else if(!strcmp(op, "<")){
		/* lt */
		if((p=s->c->eng->alloc(s->c, sizeof(*p)))==NULL)
			exit(-1);
		p->prid=prid;
		p->val=atoi(arg);
		p->next=s->lt;
		s->lt=p;

	} else if(!strcmp(op, ">")){
		/* gt */
		if((p=s->c->eng->alloc(s->c, sizeof(*p)))==NULL)
			exit(-1);
		p->prid=prid;
		p->val=atoi(arg);
		p->next=s->gt;
		s->gt=p;

	} else if(!strcmp(op, "mask")){ 
		/* mask any: val & arg != 0 */
		if((p=s->c->eng->alloc(s->c, sizeof(*p)))==NULL)
			exit(-1);
		p->prid=prid;
		p->val=atoi(arg);
		p->next=s->mask;
		s->mask=p;

	} else if(!strcmp(op, "mask_all")){
		/* mask all: val & arg1 = arg2*/
		if((r=s->c->eng->alloc(s->c, sizeof(*r)))==NULL)
			exit(-1);
		r->prid=prid;
		split_kv(arg, a, ':');
		if(a[1]==NULL){
			fprintf(stderr, "Syntax error: expecting argument \"<val>:<val>\" for mask_all operator.\n");
			return(-1);
		}
		r->val1=atoi(a[0]);
		r->val2=atoi(a[1]);
		r->next=s->mask_all;
		s->mask_all=r;

	} else if(!strcmp(op, ":")){
		/* range */
		if((r=s->c->eng->alloc(s->c, sizeof(*r)))==NULL)
			exit(-1);
		r->prid=prid;
		split_kv(arg, a, '-');
		if(a[1]==NULL){
			fprintf(stderr, "Syntax error: expecting argument \"<val>-<val>\" for range operator.\n");
			return(-1);
		}
		r->val1=atoi(a[0]);
		r->val2=atoi(a[1]);
		if(r->val1 > r->val2){ /* we're cool. */
			u=r->val1;
			r->val1=r->val2;
			r->val2=u;
		}
		r->next=s->range;
		s->range=r;

	} else if(!strcmp(op, "%")){
		/* modulo*/
		if((r=s->c->eng->alloc(s->c, sizeof(*r)))==NULL)
			exit(-1);
		r->prid=prid;
		split_kv(arg, a, '=');
		if(a[1]==NULL){
			fprintf(stderr, "Syntax error: expecting argument \"<val>=<val>\" for modulo operator.\n");
			return(-1);
		}
		r->val1=atoi(a[0]);
		r->val2=atoi(a[1]);
		r->next=s->modulo;
		s->modulo=r;

	} else {
		return(-1);
	}
	/* need to return list offset */
	return(0);
}

void 
hdt_int32_del(hdt_set set, hdt_prim prim)
{
	/* need a list of prims in hdt_int32_set */
	/* not implemented */
}

#define PRINTF(x...)

int
hdt_int32_match(hdt_set set, hdt_prim *prims, u_int32_t arg, unsigned char *rbuf, char newbits)
{
	struct hdt_int32_prim *i;
	struct hdt_int32_two_prim *r;
	struct hdt_int32_set *s = sets[set];
	unsigned long b, e;

	rdtscl(b);
	PRINTF("hdt_int32_match: matching.\n");

	if(!s){
		printf("AIEE!  hdt_int32_match() called with a null set!\n");
		return(-1);
	}

	/* equals */
	for(i=s->equals; i; i=i->next){
		if(i->val==arg){
			PRINTF("%i: %i = %i\n", i->prid, i->val, arg);
			rbuf[i->prid] |= newbits;
		} else {
			PRINTF("%i: %i != %i\n", i->prid, i->val, arg);
		}
	}

	/* lt */
	/* FIXME: If this were an ordered list, the cost of this would
	 * approach O(log(n)), instead of O(n) as it is now. */
	for(i=s->lt; i; i=i->next){
		if(arg<i->val){
			PRINTF("%i: %i < %i\n", i->prid, i->val, arg);
			rbuf[i->prid] |= newbits;
		} else {
			PRINTF("%i: %i !< %i\n", i->prid, i->val, arg);
		}
	}

	/* gt */
	/* FIXME: If this were an ordered list, the cost of this would
	 * approach O(log(n)), instead of O(n) as it is now. */
	for(i=s->gt; i; i=i->next){
		if(arg>i->val){
			PRINTF("%i: %i > %i\n", i->prid, i->val, arg);
			rbuf[i->prid] |= newbits;
		} else {
			PRINTF("%i: %i !> %i\n", i->prid, i->val, arg);
		}
	}

	/* mask */
	for(i=s->mask; i; i=i->next){
		if((i->val&arg) == i->val){
			PRINTF("MASK: %i: %i & %i == %i\n", i->prid, i->val, arg, i->val);
			rbuf[i->prid] |= newbits;
		} else {
			PRINTF("MASK: %i: %i & %i != %i\n", i->prid, i->val, arg, i->val);
		}
	}

	/* mask_all*/
	for(r=s->mask_all; r; r=r->next){
		if( (r->val1 & arg) == r->val2){
			PRINTF("MASK_ALL: %i: %i & %i == %i\n", r->prid, r->val1, arg, r->val2);
			rbuf[r->prid] |= newbits;
		} else {
			PRINTF("MASK_ALL: %i: %i & %i != %i\n", r->prid, r->val1, arg, r->val2);
		}
	}

	/* range */
	/* FIXME: If this list were ordered by min, then you could easily
	 * cut the work, on average, in half. */
	for(r=s->range; r; r=r->next){
		if( (r->val1 <= arg) && (r->val2 >= arg) ){
			PRINTF("%i: %i < %i < %i\n", r->prid, r->val1, arg, r->val2);
			rbuf[r->prid] |= newbits;
		} else {
			PRINTF("%i: %i !< %i !< %i\n", r->prid, r->val1, arg, r->val2);
		}
	}

	/* modulo*/
	for(r=s->modulo; r; r=r->next){
		if( (arg % r->val1) == r->val2 ) {
			PRINTF("%i: %i %% %i = %i\n", r->prid, arg, r->val1, r->val2);
			rbuf[r->prid] |= newbits;
		} else {
			PRINTF("%i: %i %% %i != %i\n", r->prid, arg, r->val1, r->val2);
		}
	}

	PRINTF("hdt_int32_match: done matching.\n");
	rdtscl(e);
	printf("hdt_int32_match() took %ul cycles.\n", (e-b));

	return(0);
}

char *
hdt_int32_report(u_int32_t val, struct memcon_context *c)
{
	/* 2147483648 * 2 is 10 digits */

	char *r;

	r=c->eng->alloc(c, 12);

	snprintf(r, 11, "%i", val);

	return(r);
}
