#include "ip_lookup.h"
#include "hank.h"
#include <stdio.h>

/* routines to help in managing ip addresses */

static char rcsid[] = "$Id: ip_lookup.c,v 1.5 2001/11/10 01:45:33 tlewis Exp $";

#define PRINTF(a...)

struct ip_btree {
	struct ip_btree *parent;
	struct ip_btree *zero;
	struct ip_btree *one;
	prim_id id;
	char depth;
};

static struct ip_btree sets[1024];
static int nsets = 1;

int nnode=1;
int nprim=0;

prim_id p = 0;

hdt_set hdt_ipv4_addr_create(void)
{
	hdt_set n=nsets++;

	return(n);
}

void 
hdt_ipv4_addr_destroy(hdt_set set)
{
	/* FIXME: unsupported */
}


static void
return_prim_slot(prim_id p)
{
	prims[p-1].id = 0;
	prims[p-1].next=NULL;
	prims[p-1].prev=NULL;
	prims[p-1].parent=NULL;
	nprim--;
}

static struct ip_btree *
create_btree_node(char depth)
{
	struct ip_btree *r;

	r=malloc(sizeof(*r));
	if(r==NULL)
		exit(-1);
	r->zero=NULL;
	r->one=NULL;
	r->matches=NULL;
	r->depth=depth;
	nnode++;
	return(r);
}

static void
release_btree_node(struct ip_btree *r)
{
	free(r);
	nnode--;
}

hdt_prim 
hdt_ipv4_addr_add(hdt_set set, prim_id prid, char *op, char *arg)
{
	struct ipv4_network *n=FOO;
	struct ip_btree *b=&sets[set];
	int i, j;
	char buf[33];
	char kv[2];
	unsigned char mask;
	u_int32_t addr;
	
	split_kv(arg, kv, '/');

	addr = inet_addr(kv[0]);
	mask = atoi(mask);

	if((mask>32)||(mask<0))
		return(0);

	/* grab prim struct */
	l=get_prim_slot();
	if(l==NULL)
		return(0);

	bit_dump(addr, buf, sizeof(buf));
	PRINTF("(Net(%s): (", buf);
	/* find this network's home node in the btree */
	for(i=1;i<mask;i++){
		j=((addr>>(32-i))&1);
		if(j==0){
			if(b->zero==NULL){
				b->zero=create_btree_node(i);
				b->zero->parent=b;
			}
			PRINTF("0");
			b=b->zero;
		} else {
			if(b->one==NULL){
				b->one=create_btree_node(i);
				b->one->parent=b;
			}
			PRINTF("1");
			b=b->one;
		}
	}
	PRINTF(")) ");

	/* append network to node's network list */
	if(b->matches==NULL){
		b->matches=l;
	} else {
		m=b->matches;
		while(m->next!=NULL)
			m=m->next;
		m->next=l;
		l->prev=m;
	}

	/* set parent */
	l->parent=b;

	/* return id */
	return(l->id);
}

void 
hdt_ipv4_del(hdt_set set, hdt_prim p)
{
	struct ip_btree *b=&sets[set], *c;
	int i;

	/* find prim */
	l=&prims[p-1];
	/* find btree node of prim */
	b=l->parent;

	/* remove from matches list */
	if(b->matches==l)
		b->matches=l->next;
	/* zap precursor */
	if(l->prev)
		l->prev->next=l->next;
	/* zap successor */
	if(l->next)
		l->next->prev=l->prev;

	/* recursively: if matches list now empty, then remove node */
	for(i=b->depth;i>0;i++){
		c=b->parent;
		if(b->matches==NULL) {
			/* zap */
			release_btree_node(b);
		} else {
			break;
		}
		b=c;
	}

	/* return prim to free list */
	return_prim_slot(p);

	return(0);
}

struct prim_can *
find_matches(u_int32_t addr)
{
	int i, j, k=0;
	struct prim_can *r=NULL;
	struct ip_btree *b;
	prim_id *res;
	char buf[33];
	unsigned long start, end;

	addr=ntohl(addr);

	b=&root;

	res=malloc(nprim*sizeof(prim_id));
	memset(res, 0, (nprim*sizeof(prim_id)));

	PRINTF("IPv4 Matches (%i nodes, %i prims): ", nnode, nprim);
	/* find this network's home node in the btree */
	bit_dump(addr, buf, sizeof(buf));
	PRINTF("(%s)(", buf);
	rdtscl(start);
	for(i=1;i<=32;i++){
		/* add all matches at this node onto list */
		for(l=b->matches;l;l=l->next)
			res[k++]=l->id;

		j=((addr>>(32-i))&1);
		if(j==0){
			PRINTF("0");
			b=b->zero;
		} else {
			PRINTF("1");
			b=b->one;
		}
		if(b==NULL)
			break;
	}
	PRINTF(") ");
	rdtscl(end);

	for(j=0;res[j]!=0;j++){
		printf("%i, ", res[j]);
	}
	printf("\nDepth/Cycles: %i/%li\n", i, (end-start));

	free(res);

	return(r);
}


