#ifndef   PAENGINE_H
#define   PAENGINE_H

/*******************************************************************
 * paengine.h: support for packet acquisition engines              *
 * (c) 2001, Todd Lewis, <tlewis@mindspring.com>                   *
 * Released under GNU GPL v2.                                            *
 * Earlier paengine work supported by: www.secureworks.net         *
 *******************************************************************/

#include "event.h"
#include "packet.h"
#include "smodule.h"
#include "types.h"

/*! smodule version numbers. */
/*! Bump these appropriately if the data structures in this header file are changed. */
#define PAVMAJ 2 /*! v1 was the prototype */
#define PAVMIN 1

/*! Capabilities of pa engines. */
/*! These are two's complements, fit for ORing together into a set. */
#define PA_CAP_NONE     0      /* !< no capabilities */
#define PA_CAP_DUMP     1 << 0 /* !< can dump packets */
#define PA_CAP_DISCARD  1 << 1 /* !< can discard packets */
#define PA_CAP_REWRITE  1 << 2 /* !< can rewrite packets */

/*! return codes from get_packet */
#define PA_PACKET_CODE_OK 0         /* !< Packet returned succesfully */
#define PA_PACKET_CODE_NOPACKET 1   /* !< Non-error condition, but no packet returned; cf. EAGAIN */
#define PA_PACKET_CODE_EOF 2        /* !< End of file reached, as with a pcap capture file */
#define PA_PACKET_CODE_ERR 7        /* !< Other, misc. error; see docs for that pae */

/*! packet verdicts */
#define PA_PACKET_VERDICT_NO_ACTION 1<<0 /* !< No action specified for packet */
#define PA_PACKET_VERDICT_ACCEPT    1<<1 /* !< Accept packet */
#define PA_PACKET_VERDICT_DROP      1<<2 /* !< Drop packet */
#define PA_PACKET_VERDICT_REJECT    1<<3 /* !< Reject packet (send nasty ICMP, etc.) */
#define PA_PACKET_VERDICT_REWRITE   1<<4 /* !< Rewrite the packet  */

/*! \struct
    \brief The main packet acquisition module.

This is the structure exported as the "interface" from the smodule.
 */

struct paengine {
	struct smodule s;
   int (*init)(void); /* initialize driver; 0 on success */
   struct packet* (*get_packet)(event_id id); /* id of initial event */
   int (*dispose_packet)(struct packet *);
   int (*close)(void); /* close driver; 0 on success */
   int (*get_capabilities)(void); /* caps of the driver; called after conf() */
   int (*configure)(char * config); /* config file entry for engine; 0 on
                                     * success, other on error */
};

#endif /* PAENGINE_H */
