/* Copyright 2001, Todd Lewis <tlewis@mindspring.com>
 * All rights reserved. */

#include "hank.h"
#include "hdt_blob.h"
#include "hdt_int32.h"
#include "packet.h"
#include "protocol_engine.h"
#include "protocols.h"
#include "support.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.h"

#include "ntp.hi"


static char rcsid[] = "$Id: ntp.xml,v 1.3 2001/10/29 07:35:20 tlewis Exp $";


extern struct smodule *ntp[];


extern struct protocol_engine ntp_pe;


static u_int32_t protocols[] = {

  ((HPN_TCP << 24) | 123),
  ((HPN_UDP << 24) | 123),


  0
};

struct simple_prim
{
  prim_id prim;
  struct simple_prim *next;
};

static struct
{
  struct simple_prim *proto_self;


  struct simple_prim *present_Hdr;


}
present_sections;

enum NTP_FIELD
{
  NTP_SELF,


  NTP_FIELD_Hdr,
  NTP_FIELD_Hdr_li_vn_mode,
  NTP_FIELD_Hdr_stratum,
  NTP_FIELD_Hdr_ppoll,
  NTP_FIELD_Hdr_precision,
  NTP_FIELD_Hdr_rootdelay,
  NTP_FIELD_Hdr_rootdispersion,
  NTP_FIELD_Hdr_refid,
  NTP_FIELD_Hdr_reftime,
  NTP_FIELD_Hdr_org,
  NTP_FIELD_Hdr_rec,
  NTP_FIELD_Hdr_xmt,
  NTP_FIELD_Hdr_keyid1,
  NTP_FIELD_Hdr_keyid2,
  NTP_FIELD_Hdr_keyid3,
  NTP_FIELD_Hdr_mac,


  NTP_FIELD_INVALID
};

enum NTP_SECTIONS
{


  NTP_SECTION_Hdr,


  NTP_SECTION_INVALID
};

struct NTP_null
{
  int empty;


};
struct NTP_Hdr
{
  u_int8_t li_vn_mode;
  u_int8_t stratum;
  u_int8_t ppoll;
  u_int8_t precision;
  u_int32_t rootdelay;
  u_int32_t rootdispersion;
  u_int32_t refid;
  u_int8_t reftime;
  u_int8_t org;
  u_int8_t rec;
  u_int8_t xmt;
  u_int32_t keyid1;
  u_int32_t keyid2;
  u_int32_t keyid3;
  u_int32_t mac:24;


};


static char hexes[] = "0123456789abcdef";

static int gprid = 0;

static hdt_set hdt_sets[NTP_FIELD_INVALID];


static int ntp_stock_reg_prim (int field, char *op, char *arg);
static int ntp_stock_unreg_prim (prim_id i);
static int ntp_stock_decompose (struct packet *p);
static int ntp_stock_find_matches (struct capsule *cp, struct prim_can *c);
static int ntp_stock_rewrite (struct capsule *c, void *new_payload,
			      u_int32_t len);
static char *ntp_stock_report_field (struct capsule *p, int field,
				     struct memcon_context *m);
static int ntp_field_lookup (char *section, char *field);

struct protocol_engine ntp_pe = {
  {
   "pe",
   "ntp",
   PEVMIN,
   PEVMAJ,
   &ntp_pe},
  &ntp_stock_decompose,
  &ntp_stock_reg_prim,
  &ntp_stock_unreg_prim,
  &ntp_stock_find_matches,
  &ntp_stock_rewrite,
  &ntp_stock_report_field,
  &ntp_field_lookup,
  &protocols[0],
  0
};


struct smodule *ntp[] = {

  &(ntp_pe.s),
  NULL
};


static int
ntp_field_lookup (char *section, char *field)
{
  int p = -1;

  if ((section == NULL) && (field == NULL))
    {
      p = NTP_SELF;
    }
  else if (section != NULL && !(strcmp (section, "INVALID")))
    {
      p = NTP_FIELD_INVALID;


    }
  else if ((section == NULL) || (!(strcmp (section, "Hdr"))))
    {
      if (field == NULL)
	p = NTP_FIELD_Hdr;
      else if (!(strcmp (field, "li_vn_mode")))
	p = NTP_FIELD_Hdr_li_vn_mode;
      else if (!(strcmp (field, "stratum")))
	p = NTP_FIELD_Hdr_stratum;
      else if (!(strcmp (field, "ppoll")))
	p = NTP_FIELD_Hdr_ppoll;
      else if (!(strcmp (field, "precision")))
	p = NTP_FIELD_Hdr_precision;
      else if (!(strcmp (field, "rootdelay")))
	p = NTP_FIELD_Hdr_rootdelay;
      else if (!(strcmp (field, "rootdispersion")))
	p = NTP_FIELD_Hdr_rootdispersion;
      else if (!(strcmp (field, "refid")))
	p = NTP_FIELD_Hdr_refid;
      else if (!(strcmp (field, "reftime")))
	p = NTP_FIELD_Hdr_reftime;
      else if (!(strcmp (field, "org")))
	p = NTP_FIELD_Hdr_org;
      else if (!(strcmp (field, "rec")))
	p = NTP_FIELD_Hdr_rec;
      else if (!(strcmp (field, "xmt")))
	p = NTP_FIELD_Hdr_xmt;
      else if (!(strcmp (field, "keyid1")))
	p = NTP_FIELD_Hdr_keyid1;
      else if (!(strcmp (field, "keyid2")))
	p = NTP_FIELD_Hdr_keyid2;
      else if (!(strcmp (field, "keyid3")))
	p = NTP_FIELD_Hdr_keyid3;
      else if (!(strcmp (field, "mac")))
	p = NTP_FIELD_Hdr_mac;


    }

  if (p == -1)
    {
      printf ("NTP: Unrecognized field: \" %s \".  Returning error.\n",
	      field);
      return (NTP_FIELD_INVALID);
    }

  ;
  return (p);
}


static int
ntp_stock_decompose (struct packet *p)
{
  struct capsule *e = p->encaps, *f;
  struct NTP_Hdr *h;
  struct section *s;


  while (e->next)
    e = e->next;

  h = (struct NTP_Hdr *) e->payload;


  append_encap (p, protocols[0], e->payload_length, e->payload);
  f = e->next;
  f->pe = &ntp_pe;

  if ((s = p->c->eng->alloc (p->c, sizeof (*s))) == NULL)
    exit (-1);
  s->section_id = NTP_SECTION_Hdr;
  s->content = h;
  s->next = NULL;
  f->sections = s;
  f->length = e->payload_length;


  return (0);
}

static int
ntp_stock_reg_prim (int field, char *op, char *arg)
{
  int prid = ++gprid;


  ;

  if (field == NTP_FIELD_INVALID)
    {
      printf
	("NTP: Error.  Received invalid field in reg_prim(), op=%s, arg=%s.\n",
	 op, arg);
      return (-1);
    }


  if (!strcmp (op, "present"))
    {
      struct simple_prim *p;

      if (!(p = malloc (sizeof (*p))))
	exit (-1);
      p->prim = prid;

      if (field == NTP_SELF)
	{

	  p->next = present_sections.proto_self;
	  present_sections.proto_self = p;
	  return (prid);


	}
      else if (field == NTP_FIELD_Hdr)
	{
	  p->next = present_sections.present_Hdr;
	  present_sections.present_Hdr = p;
	  return (prid);


	}

      printf ("That shit (%i) ain't keepin' it real.\n", field);
      return (-1);
    }


  if (field == NTP_FIELD_Hdr_li_vn_mode)
    {
      if (hdt_sets[NTP_FIELD_Hdr_li_vn_mode] == 0)
	{
	  hdt_sets[NTP_FIELD_Hdr_li_vn_mode] = hdt_int32_create ();
	  if (hdt_sets[NTP_FIELD_Hdr_li_vn_mode] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "li_vn_mode");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[NTP_FIELD_Hdr_li_vn_mode], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ntp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == NTP_FIELD_Hdr_stratum)
    {
      if (hdt_sets[NTP_FIELD_Hdr_stratum] == 0)
	{
	  hdt_sets[NTP_FIELD_Hdr_stratum] = hdt_int32_create ();
	  if (hdt_sets[NTP_FIELD_Hdr_stratum] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "stratum");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[NTP_FIELD_Hdr_stratum], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ntp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == NTP_FIELD_Hdr_ppoll)
    {
      if (hdt_sets[NTP_FIELD_Hdr_ppoll] == 0)
	{
	  hdt_sets[NTP_FIELD_Hdr_ppoll] = hdt_int32_create ();
	  if (hdt_sets[NTP_FIELD_Hdr_ppoll] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "ppoll");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[NTP_FIELD_Hdr_ppoll], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ntp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == NTP_FIELD_Hdr_precision)
    {
      if (hdt_sets[NTP_FIELD_Hdr_precision] == 0)
	{
	  hdt_sets[NTP_FIELD_Hdr_precision] = hdt_int32_create ();
	  if (hdt_sets[NTP_FIELD_Hdr_precision] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "precision");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[NTP_FIELD_Hdr_precision], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ntp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == NTP_FIELD_Hdr_rootdelay)
    {
      if (hdt_sets[NTP_FIELD_Hdr_rootdelay] == 0)
	{
	  hdt_sets[NTP_FIELD_Hdr_rootdelay] = hdt_int32_create ();
	  if (hdt_sets[NTP_FIELD_Hdr_rootdelay] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "rootdelay");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[NTP_FIELD_Hdr_rootdelay], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ntp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == NTP_FIELD_Hdr_rootdispersion)
    {
      if (hdt_sets[NTP_FIELD_Hdr_rootdispersion] == 0)
	{
	  hdt_sets[NTP_FIELD_Hdr_rootdispersion] = hdt_int32_create ();
	  if (hdt_sets[NTP_FIELD_Hdr_rootdispersion] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "rootdispersion");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[NTP_FIELD_Hdr_rootdispersion], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ntp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == NTP_FIELD_Hdr_refid)
    {
      if (hdt_sets[NTP_FIELD_Hdr_refid] == 0)
	{
	  hdt_sets[NTP_FIELD_Hdr_refid] = hdt_int32_create ();
	  if (hdt_sets[NTP_FIELD_Hdr_refid] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "refid");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[NTP_FIELD_Hdr_refid], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ntp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == NTP_FIELD_Hdr_reftime)
    {
      if (hdt_sets[NTP_FIELD_Hdr_reftime] == 0)
	{
	  hdt_sets[NTP_FIELD_Hdr_reftime] = hdt_int32_create ();
	  if (hdt_sets[NTP_FIELD_Hdr_reftime] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "reftime");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[NTP_FIELD_Hdr_reftime], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ntp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == NTP_FIELD_Hdr_org)
    {
      if (hdt_sets[NTP_FIELD_Hdr_org] == 0)
	{
	  hdt_sets[NTP_FIELD_Hdr_org] = hdt_int32_create ();
	  if (hdt_sets[NTP_FIELD_Hdr_org] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "org");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[NTP_FIELD_Hdr_org], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ntp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == NTP_FIELD_Hdr_rec)
    {
      if (hdt_sets[NTP_FIELD_Hdr_rec] == 0)
	{
	  hdt_sets[NTP_FIELD_Hdr_rec] = hdt_int32_create ();
	  if (hdt_sets[NTP_FIELD_Hdr_rec] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "rec");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[NTP_FIELD_Hdr_rec], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ntp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == NTP_FIELD_Hdr_xmt)
    {
      if (hdt_sets[NTP_FIELD_Hdr_xmt] == 0)
	{
	  hdt_sets[NTP_FIELD_Hdr_xmt] = hdt_int32_create ();
	  if (hdt_sets[NTP_FIELD_Hdr_xmt] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "xmt");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[NTP_FIELD_Hdr_xmt], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ntp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == NTP_FIELD_Hdr_keyid1)
    {
      if (hdt_sets[NTP_FIELD_Hdr_keyid1] == 0)
	{
	  hdt_sets[NTP_FIELD_Hdr_keyid1] = hdt_int32_create ();
	  if (hdt_sets[NTP_FIELD_Hdr_keyid1] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "keyid1");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[NTP_FIELD_Hdr_keyid1], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ntp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == NTP_FIELD_Hdr_keyid2)
    {
      if (hdt_sets[NTP_FIELD_Hdr_keyid2] == 0)
	{
	  hdt_sets[NTP_FIELD_Hdr_keyid2] = hdt_int32_create ();
	  if (hdt_sets[NTP_FIELD_Hdr_keyid2] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "keyid2");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[NTP_FIELD_Hdr_keyid2], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ntp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == NTP_FIELD_Hdr_keyid3)
    {
      if (hdt_sets[NTP_FIELD_Hdr_keyid3] == 0)
	{
	  hdt_sets[NTP_FIELD_Hdr_keyid3] = hdt_int32_create ();
	  if (hdt_sets[NTP_FIELD_Hdr_keyid3] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "keyid3");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[NTP_FIELD_Hdr_keyid3], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ntp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == NTP_FIELD_Hdr_mac)
    {
      if (hdt_sets[NTP_FIELD_Hdr_mac] == 0)
	{
	  hdt_sets[NTP_FIELD_Hdr_mac] = hdt_int32_create ();
	  if (hdt_sets[NTP_FIELD_Hdr_mac] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "mac");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[NTP_FIELD_Hdr_mac], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ntp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }


  return (-1);
}

static int
ntp_stock_unreg_prim (prim_id i)
{

  return (-1);
}

static int
ntp_stock_rewrite (struct capsule *c, void *new_payload, u_int32_t len)
{
  if ((c != NULL) && (new_payload != NULL))
    len++;
  return (-1);
}

static char *
ntp_stock_report_field (struct capsule *cp, int field,
			struct memcon_context *m)
{
  u_int32_t t;
  char *r, buf[1024];
  struct section *sec, *sec2;

  buf[0] = 0;

  switch (field)
    {


    case (NTP_FIELD_Hdr_li_vn_mode):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == NTP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct NTP_Hdr *) (sec->content))->li_vn_mode);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (NTP_FIELD_Hdr_stratum):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == NTP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct NTP_Hdr *) (sec->content))->stratum);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (NTP_FIELD_Hdr_ppoll):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == NTP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct NTP_Hdr *) (sec->content))->ppoll);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (NTP_FIELD_Hdr_precision):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == NTP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct NTP_Hdr *) (sec->content))->precision);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (NTP_FIELD_Hdr_rootdelay):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == NTP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct NTP_Hdr *) (sec->content))->rootdelay));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (NTP_FIELD_Hdr_rootdispersion):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == NTP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct NTP_Hdr *) (sec->content))->
			   rootdispersion));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (NTP_FIELD_Hdr_refid):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == NTP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct NTP_Hdr *) (sec->content))->refid));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (NTP_FIELD_Hdr_reftime):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == NTP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct NTP_Hdr *) (sec->content))->reftime);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (NTP_FIELD_Hdr_org):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == NTP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct NTP_Hdr *) (sec->content))->org);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (NTP_FIELD_Hdr_rec):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == NTP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct NTP_Hdr *) (sec->content))->rec);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (NTP_FIELD_Hdr_xmt):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == NTP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct NTP_Hdr *) (sec->content))->xmt);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (NTP_FIELD_Hdr_keyid1):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == NTP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct NTP_Hdr *) (sec->content))->keyid1));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (NTP_FIELD_Hdr_keyid2):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == NTP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct NTP_Hdr *) (sec->content))->keyid2));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (NTP_FIELD_Hdr_keyid3):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == NTP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct NTP_Hdr *) (sec->content))->keyid3));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (NTP_FIELD_Hdr_mac):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == NTP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct NTP_Hdr *) (sec->content))->mac));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;


    default:

      return (NULL);
    }
  t = (strlen (buf) + 1);
  r = m->eng->alloc (m, t);
  strncpy (r, buf, t);
  return (r);
}

static int
ntp_stock_find_matches (struct capsule *cp, struct prim_can *p)
{
  struct section *s;
  struct memcon_context *c = cp->parent->c;
  u_int32_t v;
  unsigned long beg, end;

  if (p)
    return (-1);

  if (cp->matches == NULL && gprid)
    {
      if (!(cp->matches = c->eng->alloc (c, gprid + 1)))
	exit (-1);
      memset (cp->matches, 0, gprid + 1);

      cp->matches[0] = 1;
    }


  for (s = cp->sections; s; s = s->next)
    {
      struct simple_prim *sp;

      for (sp = present_sections.proto_self; sp; sp = sp->next)
	{
	  ;
	  cp->matches[sp->prim] |= 1;
	}


      switch (s->section_id)
	{
	case NTP_SECTION_INVALID:
	  return (-1);


	  break;
	case NTP_SECTION_Hdr:;
	  for (sp = present_sections.present_Hdr; sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[NTP_FIELD_Hdr_li_vn_mode])
	    {;
	      v = ((struct NTP_Hdr *) (s->content))->li_vn_mode;;
	      if (hdt_int32_match
		  (hdt_sets[NTP_FIELD_Hdr_li_vn_mode], NULL, v, cp->matches,
		   1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "li_vn_mode", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[NTP_FIELD_Hdr_stratum])
	    {;
	      v = ((struct NTP_Hdr *) (s->content))->stratum;;
	      if (hdt_int32_match
		  (hdt_sets[NTP_FIELD_Hdr_stratum], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "stratum", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[NTP_FIELD_Hdr_ppoll])
	    {;
	      v = ((struct NTP_Hdr *) (s->content))->ppoll;;
	      if (hdt_int32_match
		  (hdt_sets[NTP_FIELD_Hdr_ppoll], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "ppoll", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[NTP_FIELD_Hdr_precision])
	    {;
	      v = ((struct NTP_Hdr *) (s->content))->precision;;
	      if (hdt_int32_match
		  (hdt_sets[NTP_FIELD_Hdr_precision], NULL, v, cp->matches,
		   1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "precision", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[NTP_FIELD_Hdr_rootdelay])
	    {;
	      v = ntohl (((struct NTP_Hdr *) (s->content))->rootdelay);;
	      if (hdt_int32_match
		  (hdt_sets[NTP_FIELD_Hdr_rootdelay], NULL, v, cp->matches,
		   1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "rootdelay", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[NTP_FIELD_Hdr_rootdispersion])
	    {;
	      v = ntohl (((struct NTP_Hdr *) (s->content))->rootdispersion);;
	      if (hdt_int32_match
		  (hdt_sets[NTP_FIELD_Hdr_rootdispersion], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "rootdispersion", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[NTP_FIELD_Hdr_refid])
	    {;
	      v = ntohl (((struct NTP_Hdr *) (s->content))->refid);;
	      if (hdt_int32_match
		  (hdt_sets[NTP_FIELD_Hdr_refid], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "refid", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[NTP_FIELD_Hdr_reftime])
	    {;
	      v = ((struct NTP_Hdr *) (s->content))->reftime;;
	      if (hdt_int32_match
		  (hdt_sets[NTP_FIELD_Hdr_reftime], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "reftime", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[NTP_FIELD_Hdr_org])
	    {;
	      v = ((struct NTP_Hdr *) (s->content))->org;;
	      if (hdt_int32_match
		  (hdt_sets[NTP_FIELD_Hdr_org], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr", "org",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[NTP_FIELD_Hdr_rec])
	    {;
	      v = ((struct NTP_Hdr *) (s->content))->rec;;
	      if (hdt_int32_match
		  (hdt_sets[NTP_FIELD_Hdr_rec], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr", "rec",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[NTP_FIELD_Hdr_xmt])
	    {;
	      v = ((struct NTP_Hdr *) (s->content))->xmt;;
	      if (hdt_int32_match
		  (hdt_sets[NTP_FIELD_Hdr_xmt], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr", "xmt",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[NTP_FIELD_Hdr_keyid1])
	    {;
	      v = ntohl (((struct NTP_Hdr *) (s->content))->keyid1);;
	      if (hdt_int32_match
		  (hdt_sets[NTP_FIELD_Hdr_keyid1], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "keyid1", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[NTP_FIELD_Hdr_keyid2])
	    {;
	      v = ntohl (((struct NTP_Hdr *) (s->content))->keyid2);;
	      if (hdt_int32_match
		  (hdt_sets[NTP_FIELD_Hdr_keyid2], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "keyid2", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[NTP_FIELD_Hdr_keyid3])
	    {;
	      v = ntohl (((struct NTP_Hdr *) (s->content))->keyid3);;
	      if (hdt_int32_match
		  (hdt_sets[NTP_FIELD_Hdr_keyid3], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "keyid3", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[NTP_FIELD_Hdr_mac])
	    {;
	      v = ntohl (((struct NTP_Hdr *) (s->content))->mac);;
	      if (hdt_int32_match
		  (hdt_sets[NTP_FIELD_Hdr_mac], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr", "mac",
		  (end - beg));


	  break;
	default:
	  printf ("Unrecognized section: %i (not %i).\n", s->section_id,
		  NTP_SECTION_Hdr);
	  break;
	}
    }

  return (-1);
}
