/* Copyright 2001, Todd Lewis <tlewis@mindspring.com>
 * All rights reserved. */

#include "hank.h"
#include "hdt_blob.h"
#include "hdt_int32.h"
#include "packet.h"
#include "protocol_engine.h"
#include "protocols.h"
#include "support.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.h"

#include "tcp.hi"


static char rcsid[] = "$Id: tcp.xml,v 1.5 2001/10/31 21:20:01 tlewis Exp $";


extern struct smodule *tcp[];


extern struct protocol_engine tcp_pe;


static u_int32_t protocols[] = {

  ((HPN_IPV4 << 24) | 6),
  ((HPN_IPV6 << 24) | 6),


  0
};

struct simple_prim
{
  prim_id prim;
  struct simple_prim *next;
};

static struct
{
  struct simple_prim *proto_self;


  struct simple_prim *present_Hdr;


}
present_sections;

enum TCP_FIELD
{
  TCP_SELF,


  TCP_FIELD_Hdr,
  TCP_FIELD_Hdr_source_port,
  TCP_FIELD_Hdr_destination_port,
  TCP_FIELD_Hdr_sequence_no,
  TCP_FIELD_Hdr_ack_no,
  TCP_FIELD_Hdr_res1,
  TCP_FIELD_Hdr_doff,
  TCP_FIELD_Hdr_flags,
  TCP_FIELD_Hdr_window,
  TCP_FIELD_Hdr_check,
  TCP_FIELD_Hdr_urg_ptr,
  TCP_FIELD_Hdr_payload,


  TCP_FIELD_INVALID
};

enum TCP_SECTIONS
{


  TCP_SECTION_Hdr,


  TCP_SECTION_INVALID
};

struct TCP_null
{
  int empty;


};
struct TCP_Hdr
{
  u_int16_t source_port;
  u_int16_t destination_port;
  u_int32_t sequence_no;
  u_int32_t ack_no;
  u_int32_t res1:4;
  u_int32_t doff:4;
  u_int8_t flags;
  u_int16_t window;
  u_int16_t check;
  u_int16_t urg_ptr;
  unsigned char payload;


};


static char hexes[] = "0123456789abcdef";

static int gprid = 0;

static hdt_set hdt_sets[TCP_FIELD_INVALID];


static int tcp_stock_reg_prim (int field, char *op, char *arg);
static int tcp_stock_unreg_prim (prim_id i);
static int tcp_stock_decompose (struct packet *p);
static int tcp_stock_find_matches (struct capsule *cp, struct prim_can *c);
static int tcp_stock_rewrite (struct capsule *c, void *new_payload,
			      u_int32_t len);
static char *tcp_stock_report_field (struct capsule *p, int field,
				     struct memcon_context *m);
static int tcp_field_lookup (char *section, char *field);

struct protocol_engine tcp_pe = {
  {
   "pe",
   "tcp",
   PEVMIN,
   PEVMAJ,
   &tcp_pe},
  &tcp_stock_decompose,
  &tcp_stock_reg_prim,
  &tcp_stock_unreg_prim,
  &tcp_stock_find_matches,
  &tcp_stock_rewrite,
  &tcp_stock_report_field,
  &tcp_field_lookup,
  &protocols[0],
  0
};


struct smodule *tcp[] = {

  &(tcp_pe.s),
  NULL
};


static int
tcp_field_lookup (char *section, char *field)
{
  int p = -1;

  if ((section == NULL) && (field == NULL))
    {
      p = TCP_SELF;
    }
  else if (section != NULL && !(strcmp (section, "INVALID")))
    {
      p = TCP_FIELD_INVALID;


    }
  else if ((section == NULL) || (!(strcmp (section, "Hdr"))))
    {
      if (field == NULL)
	p = TCP_FIELD_Hdr;
      else if (!(strcmp (field, "source_port")))
	p = TCP_FIELD_Hdr_source_port;
      else if (!(strcmp (field, "destination_port")))
	p = TCP_FIELD_Hdr_destination_port;
      else if (!(strcmp (field, "sequence_no")))
	p = TCP_FIELD_Hdr_sequence_no;
      else if (!(strcmp (field, "ack_no")))
	p = TCP_FIELD_Hdr_ack_no;
      else if (!(strcmp (field, "res1")))
	p = TCP_FIELD_Hdr_res1;
      else if (!(strcmp (field, "doff")))
	p = TCP_FIELD_Hdr_doff;
      else if (!(strcmp (field, "flags")))
	p = TCP_FIELD_Hdr_flags;
      else if (!(strcmp (field, "window")))
	p = TCP_FIELD_Hdr_window;
      else if (!(strcmp (field, "check")))
	p = TCP_FIELD_Hdr_check;
      else if (!(strcmp (field, "urg_ptr")))
	p = TCP_FIELD_Hdr_urg_ptr;
      else if (!(strcmp (field, "payload")))
	p = TCP_FIELD_Hdr_payload;


    }

  if (p == -1)
    {
      printf ("TCP: Unrecognized field: \" %s \".  Returning error.\n",
	      field);
      return (TCP_FIELD_INVALID);
    }

  ;
  return (p);
}


static int
tcp_stock_decompose (struct packet *p)
{
  struct capsule *e = p->encaps, *f;
  struct TCP_Hdr *h;
  struct section *s;


  while (e->next)
    e = e->next;

  h = (struct TCP_Hdr *) e->payload;


  append_encap (p, protocols[0], e->payload_length, e->payload);
  f = e->next;
  f->pe = &tcp_pe;

  if ((s = p->c->eng->alloc (p->c, sizeof (*s))) == NULL)
    exit (-1);
  s->section_id = TCP_SECTION_Hdr;
  s->content = h;
  s->next = NULL;
  f->sections = s;
  f->length = e->payload_length;


  f->payload_length = (f->length - (offsetof (struct TCP_Hdr, payload)));
  f->payload_protocol =
    (u_int32_t) ((HPN_TCP << 24) |
		 ((ntohs (h->source_port) <
		   ntohs (h->destination_port)) ? ntohs (h->
							 source_port) :
		  ntohs (h->destination_port)));
  f->payload = &(e->payload[(offsetof (struct TCP_Hdr, payload))]);


  return (0);
}

static int
tcp_stock_reg_prim (int field, char *op, char *arg)
{
  int prid = ++gprid;


  ;

  if (field == TCP_FIELD_INVALID)
    {
      printf
	("TCP: Error.  Received invalid field in reg_prim(), op=%s, arg=%s.\n",
	 op, arg);
      return (-1);
    }


  if (!strcmp (op, "present"))
    {
      struct simple_prim *p;

      if (!(p = malloc (sizeof (*p))))
	exit (-1);
      p->prim = prid;

      if (field == TCP_SELF)
	{

	  p->next = present_sections.proto_self;
	  present_sections.proto_self = p;
	  return (prid);


	}
      else if (field == TCP_FIELD_Hdr)
	{
	  p->next = present_sections.present_Hdr;
	  present_sections.present_Hdr = p;
	  return (prid);


	}

      printf ("That shit (%i) ain't keepin' it real.\n", field);
      return (-1);
    }


  if (field == TCP_FIELD_Hdr_source_port)
    {
      if (hdt_sets[TCP_FIELD_Hdr_source_port] == 0)
	{
	  hdt_sets[TCP_FIELD_Hdr_source_port] = hdt_int32_create ();
	  if (hdt_sets[TCP_FIELD_Hdr_source_port] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "source_port");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[TCP_FIELD_Hdr_source_port], prid, op, arg))
	{
	  printf ("Error adding primitive %s in tcp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == TCP_FIELD_Hdr_destination_port)
    {
      if (hdt_sets[TCP_FIELD_Hdr_destination_port] == 0)
	{
	  hdt_sets[TCP_FIELD_Hdr_destination_port] = hdt_int32_create ();
	  if (hdt_sets[TCP_FIELD_Hdr_destination_port] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "destination_port");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[TCP_FIELD_Hdr_destination_port], prid, op, arg))
	{
	  printf ("Error adding primitive %s in tcp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == TCP_FIELD_Hdr_sequence_no)
    {
      if (hdt_sets[TCP_FIELD_Hdr_sequence_no] == 0)
	{
	  hdt_sets[TCP_FIELD_Hdr_sequence_no] = hdt_int32_create ();
	  if (hdt_sets[TCP_FIELD_Hdr_sequence_no] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "sequence_no");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[TCP_FIELD_Hdr_sequence_no], prid, op, arg))
	{
	  printf ("Error adding primitive %s in tcp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == TCP_FIELD_Hdr_ack_no)
    {
      if (hdt_sets[TCP_FIELD_Hdr_ack_no] == 0)
	{
	  hdt_sets[TCP_FIELD_Hdr_ack_no] = hdt_int32_create ();
	  if (hdt_sets[TCP_FIELD_Hdr_ack_no] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "ack_no");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[TCP_FIELD_Hdr_ack_no], prid, op, arg))
	{
	  printf ("Error adding primitive %s in tcp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == TCP_FIELD_Hdr_res1)
    {
      if (hdt_sets[TCP_FIELD_Hdr_res1] == 0)
	{
	  hdt_sets[TCP_FIELD_Hdr_res1] = hdt_int32_create ();
	  if (hdt_sets[TCP_FIELD_Hdr_res1] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "res1");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[TCP_FIELD_Hdr_res1], prid, op, arg))
	{
	  printf ("Error adding primitive %s in tcp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == TCP_FIELD_Hdr_doff)
    {
      if (hdt_sets[TCP_FIELD_Hdr_doff] == 0)
	{
	  hdt_sets[TCP_FIELD_Hdr_doff] = hdt_int32_create ();
	  if (hdt_sets[TCP_FIELD_Hdr_doff] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "doff");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[TCP_FIELD_Hdr_doff], prid, op, arg))
	{
	  printf ("Error adding primitive %s in tcp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == TCP_FIELD_Hdr_flags)
    {
      if (hdt_sets[TCP_FIELD_Hdr_flags] == 0)
	{
	  hdt_sets[TCP_FIELD_Hdr_flags] = hdt_int32_create ();
	  if (hdt_sets[TCP_FIELD_Hdr_flags] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "flags");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[TCP_FIELD_Hdr_flags], prid, op, arg))
	{
	  printf ("Error adding primitive %s in tcp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == TCP_FIELD_Hdr_window)
    {
      if (hdt_sets[TCP_FIELD_Hdr_window] == 0)
	{
	  hdt_sets[TCP_FIELD_Hdr_window] = hdt_int32_create ();
	  if (hdt_sets[TCP_FIELD_Hdr_window] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "window");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[TCP_FIELD_Hdr_window], prid, op, arg))
	{
	  printf ("Error adding primitive %s in tcp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == TCP_FIELD_Hdr_check)
    {
      if (hdt_sets[TCP_FIELD_Hdr_check] == 0)
	{
	  hdt_sets[TCP_FIELD_Hdr_check] = hdt_int32_create ();
	  if (hdt_sets[TCP_FIELD_Hdr_check] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "check");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[TCP_FIELD_Hdr_check], prid, op, arg))
	{
	  printf ("Error adding primitive %s in tcp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == TCP_FIELD_Hdr_urg_ptr)
    {
      if (hdt_sets[TCP_FIELD_Hdr_urg_ptr] == 0)
	{
	  hdt_sets[TCP_FIELD_Hdr_urg_ptr] = hdt_int32_create ();
	  if (hdt_sets[TCP_FIELD_Hdr_urg_ptr] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "urg_ptr");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[TCP_FIELD_Hdr_urg_ptr], prid, op, arg))
	{
	  printf ("Error adding primitive %s in tcp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == TCP_FIELD_Hdr_payload)
    {
      if (hdt_sets[TCP_FIELD_Hdr_payload] == 0)
	{
	  hdt_sets[TCP_FIELD_Hdr_payload] = hdt_blob_create ();;
	  if (hdt_sets[TCP_FIELD_Hdr_payload] == 0)
	    {
	      printf ("AIEE!  Can't add match set for %s.%s!  Exiting.\n",
		      "Hdr", "payload");
	      exit (-1);
	    }
	}
      if (hdt_blob_add (hdt_sets[TCP_FIELD_Hdr_payload], prid, op, arg))
	{
	  printf ("Error adding primitive %s in tcp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }


  return (-1);
}

static int
tcp_stock_unreg_prim (prim_id i)
{

  return (-1);
}

static int
tcp_stock_rewrite (struct capsule *c, void *new_payload, u_int32_t len)
{
  if ((c != NULL) && (new_payload != NULL))
    len++;
  return (-1);
}

static char *
tcp_stock_report_field (struct capsule *cp, int field,
			struct memcon_context *m)
{
  u_int32_t t;
  char *r, buf[1024];
  struct section *sec, *sec2;

  buf[0] = 0;

  switch (field)
    {


    case (TCP_FIELD_Hdr_source_port):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == TCP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct TCP_Hdr *) (sec->content))->source_port));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (TCP_FIELD_Hdr_destination_port):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == TCP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct TCP_Hdr *) (sec->content))->
			   destination_port));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (TCP_FIELD_Hdr_sequence_no):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == TCP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct TCP_Hdr *) (sec->content))->sequence_no));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (TCP_FIELD_Hdr_ack_no):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == TCP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct TCP_Hdr *) (sec->content))->ack_no));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (TCP_FIELD_Hdr_res1):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == TCP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct TCP_Hdr *) (sec->content))->res1);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (TCP_FIELD_Hdr_doff):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == TCP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct TCP_Hdr *) (sec->content))->doff);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (TCP_FIELD_Hdr_flags):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == TCP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct TCP_Hdr *) (sec->content))->flags);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (TCP_FIELD_Hdr_window):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == TCP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct TCP_Hdr *) (sec->content))->window));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (TCP_FIELD_Hdr_check):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == TCP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct TCP_Hdr *) (sec->content))->check));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (TCP_FIELD_Hdr_urg_ptr):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == TCP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct TCP_Hdr *) (sec->content))->urg_ptr));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;


    default:

      return (NULL);
    }
  t = (strlen (buf) + 1);
  r = m->eng->alloc (m, t);
  strncpy (r, buf, t);
  return (r);
}

static int
tcp_stock_find_matches (struct capsule *cp, struct prim_can *p)
{
  struct section *s;
  struct memcon_context *c = cp->parent->c;
  u_int32_t v;
  unsigned long beg, end;

  if (p)
    return (-1);

  if (cp->matches == NULL && gprid)
    {
      if (!(cp->matches = c->eng->alloc (c, gprid + 1)))
	exit (-1);
      memset (cp->matches, 0, gprid + 1);

      cp->matches[0] = 1;
    }


  for (s = cp->sections; s; s = s->next)
    {
      struct simple_prim *sp;

      for (sp = present_sections.proto_self; sp; sp = sp->next)
	{
	  ;
	  cp->matches[sp->prim] |= 1;
	}


      switch (s->section_id)
	{
	case TCP_SECTION_INVALID:
	  return (-1);


	  break;
	case TCP_SECTION_Hdr:;
	  for (sp = present_sections.present_Hdr; sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[TCP_FIELD_Hdr_source_port])
	    {;
	      v = ntohs (((struct TCP_Hdr *) (s->content))->source_port);;
	      if (hdt_int32_match
		  (hdt_sets[TCP_FIELD_Hdr_source_port], NULL, v, cp->matches,
		   1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "source_port", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[TCP_FIELD_Hdr_destination_port])
	    {;
	      v =
		ntohs (((struct TCP_Hdr *) (s->content))->destination_port);;
	      if (hdt_int32_match
		  (hdt_sets[TCP_FIELD_Hdr_destination_port], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "destination_port", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[TCP_FIELD_Hdr_sequence_no])
	    {;
	      v = ntohl (((struct TCP_Hdr *) (s->content))->sequence_no);;
	      if (hdt_int32_match
		  (hdt_sets[TCP_FIELD_Hdr_sequence_no], NULL, v, cp->matches,
		   1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "sequence_no", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[TCP_FIELD_Hdr_ack_no])
	    {;
	      v = ntohl (((struct TCP_Hdr *) (s->content))->ack_no);;
	      if (hdt_int32_match
		  (hdt_sets[TCP_FIELD_Hdr_ack_no], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "ack_no", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[TCP_FIELD_Hdr_res1])
	    {;
	      v = ((struct TCP_Hdr *) (s->content))->res1;;
	      if (hdt_int32_match
		  (hdt_sets[TCP_FIELD_Hdr_res1], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr", "res1",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[TCP_FIELD_Hdr_doff])
	    {;
	      v = ((struct TCP_Hdr *) (s->content))->doff;;
	      if (hdt_int32_match
		  (hdt_sets[TCP_FIELD_Hdr_doff], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr", "doff",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[TCP_FIELD_Hdr_flags])
	    {;
	      v = ((struct TCP_Hdr *) (s->content))->flags;;
	      if (hdt_int32_match
		  (hdt_sets[TCP_FIELD_Hdr_flags], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "flags", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[TCP_FIELD_Hdr_window])
	    {;
	      v = ntohs (((struct TCP_Hdr *) (s->content))->window);;
	      if (hdt_int32_match
		  (hdt_sets[TCP_FIELD_Hdr_window], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "window", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[TCP_FIELD_Hdr_check])
	    {;
	      v = ntohs (((struct TCP_Hdr *) (s->content))->check);;
	      if (hdt_int32_match
		  (hdt_sets[TCP_FIELD_Hdr_check], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "check", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[TCP_FIELD_Hdr_urg_ptr])
	    {;
	      v = ntohs (((struct TCP_Hdr *) (s->content))->urg_ptr);;
	      if (hdt_int32_match
		  (hdt_sets[TCP_FIELD_Hdr_urg_ptr], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "urg_ptr", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[TCP_FIELD_Hdr_payload])
	    {
	      struct blob bb;
	      bb.blob = &(((struct TCP_Hdr *) (s->content))->payload);
	      bb.len = cp->length - offsetof (struct TCP_Hdr, payload);;
	      if (hdt_blob_match
		  (hdt_sets[TCP_FIELD_Hdr_payload], NULL, &bb, cp->matches,
		   1))
		{
		  printf ("AIEE!  Can't find blob match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "payload", (end - beg));


	  break;
	default:
	  printf ("Unrecognized section: %i (not %i).\n", s->section_id,
		  TCP_SECTION_Hdr);
	  break;
	}
    }

  return (-1);
}
