/* Copyright 2001, Todd Lewis <tlewis@mindspring.com>
 * All rights reserved. */

#include "hank.h"
#include "hdt_blob.h"
#include "hdt_int32.h"
#include "packet.h"
#include "protocol_engine.h"
#include "protocols.h"
#include "support.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.h"

#include "arp.hi"


static char rcsid[] = "$Id: arp.xml,v 1.2 2001/10/29 06:23:30 tlewis Exp $";


extern struct smodule *arp[];


extern struct protocol_engine arp_pe;


static u_int32_t protocols[] = {

  ((HPN_802 << 24) | 2054),


  0
};

struct simple_prim
{
  prim_id prim;
  struct simple_prim *next;
};

static struct
{
  struct simple_prim *proto_self;


  struct simple_prim *present_Hdr;


}
present_sections;

enum ARP_FIELD
{
  ARP_SELF,


  ARP_FIELD_Hdr,
  ARP_FIELD_Hdr_hardware_address_space,
  ARP_FIELD_Hdr_protocol_address_space,
  ARP_FIELD_Hdr_hardware_address_length,
  ARP_FIELD_Hdr_protocol_address_length,
  ARP_FIELD_Hdr_operation_code,
  ARP_FIELD_Hdr_sender_hardware_address,
  ARP_FIELD_Hdr_sender_protocol_address,
  ARP_FIELD_Hdr_target_hardware_address,
  ARP_FIELD_Hdr_target_protocol_address,


  ARP_FIELD_INVALID
};

enum ARP_SECTIONS
{


  ARP_SECTION_Hdr,


  ARP_SECTION_INVALID
};

struct ARP_null
{
  int empty;


};
struct ARP_Hdr
{
  u_int16_t hardware_address_space;
  u_int16_t protocol_address_space;
  u_int8_t hardware_address_length;
  u_int8_t protocol_address_length;
  u_int16_t operation_code;
  u_int32_t sender_hardware_address:6;
  u_int32_t sender_protocol_address;
  u_int32_t target_hardware_address:6;
  u_int32_t target_protocol_address;


};


static char hexes[] = "0123456789abcdef";

static int gprid = 0;

static hdt_set hdt_sets[ARP_FIELD_INVALID];


static int arp_stock_reg_prim (int field, char *op, char *arg);
static int arp_stock_unreg_prim (prim_id i);
static int arp_stock_decompose (struct packet *p);
static int arp_stock_find_matches (struct capsule *cp, struct prim_can *c);
static int arp_stock_rewrite (struct capsule *c, void *new_payload,
			      u_int32_t len);
static char *arp_stock_report_field (struct capsule *p, int field,
				     struct memcon_context *m);
static int arp_field_lookup (char *section, char *field);

struct protocol_engine arp_pe = {
  {
   "pe",
   "arp",
   PEVMIN,
   PEVMAJ,
   &arp_pe},
  &arp_stock_decompose,
  &arp_stock_reg_prim,
  &arp_stock_unreg_prim,
  &arp_stock_find_matches,
  &arp_stock_rewrite,
  &arp_stock_report_field,
  &arp_field_lookup,
  &protocols[0],
  0
};


struct smodule *arp[] = {

  &(arp_pe.s),
  NULL
};


static int
arp_field_lookup (char *section, char *field)
{
  int p = -1;

  if ((section == NULL) && (field == NULL))
    {
      p = ARP_SELF;
    }
  else if (section != NULL && !(strcmp (section, "INVALID")))
    {
      p = ARP_FIELD_INVALID;


    }
  else if ((section == NULL) || (!(strcmp (section, "Hdr"))))
    {
      if (field == NULL)
	p = ARP_FIELD_Hdr;
      else if (!(strcmp (field, "hardware_address_space")))
	p = ARP_FIELD_Hdr_hardware_address_space;
      else if (!(strcmp (field, "protocol_address_space")))
	p = ARP_FIELD_Hdr_protocol_address_space;
      else if (!(strcmp (field, "hardware_address_length")))
	p = ARP_FIELD_Hdr_hardware_address_length;
      else if (!(strcmp (field, "protocol_address_length")))
	p = ARP_FIELD_Hdr_protocol_address_length;
      else if (!(strcmp (field, "operation_code")))
	p = ARP_FIELD_Hdr_operation_code;
      else if (!(strcmp (field, "sender_hardware_address")))
	p = ARP_FIELD_Hdr_sender_hardware_address;
      else if (!(strcmp (field, "sender_protocol_address")))
	p = ARP_FIELD_Hdr_sender_protocol_address;
      else if (!(strcmp (field, "target_hardware_address")))
	p = ARP_FIELD_Hdr_target_hardware_address;
      else if (!(strcmp (field, "target_protocol_address")))
	p = ARP_FIELD_Hdr_target_protocol_address;


    }

  if (p == -1)
    {
      printf ("ARP: Unrecognized field: \" %s \".  Returning error.\n",
	      field);
      return (ARP_FIELD_INVALID);
    }

  ;
  return (p);
}


static int
arp_stock_decompose (struct packet *p)
{
  struct capsule *e = p->encaps, *f;
  struct ARP_Hdr *h;
  struct section *s;


  while (e->next)
    e = e->next;

  h = (struct ARP_Hdr *) e->payload;


  append_encap (p, protocols[0], e->payload_length, e->payload);
  f = e->next;
  f->pe = &arp_pe;

  if ((s = p->c->eng->alloc (p->c, sizeof (*s))) == NULL)
    exit (-1);
  s->section_id = ARP_SECTION_Hdr;
  s->content = h;
  s->next = NULL;
  f->sections = s;
  f->length = e->payload_length;


  return (0);
}

static int
arp_stock_reg_prim (int field, char *op, char *arg)
{
  int prid = ++gprid;


  ;

  if (field == ARP_FIELD_INVALID)
    {
      printf
	("ARP: Error.  Received invalid field in reg_prim(), op=%s, arg=%s.\n",
	 op, arg);
      return (-1);
    }


  if (!strcmp (op, "present"))
    {
      struct simple_prim *p;

      if (!(p = malloc (sizeof (*p))))
	exit (-1);
      p->prim = prid;

      if (field == ARP_SELF)
	{

	  p->next = present_sections.proto_self;
	  present_sections.proto_self = p;
	  return (prid);


	}
      else if (field == ARP_FIELD_Hdr)
	{
	  p->next = present_sections.present_Hdr;
	  present_sections.present_Hdr = p;
	  return (prid);


	}

      printf ("That shit (%i) ain't keepin' it real.\n", field);
      return (-1);
    }


  if (field == ARP_FIELD_Hdr_hardware_address_space)
    {
      if (hdt_sets[ARP_FIELD_Hdr_hardware_address_space] == 0)
	{
	  hdt_sets[ARP_FIELD_Hdr_hardware_address_space] =
	    hdt_int32_create ();
	  if (hdt_sets[ARP_FIELD_Hdr_hardware_address_space] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "hardware_address_space");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ARP_FIELD_Hdr_hardware_address_space], prid, op, arg))
	{
	  printf ("Error adding primitive %s in arp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ARP_FIELD_Hdr_protocol_address_space)
    {
      if (hdt_sets[ARP_FIELD_Hdr_protocol_address_space] == 0)
	{
	  hdt_sets[ARP_FIELD_Hdr_protocol_address_space] =
	    hdt_int32_create ();
	  if (hdt_sets[ARP_FIELD_Hdr_protocol_address_space] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "protocol_address_space");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ARP_FIELD_Hdr_protocol_address_space], prid, op, arg))
	{
	  printf ("Error adding primitive %s in arp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ARP_FIELD_Hdr_hardware_address_length)
    {
      if (hdt_sets[ARP_FIELD_Hdr_hardware_address_length] == 0)
	{
	  hdt_sets[ARP_FIELD_Hdr_hardware_address_length] =
	    hdt_int32_create ();
	  if (hdt_sets[ARP_FIELD_Hdr_hardware_address_length] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "hardware_address_length");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ARP_FIELD_Hdr_hardware_address_length], prid, op, arg))
	{
	  printf ("Error adding primitive %s in arp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ARP_FIELD_Hdr_protocol_address_length)
    {
      if (hdt_sets[ARP_FIELD_Hdr_protocol_address_length] == 0)
	{
	  hdt_sets[ARP_FIELD_Hdr_protocol_address_length] =
	    hdt_int32_create ();
	  if (hdt_sets[ARP_FIELD_Hdr_protocol_address_length] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "protocol_address_length");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ARP_FIELD_Hdr_protocol_address_length], prid, op, arg))
	{
	  printf ("Error adding primitive %s in arp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ARP_FIELD_Hdr_operation_code)
    {
      if (hdt_sets[ARP_FIELD_Hdr_operation_code] == 0)
	{
	  hdt_sets[ARP_FIELD_Hdr_operation_code] = hdt_int32_create ();
	  if (hdt_sets[ARP_FIELD_Hdr_operation_code] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "operation_code");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ARP_FIELD_Hdr_operation_code], prid, op, arg))
	{
	  printf ("Error adding primitive %s in arp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ARP_FIELD_Hdr_sender_hardware_address)
    {
      if (hdt_sets[ARP_FIELD_Hdr_sender_hardware_address] == 0)
	{
	  hdt_sets[ARP_FIELD_Hdr_sender_hardware_address] =
	    hdt_int32_create ();
	  if (hdt_sets[ARP_FIELD_Hdr_sender_hardware_address] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "sender_hardware_address");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ARP_FIELD_Hdr_sender_hardware_address], prid, op, arg))
	{
	  printf ("Error adding primitive %s in arp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ARP_FIELD_Hdr_sender_protocol_address)
    {
      if (hdt_sets[ARP_FIELD_Hdr_sender_protocol_address] == 0)
	{
	  hdt_sets[ARP_FIELD_Hdr_sender_protocol_address] =
	    hdt_int32_create ();
	  if (hdt_sets[ARP_FIELD_Hdr_sender_protocol_address] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "sender_protocol_address");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ARP_FIELD_Hdr_sender_protocol_address], prid, op, arg))
	{
	  printf ("Error adding primitive %s in arp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ARP_FIELD_Hdr_target_hardware_address)
    {
      if (hdt_sets[ARP_FIELD_Hdr_target_hardware_address] == 0)
	{
	  hdt_sets[ARP_FIELD_Hdr_target_hardware_address] =
	    hdt_int32_create ();
	  if (hdt_sets[ARP_FIELD_Hdr_target_hardware_address] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "target_hardware_address");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ARP_FIELD_Hdr_target_hardware_address], prid, op, arg))
	{
	  printf ("Error adding primitive %s in arp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ARP_FIELD_Hdr_target_protocol_address)
    {
      if (hdt_sets[ARP_FIELD_Hdr_target_protocol_address] == 0)
	{
	  hdt_sets[ARP_FIELD_Hdr_target_protocol_address] =
	    hdt_int32_create ();
	  if (hdt_sets[ARP_FIELD_Hdr_target_protocol_address] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "target_protocol_address");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ARP_FIELD_Hdr_target_protocol_address], prid, op, arg))
	{
	  printf ("Error adding primitive %s in arp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }


  return (-1);
}

static int
arp_stock_unreg_prim (prim_id i)
{

  return (-1);
}

static int
arp_stock_rewrite (struct capsule *c, void *new_payload, u_int32_t len)
{
  if ((c != NULL) && (new_payload != NULL))
    len++;
  return (-1);
}

static char *
arp_stock_report_field (struct capsule *cp, int field,
			struct memcon_context *m)
{
  u_int32_t t;
  char *r, buf[1024];
  struct section *sec, *sec2;

  buf[0] = 0;

  switch (field)
    {


    case (ARP_FIELD_Hdr_hardware_address_space):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ARP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct ARP_Hdr *) (sec->content))->
			   hardware_address_space));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (ARP_FIELD_Hdr_protocol_address_space):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ARP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct ARP_Hdr *) (sec->content))->
			   protocol_address_space));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (ARP_FIELD_Hdr_hardware_address_length):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ARP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct ARP_Hdr *) (sec->content))->
		    hardware_address_length);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (ARP_FIELD_Hdr_protocol_address_length):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ARP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct ARP_Hdr *) (sec->content))->
		    protocol_address_length);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (ARP_FIELD_Hdr_operation_code):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ARP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct ARP_Hdr *) (sec->content))->
			   operation_code));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (ARP_FIELD_Hdr_sender_hardware_address):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ARP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct ARP_Hdr *) (sec->content))->
		    sender_hardware_address);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (ARP_FIELD_Hdr_sender_protocol_address):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ARP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct ARP_Hdr *) (sec->content))->
			   sender_protocol_address));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (ARP_FIELD_Hdr_target_hardware_address):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ARP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct ARP_Hdr *) (sec->content))->
		    target_hardware_address);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (ARP_FIELD_Hdr_target_protocol_address):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ARP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct ARP_Hdr *) (sec->content))->
			   target_protocol_address));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;


    default:

      return (NULL);
    }
  t = (strlen (buf) + 1);
  r = m->eng->alloc (m, t);
  strncpy (r, buf, t);
  return (r);
}

static int
arp_stock_find_matches (struct capsule *cp, struct prim_can *p)
{
  struct section *s;
  struct memcon_context *c = cp->parent->c;
  u_int32_t v;
  unsigned long beg, end;

  if (p)
    return (-1);

  if (cp->matches == NULL && gprid)
    {
      if (!(cp->matches = c->eng->alloc (c, gprid + 1)))
	exit (-1);
      memset (cp->matches, 0, gprid + 1);

      cp->matches[0] = 1;
    }


  for (s = cp->sections; s; s = s->next)
    {
      struct simple_prim *sp;

      for (sp = present_sections.proto_self; sp; sp = sp->next)
	{
	  ;
	  cp->matches[sp->prim] |= 1;
	}


      switch (s->section_id)
	{
	case ARP_SECTION_INVALID:
	  return (-1);


	  break;
	case ARP_SECTION_Hdr:;
	  for (sp = present_sections.present_Hdr; sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[ARP_FIELD_Hdr_hardware_address_space])
	    {;
	      v =
		ntohs (((struct ARP_Hdr *) (s->content))->
		       hardware_address_space);;
	      if (hdt_int32_match
		  (hdt_sets[ARP_FIELD_Hdr_hardware_address_space], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "hardware_address_space", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ARP_FIELD_Hdr_protocol_address_space])
	    {;
	      v =
		ntohs (((struct ARP_Hdr *) (s->content))->
		       protocol_address_space);;
	      if (hdt_int32_match
		  (hdt_sets[ARP_FIELD_Hdr_protocol_address_space], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "protocol_address_space", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ARP_FIELD_Hdr_hardware_address_length])
	    {;
	      v = ((struct ARP_Hdr *) (s->content))->hardware_address_length;;
	      if (hdt_int32_match
		  (hdt_sets[ARP_FIELD_Hdr_hardware_address_length], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "hardware_address_length", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ARP_FIELD_Hdr_protocol_address_length])
	    {;
	      v = ((struct ARP_Hdr *) (s->content))->protocol_address_length;;
	      if (hdt_int32_match
		  (hdt_sets[ARP_FIELD_Hdr_protocol_address_length], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "protocol_address_length", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ARP_FIELD_Hdr_operation_code])
	    {;
	      v = ntohs (((struct ARP_Hdr *) (s->content))->operation_code);;
	      if (hdt_int32_match
		  (hdt_sets[ARP_FIELD_Hdr_operation_code], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "operation_code", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ARP_FIELD_Hdr_sender_hardware_address])
	    {;
	      v = ((struct ARP_Hdr *) (s->content))->sender_hardware_address;;
	      if (hdt_int32_match
		  (hdt_sets[ARP_FIELD_Hdr_sender_hardware_address], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "sender_hardware_address", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ARP_FIELD_Hdr_sender_protocol_address])
	    {;
	      v =
		ntohl (((struct ARP_Hdr *) (s->content))->
		       sender_protocol_address);;
	      if (hdt_int32_match
		  (hdt_sets[ARP_FIELD_Hdr_sender_protocol_address], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "sender_protocol_address", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ARP_FIELD_Hdr_target_hardware_address])
	    {;
	      v = ((struct ARP_Hdr *) (s->content))->target_hardware_address;;
	      if (hdt_int32_match
		  (hdt_sets[ARP_FIELD_Hdr_target_hardware_address], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "target_hardware_address", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ARP_FIELD_Hdr_target_protocol_address])
	    {;
	      v =
		ntohl (((struct ARP_Hdr *) (s->content))->
		       target_protocol_address);;
	      if (hdt_int32_match
		  (hdt_sets[ARP_FIELD_Hdr_target_protocol_address], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "target_protocol_address", (end - beg));


	  break;
	default:
	  printf ("Unrecognized section: %i (not %i).\n", s->section_id,
		  ARP_SECTION_Hdr);
	  break;
	}
    }

  return (-1);
}
