/* Copyright 2001, Todd Lewis <tlewis@mindspring.com>
 * All rights reserved. */

#include "hank.h"
#include "hdt_blob.h"
#include "hdt_int32.h"
#include "packet.h"
#include "protocol_engine.h"
#include "protocols.h"
#include "support.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.h"

#include "dns.hi"


static char rcsid[] = "$Id: dns.xml,v 1.2 2001/10/29 06:52:01 tlewis Exp $";


extern struct smodule *dns[];


extern struct protocol_engine dns_pe;


static u_int32_t protocols[] = {

  ((HPN_UDP << 24) | 53),
  ((HPN_TCP << 24) | 53),


  0
};

struct simple_prim
{
  prim_id prim;
  struct simple_prim *next;
};

static struct
{
  struct simple_prim *proto_self;


  struct simple_prim *present_Hdr;


}
present_sections;

enum DNS_FIELD
{
  DNS_SELF,


  DNS_FIELD_Hdr,
  DNS_FIELD_Hdr_id,
  DNS_FIELD_Hdr_rd,
  DNS_FIELD_Hdr_tc,
  DNS_FIELD_Hdr_aa,
  DNS_FIELD_Hdr_opcode,
  DNS_FIELD_Hdr_qr,
  DNS_FIELD_Hdr_tcode,
  DNS_FIELD_Hdr_z,
  DNS_FIELD_Hdr_ra,
  DNS_FIELD_Hdr_qdcount,
  DNS_FIELD_Hdr_ancount,
  DNS_FIELD_Hdr_nscount,
  DNS_FIELD_Hdr_arcount,


  DNS_FIELD_INVALID
};

enum DNS_SECTIONS
{


  DNS_SECTION_Hdr,


  DNS_SECTION_INVALID
};

struct DNS_null
{
  int empty;


};
struct DNS_Hdr
{
  u_int16_t id;
  u_int32_t rd:1;
  u_int32_t tc:1;
  u_int32_t aa:1;
  u_int32_t opcode:4;
  u_int16_t qr;
  u_int32_t tcode:4;
  u_int32_t z:3;
  u_int32_t ra:1;
  u_int16_t qdcount;
  u_int16_t ancount;
  u_int16_t nscount;
  u_int16_t arcount;


};


static char hexes[] = "0123456789abcdef";

static int gprid = 0;

static hdt_set hdt_sets[DNS_FIELD_INVALID];


static int dns_stock_reg_prim (int field, char *op, char *arg);
static int dns_stock_unreg_prim (prim_id i);
static int dns_stock_decompose (struct packet *p);
static int dns_stock_find_matches (struct capsule *cp, struct prim_can *c);
static int dns_stock_rewrite (struct capsule *c, void *new_payload,
			      u_int32_t len);
static char *dns_stock_report_field (struct capsule *p, int field,
				     struct memcon_context *m);
static int dns_field_lookup (char *section, char *field);

struct protocol_engine dns_pe = {
  {
   "pe",
   "dns",
   PEVMIN,
   PEVMAJ,
   &dns_pe},
  &dns_stock_decompose,
  &dns_stock_reg_prim,
  &dns_stock_unreg_prim,
  &dns_stock_find_matches,
  &dns_stock_rewrite,
  &dns_stock_report_field,
  &dns_field_lookup,
  &protocols[0],
  0
};


struct smodule *dns[] = {

  &(dns_pe.s),
  NULL
};


static int
dns_field_lookup (char *section, char *field)
{
  int p = -1;

  if ((section == NULL) && (field == NULL))
    {
      p = DNS_SELF;
    }
  else if (section != NULL && !(strcmp (section, "INVALID")))
    {
      p = DNS_FIELD_INVALID;


    }
  else if ((section == NULL) || (!(strcmp (section, "Hdr"))))
    {
      if (field == NULL)
	p = DNS_FIELD_Hdr;
      else if (!(strcmp (field, "id")))
	p = DNS_FIELD_Hdr_id;
      else if (!(strcmp (field, "rd")))
	p = DNS_FIELD_Hdr_rd;
      else if (!(strcmp (field, "tc")))
	p = DNS_FIELD_Hdr_tc;
      else if (!(strcmp (field, "aa")))
	p = DNS_FIELD_Hdr_aa;
      else if (!(strcmp (field, "opcode")))
	p = DNS_FIELD_Hdr_opcode;
      else if (!(strcmp (field, "qr")))
	p = DNS_FIELD_Hdr_qr;
      else if (!(strcmp (field, "tcode")))
	p = DNS_FIELD_Hdr_tcode;
      else if (!(strcmp (field, "z")))
	p = DNS_FIELD_Hdr_z;
      else if (!(strcmp (field, "ra")))
	p = DNS_FIELD_Hdr_ra;
      else if (!(strcmp (field, "qdcount")))
	p = DNS_FIELD_Hdr_qdcount;
      else if (!(strcmp (field, "ancount")))
	p = DNS_FIELD_Hdr_ancount;
      else if (!(strcmp (field, "nscount")))
	p = DNS_FIELD_Hdr_nscount;
      else if (!(strcmp (field, "arcount")))
	p = DNS_FIELD_Hdr_arcount;


    }

  if (p == -1)
    {
      printf ("DNS: Unrecognized field: \" %s \".  Returning error.\n",
	      field);
      return (DNS_FIELD_INVALID);
    }

  ;
  return (p);
}


static int
dns_stock_decompose (struct packet *p)
{
  struct capsule *e = p->encaps, *f;
  struct DNS_Hdr *h;
  struct section *s;


  while (e->next)
    e = e->next;

  h = (struct DNS_Hdr *) e->payload;


  append_encap (p, protocols[0], e->payload_length, e->payload);
  f = e->next;
  f->pe = &dns_pe;

  if ((s = p->c->eng->alloc (p->c, sizeof (*s))) == NULL)
    exit (-1);
  s->section_id = DNS_SECTION_Hdr;
  s->content = h;
  s->next = NULL;
  f->sections = s;
  f->length = e->payload_length;


  return (0);
}

static int
dns_stock_reg_prim (int field, char *op, char *arg)
{
  int prid = ++gprid;


  ;

  if (field == DNS_FIELD_INVALID)
    {
      printf
	("DNS: Error.  Received invalid field in reg_prim(), op=%s, arg=%s.\n",
	 op, arg);
      return (-1);
    }


  if (!strcmp (op, "present"))
    {
      struct simple_prim *p;

      if (!(p = malloc (sizeof (*p))))
	exit (-1);
      p->prim = prid;

      if (field == DNS_SELF)
	{

	  p->next = present_sections.proto_self;
	  present_sections.proto_self = p;
	  return (prid);


	}
      else if (field == DNS_FIELD_Hdr)
	{
	  p->next = present_sections.present_Hdr;
	  present_sections.present_Hdr = p;
	  return (prid);


	}

      printf ("That shit (%i) ain't keepin' it real.\n", field);
      return (-1);
    }


  if (field == DNS_FIELD_Hdr_id)
    {
      if (hdt_sets[DNS_FIELD_Hdr_id] == 0)
	{
	  hdt_sets[DNS_FIELD_Hdr_id] = hdt_int32_create ();
	  if (hdt_sets[DNS_FIELD_Hdr_id] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "id");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[DNS_FIELD_Hdr_id], prid, op, arg))
	{
	  printf ("Error adding primitive %s in dns!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == DNS_FIELD_Hdr_rd)
    {
      if (hdt_sets[DNS_FIELD_Hdr_rd] == 0)
	{
	  hdt_sets[DNS_FIELD_Hdr_rd] = hdt_int32_create ();
	  if (hdt_sets[DNS_FIELD_Hdr_rd] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "rd");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[DNS_FIELD_Hdr_rd], prid, op, arg))
	{
	  printf ("Error adding primitive %s in dns!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == DNS_FIELD_Hdr_tc)
    {
      if (hdt_sets[DNS_FIELD_Hdr_tc] == 0)
	{
	  hdt_sets[DNS_FIELD_Hdr_tc] = hdt_int32_create ();
	  if (hdt_sets[DNS_FIELD_Hdr_tc] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "tc");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[DNS_FIELD_Hdr_tc], prid, op, arg))
	{
	  printf ("Error adding primitive %s in dns!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == DNS_FIELD_Hdr_aa)
    {
      if (hdt_sets[DNS_FIELD_Hdr_aa] == 0)
	{
	  hdt_sets[DNS_FIELD_Hdr_aa] = hdt_int32_create ();
	  if (hdt_sets[DNS_FIELD_Hdr_aa] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "aa");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[DNS_FIELD_Hdr_aa], prid, op, arg))
	{
	  printf ("Error adding primitive %s in dns!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == DNS_FIELD_Hdr_opcode)
    {
      if (hdt_sets[DNS_FIELD_Hdr_opcode] == 0)
	{
	  hdt_sets[DNS_FIELD_Hdr_opcode] = hdt_int32_create ();
	  if (hdt_sets[DNS_FIELD_Hdr_opcode] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "opcode");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[DNS_FIELD_Hdr_opcode], prid, op, arg))
	{
	  printf ("Error adding primitive %s in dns!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == DNS_FIELD_Hdr_qr)
    {
      if (hdt_sets[DNS_FIELD_Hdr_qr] == 0)
	{
	  hdt_sets[DNS_FIELD_Hdr_qr] = hdt_int32_create ();
	  if (hdt_sets[DNS_FIELD_Hdr_qr] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "qr");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[DNS_FIELD_Hdr_qr], prid, op, arg))
	{
	  printf ("Error adding primitive %s in dns!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == DNS_FIELD_Hdr_tcode)
    {
      if (hdt_sets[DNS_FIELD_Hdr_tcode] == 0)
	{
	  hdt_sets[DNS_FIELD_Hdr_tcode] = hdt_int32_create ();
	  if (hdt_sets[DNS_FIELD_Hdr_tcode] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "tcode");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[DNS_FIELD_Hdr_tcode], prid, op, arg))
	{
	  printf ("Error adding primitive %s in dns!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == DNS_FIELD_Hdr_z)
    {
      if (hdt_sets[DNS_FIELD_Hdr_z] == 0)
	{
	  hdt_sets[DNS_FIELD_Hdr_z] = hdt_int32_create ();
	  if (hdt_sets[DNS_FIELD_Hdr_z] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "z");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[DNS_FIELD_Hdr_z], prid, op, arg))
	{
	  printf ("Error adding primitive %s in dns!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == DNS_FIELD_Hdr_ra)
    {
      if (hdt_sets[DNS_FIELD_Hdr_ra] == 0)
	{
	  hdt_sets[DNS_FIELD_Hdr_ra] = hdt_int32_create ();
	  if (hdt_sets[DNS_FIELD_Hdr_ra] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "ra");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[DNS_FIELD_Hdr_ra], prid, op, arg))
	{
	  printf ("Error adding primitive %s in dns!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == DNS_FIELD_Hdr_qdcount)
    {
      if (hdt_sets[DNS_FIELD_Hdr_qdcount] == 0)
	{
	  hdt_sets[DNS_FIELD_Hdr_qdcount] = hdt_int32_create ();
	  if (hdt_sets[DNS_FIELD_Hdr_qdcount] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "qdcount");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[DNS_FIELD_Hdr_qdcount], prid, op, arg))
	{
	  printf ("Error adding primitive %s in dns!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == DNS_FIELD_Hdr_ancount)
    {
      if (hdt_sets[DNS_FIELD_Hdr_ancount] == 0)
	{
	  hdt_sets[DNS_FIELD_Hdr_ancount] = hdt_int32_create ();
	  if (hdt_sets[DNS_FIELD_Hdr_ancount] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "ancount");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[DNS_FIELD_Hdr_ancount], prid, op, arg))
	{
	  printf ("Error adding primitive %s in dns!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == DNS_FIELD_Hdr_nscount)
    {
      if (hdt_sets[DNS_FIELD_Hdr_nscount] == 0)
	{
	  hdt_sets[DNS_FIELD_Hdr_nscount] = hdt_int32_create ();
	  if (hdt_sets[DNS_FIELD_Hdr_nscount] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "nscount");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[DNS_FIELD_Hdr_nscount], prid, op, arg))
	{
	  printf ("Error adding primitive %s in dns!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == DNS_FIELD_Hdr_arcount)
    {
      if (hdt_sets[DNS_FIELD_Hdr_arcount] == 0)
	{
	  hdt_sets[DNS_FIELD_Hdr_arcount] = hdt_int32_create ();
	  if (hdt_sets[DNS_FIELD_Hdr_arcount] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "arcount");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[DNS_FIELD_Hdr_arcount], prid, op, arg))
	{
	  printf ("Error adding primitive %s in dns!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }


  return (-1);
}

static int
dns_stock_unreg_prim (prim_id i)
{

  return (-1);
}

static int
dns_stock_rewrite (struct capsule *c, void *new_payload, u_int32_t len)
{
  if ((c != NULL) && (new_payload != NULL))
    len++;
  return (-1);
}

static char *
dns_stock_report_field (struct capsule *cp, int field,
			struct memcon_context *m)
{
  u_int32_t t;
  char *r, buf[1024];
  struct section *sec, *sec2;

  buf[0] = 0;

  switch (field)
    {


    case (DNS_FIELD_Hdr_id):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == DNS_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct DNS_Hdr *) (sec->content))->id));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (DNS_FIELD_Hdr_rd):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == DNS_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct DNS_Hdr *) (sec->content))->rd);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (DNS_FIELD_Hdr_tc):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == DNS_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct DNS_Hdr *) (sec->content))->tc);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (DNS_FIELD_Hdr_aa):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == DNS_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct DNS_Hdr *) (sec->content))->aa);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (DNS_FIELD_Hdr_opcode):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == DNS_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct DNS_Hdr *) (sec->content))->opcode);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (DNS_FIELD_Hdr_qr):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == DNS_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct DNS_Hdr *) (sec->content))->qr));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (DNS_FIELD_Hdr_tcode):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == DNS_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct DNS_Hdr *) (sec->content))->tcode);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (DNS_FIELD_Hdr_z):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == DNS_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct DNS_Hdr *) (sec->content))->z);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (DNS_FIELD_Hdr_ra):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == DNS_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct DNS_Hdr *) (sec->content))->ra);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (DNS_FIELD_Hdr_qdcount):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == DNS_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct DNS_Hdr *) (sec->content))->qdcount));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (DNS_FIELD_Hdr_ancount):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == DNS_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct DNS_Hdr *) (sec->content))->ancount));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (DNS_FIELD_Hdr_nscount):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == DNS_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct DNS_Hdr *) (sec->content))->nscount));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (DNS_FIELD_Hdr_arcount):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == DNS_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct DNS_Hdr *) (sec->content))->arcount));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;


    default:

      return (NULL);
    }
  t = (strlen (buf) + 1);
  r = m->eng->alloc (m, t);
  strncpy (r, buf, t);
  return (r);
}

static int
dns_stock_find_matches (struct capsule *cp, struct prim_can *p)
{
  struct section *s;
  struct memcon_context *c = cp->parent->c;
  u_int32_t v;
  unsigned long beg, end;

  if (p)
    return (-1);

  if (cp->matches == NULL && gprid)
    {
      if (!(cp->matches = c->eng->alloc (c, gprid + 1)))
	exit (-1);
      memset (cp->matches, 0, gprid + 1);

      cp->matches[0] = 1;
    }


  for (s = cp->sections; s; s = s->next)
    {
      struct simple_prim *sp;

      for (sp = present_sections.proto_self; sp; sp = sp->next)
	{
	  ;
	  cp->matches[sp->prim] |= 1;
	}


      switch (s->section_id)
	{
	case DNS_SECTION_INVALID:
	  return (-1);


	  break;
	case DNS_SECTION_Hdr:;
	  for (sp = present_sections.present_Hdr; sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[DNS_FIELD_Hdr_id])
	    {;
	      v = ntohs (((struct DNS_Hdr *) (s->content))->id);;
	      if (hdt_int32_match
		  (hdt_sets[DNS_FIELD_Hdr_id], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr", "id",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[DNS_FIELD_Hdr_rd])
	    {;
	      v = ((struct DNS_Hdr *) (s->content))->rd;;
	      if (hdt_int32_match
		  (hdt_sets[DNS_FIELD_Hdr_rd], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr", "rd",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[DNS_FIELD_Hdr_tc])
	    {;
	      v = ((struct DNS_Hdr *) (s->content))->tc;;
	      if (hdt_int32_match
		  (hdt_sets[DNS_FIELD_Hdr_tc], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr", "tc",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[DNS_FIELD_Hdr_aa])
	    {;
	      v = ((struct DNS_Hdr *) (s->content))->aa;;
	      if (hdt_int32_match
		  (hdt_sets[DNS_FIELD_Hdr_aa], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr", "aa",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[DNS_FIELD_Hdr_opcode])
	    {;
	      v = ((struct DNS_Hdr *) (s->content))->opcode;;
	      if (hdt_int32_match
		  (hdt_sets[DNS_FIELD_Hdr_opcode], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "opcode", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[DNS_FIELD_Hdr_qr])
	    {;
	      v = ntohs (((struct DNS_Hdr *) (s->content))->qr);;
	      if (hdt_int32_match
		  (hdt_sets[DNS_FIELD_Hdr_qr], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr", "qr",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[DNS_FIELD_Hdr_tcode])
	    {;
	      v = ((struct DNS_Hdr *) (s->content))->tcode;;
	      if (hdt_int32_match
		  (hdt_sets[DNS_FIELD_Hdr_tcode], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "tcode", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[DNS_FIELD_Hdr_z])
	    {;
	      v = ((struct DNS_Hdr *) (s->content))->z;;
	      if (hdt_int32_match
		  (hdt_sets[DNS_FIELD_Hdr_z], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr", "z",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[DNS_FIELD_Hdr_ra])
	    {;
	      v = ((struct DNS_Hdr *) (s->content))->ra;;
	      if (hdt_int32_match
		  (hdt_sets[DNS_FIELD_Hdr_ra], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr", "ra",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[DNS_FIELD_Hdr_qdcount])
	    {;
	      v = ntohs (((struct DNS_Hdr *) (s->content))->qdcount);;
	      if (hdt_int32_match
		  (hdt_sets[DNS_FIELD_Hdr_qdcount], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "qdcount", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[DNS_FIELD_Hdr_ancount])
	    {;
	      v = ntohs (((struct DNS_Hdr *) (s->content))->ancount);;
	      if (hdt_int32_match
		  (hdt_sets[DNS_FIELD_Hdr_ancount], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "ancount", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[DNS_FIELD_Hdr_nscount])
	    {;
	      v = ntohs (((struct DNS_Hdr *) (s->content))->nscount);;
	      if (hdt_int32_match
		  (hdt_sets[DNS_FIELD_Hdr_nscount], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "nscount", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[DNS_FIELD_Hdr_arcount])
	    {;
	      v = ntohs (((struct DNS_Hdr *) (s->content))->arcount);;
	      if (hdt_int32_match
		  (hdt_sets[DNS_FIELD_Hdr_arcount], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "arcount", (end - beg));


	  break;
	default:
	  printf ("Unrecognized section: %i (not %i).\n", s->section_id,
		  DNS_SECTION_Hdr);
	  break;
	}
    }

  return (-1);
}
