/* Copyright 2001, Todd Lewis <tlewis@mindspring.com>
 * All rights reserved. */

#include "hank.h"
#include "hdt_blob.h"
#include "hdt_int32.h"
#include "packet.h"
#include "protocol_engine.h"
#include "protocols.h"
#include "support.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.h"

#include "ethernet.hi"


static char rcsid[] =
  "$Id: ethernet.xml,v 1.3 2001/10/29 07:35:20 tlewis Exp $";


extern struct smodule *ethernet[];


extern struct protocol_engine ethernet_pe;


static u_int32_t protocols[] = {

  ((HPN_PCAP << 24) | 1),
  ((HPN_PCAP << 24) | 97),


  0
};

struct simple_prim
{
  prim_id prim;
  struct simple_prim *next;
};

static struct
{
  struct simple_prim *proto_self;


  struct simple_prim *present_Hdr;


}
present_sections;

enum ETHERNET_FIELD
{
  ETHERNET_SELF,


  ETHERNET_FIELD_Hdr,
  ETHERNET_FIELD_Hdr_destination,
  ETHERNET_FIELD_Hdr_source,
  ETHERNET_FIELD_Hdr_type,
  ETHERNET_FIELD_Hdr_payload,


  ETHERNET_FIELD_INVALID
};

enum ETHERNET_SECTIONS
{


  ETHERNET_SECTION_Hdr,


  ETHERNET_SECTION_INVALID
};

struct ETHERNET_null
{
  int empty;


};
struct ETHERNET_Hdr
{
  u_int8_t destination[6];
  u_int8_t source[6];
  u_int16_t type;
  unsigned char payload;


};


static char hexes[] = "0123456789abcdef";

static int gprid = 0;

static hdt_set hdt_sets[ETHERNET_FIELD_INVALID];


static int ethernet_stock_reg_prim (int field, char *op, char *arg);
static int ethernet_stock_unreg_prim (prim_id i);
static int ethernet_stock_decompose (struct packet *p);
static int ethernet_stock_find_matches (struct capsule *cp,
					struct prim_can *c);
static int ethernet_stock_rewrite (struct capsule *c, void *new_payload,
				   u_int32_t len);
static char *ethernet_stock_report_field (struct capsule *p, int field,
					  struct memcon_context *m);
static int ethernet_field_lookup (char *section, char *field);

struct protocol_engine ethernet_pe = {
  {
   "pe",
   "ethernet",
   PEVMIN,
   PEVMAJ,
   &ethernet_pe},
  &ethernet_stock_decompose,
  &ethernet_stock_reg_prim,
  &ethernet_stock_unreg_prim,
  &ethernet_stock_find_matches,
  &ethernet_stock_rewrite,
  &ethernet_stock_report_field,
  &ethernet_field_lookup,
  &protocols[0],
  0
};


struct smodule *ethernet[] = {

  &(ethernet_pe.s),
  NULL
};


static int
ethernet_field_lookup (char *section, char *field)
{
  int p = -1;

  if ((section == NULL) && (field == NULL))
    {
      p = ETHERNET_SELF;
    }
  else if (section != NULL && !(strcmp (section, "INVALID")))
    {
      p = ETHERNET_FIELD_INVALID;


    }
  else if ((section == NULL) || (!(strcmp (section, "Hdr"))))
    {
      if (field == NULL)
	p = ETHERNET_FIELD_Hdr;
      else if (!(strcmp (field, "destination")))
	p = ETHERNET_FIELD_Hdr_destination;
      else if (!(strcmp (field, "source")))
	p = ETHERNET_FIELD_Hdr_source;
      else if (!(strcmp (field, "type")))
	p = ETHERNET_FIELD_Hdr_type;
      else if (!(strcmp (field, "payload")))
	p = ETHERNET_FIELD_Hdr_payload;


    }

  if (p == -1)
    {
      printf ("ETHERNET: Unrecognized field: \" %s \".  Returning error.\n",
	      field);
      return (ETHERNET_FIELD_INVALID);
    }

  ;
  return (p);
}


static int
ethernet_stock_decompose (struct packet *p)
{
  struct capsule *e = p->encaps, *f;
  struct ETHERNET_Hdr *h;
  struct section *s;


  while (e->next)
    e = e->next;

  h = (struct ETHERNET_Hdr *) e->payload;


  append_encap (p, protocols[0], e->payload_length, e->payload);
  f = e->next;
  f->pe = &ethernet_pe;

  if ((s = p->c->eng->alloc (p->c, sizeof (*s))) == NULL)
    exit (-1);
  s->section_id = ETHERNET_SECTION_Hdr;
  s->content = h;
  s->next = NULL;
  f->sections = s;
  f->length = e->payload_length;


  f->payload_length = (f->length - (offsetof (struct ETHERNET_Hdr, payload)));
  f->payload_protocol = (u_int32_t) ((HPN_802 << 24) | ntohs (h->type));
  f->payload = &(e->payload[(offsetof (struct ETHERNET_Hdr, payload))]);


  return (0);
}

static int
ethernet_stock_reg_prim (int field, char *op, char *arg)
{
  int prid = ++gprid;


  ;

  if (field == ETHERNET_FIELD_INVALID)
    {
      printf
	("ETHERNET: Error.  Received invalid field in reg_prim(), op=%s, arg=%s.\n",
	 op, arg);
      return (-1);
    }


  if (!strcmp (op, "present"))
    {
      struct simple_prim *p;

      if (!(p = malloc (sizeof (*p))))
	exit (-1);
      p->prim = prid;

      if (field == ETHERNET_SELF)
	{

	  p->next = present_sections.proto_self;
	  present_sections.proto_self = p;
	  return (prid);


	}
      else if (field == ETHERNET_FIELD_Hdr)
	{
	  p->next = present_sections.present_Hdr;
	  present_sections.present_Hdr = p;
	  return (prid);


	}

      printf ("That shit (%i) ain't keepin' it real.\n", field);
      return (-1);
    }


  if (field == ETHERNET_FIELD_Hdr_type)
    {
      if (hdt_sets[ETHERNET_FIELD_Hdr_type] == 0)
	{
	  hdt_sets[ETHERNET_FIELD_Hdr_type] = hdt_int32_create ();
	  if (hdt_sets[ETHERNET_FIELD_Hdr_type] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "type");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[ETHERNET_FIELD_Hdr_type], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ethernet!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ETHERNET_FIELD_Hdr_payload)
    {
      if (hdt_sets[ETHERNET_FIELD_Hdr_payload] == 0)
	{
	  hdt_sets[ETHERNET_FIELD_Hdr_payload] = hdt_blob_create ();;
	  if (hdt_sets[ETHERNET_FIELD_Hdr_payload] == 0)
	    {
	      printf ("AIEE!  Can't add match set for %s.%s!  Exiting.\n",
		      "Hdr", "payload");
	      exit (-1);
	    }
	}
      if (hdt_blob_add (hdt_sets[ETHERNET_FIELD_Hdr_payload], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ethernet!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }


  return (-1);
}

static int
ethernet_stock_unreg_prim (prim_id i)
{

  return (-1);
}

static int
ethernet_stock_rewrite (struct capsule *c, void *new_payload, u_int32_t len)
{
  if ((c != NULL) && (new_payload != NULL))
    len++;
  return (-1);
}

static char *
ethernet_stock_report_field (struct capsule *cp, int field,
			     struct memcon_context *m)
{
  u_int32_t t;
  char *r, buf[1024];
  struct section *sec, *sec2;

  buf[0] = 0;

  switch (field)
    {


    case (ETHERNET_FIELD_Hdr_destination):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ETHERNET_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  int i;
	  unsigned char c;
	  buf[0] = '0';
	  buf[1] = 'x';
	  for (i = 0; i < 6; i++)
	    {
	      c = ((struct ETHERNET_Hdr *) (sec->content))->destination[i];
	      buf[(i * 2) + 2] = hexes[c >> 4];
	      buf[(i * 2) + 3] = hexes[c & 0xf];
	      buf[(i * 2) + 4] = 0;
	    }
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (ETHERNET_FIELD_Hdr_source):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ETHERNET_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  int i;
	  unsigned char c;
	  buf[0] = '0';
	  buf[1] = 'x';
	  for (i = 0; i < 6; i++)
	    {
	      c = ((struct ETHERNET_Hdr *) (sec->content))->source[i];
	      buf[(i * 2) + 2] = hexes[c >> 4];
	      buf[(i * 2) + 3] = hexes[c & 0xf];
	      buf[(i * 2) + 4] = 0;
	    }
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (ETHERNET_FIELD_Hdr_type):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ETHERNET_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct ETHERNET_Hdr *) (sec->content))->type));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;


    default:

      return (NULL);
    }
  t = (strlen (buf) + 1);
  r = m->eng->alloc (m, t);
  strncpy (r, buf, t);
  return (r);
}

static int
ethernet_stock_find_matches (struct capsule *cp, struct prim_can *p)
{
  struct section *s;
  struct memcon_context *c = cp->parent->c;
  u_int32_t v;
  unsigned long beg, end;

  if (p)
    return (-1);

  if (cp->matches == NULL && gprid)
    {
      if (!(cp->matches = c->eng->alloc (c, gprid + 1)))
	exit (-1);
      memset (cp->matches, 0, gprid + 1);

      cp->matches[0] = 1;
    }


  for (s = cp->sections; s; s = s->next)
    {
      struct simple_prim *sp;

      for (sp = present_sections.proto_self; sp; sp = sp->next)
	{
	  ;
	  cp->matches[sp->prim] |= 1;
	}


      switch (s->section_id)
	{
	case ETHERNET_SECTION_INVALID:
	  return (-1);


	  break;
	case ETHERNET_SECTION_Hdr:;
	  for (sp = present_sections.present_Hdr; sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }


	  rdtscl (beg);
	  if (hdt_sets[ETHERNET_FIELD_Hdr_type])
	    {;
	      v = ntohs (((struct ETHERNET_Hdr *) (s->content))->type);;
	      if (hdt_int32_match
		  (hdt_sets[ETHERNET_FIELD_Hdr_type], NULL, v, cp->matches,
		   1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr", "type",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ETHERNET_FIELD_Hdr_payload])
	    {
	      struct blob bb;
	      bb.blob = &(((struct ETHERNET_Hdr *) (s->content))->payload);
	      bb.len = cp->length - offsetof (struct ETHERNET_Hdr, payload);;
	      if (hdt_blob_match
		  (hdt_sets[ETHERNET_FIELD_Hdr_payload], NULL, &bb,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find blob match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "payload", (end - beg));


	  break;
	default:
	  printf ("Unrecognized section: %i (not %i).\n", s->section_id,
		  ETHERNET_SECTION_Hdr);
	  break;
	}
    }

  return (-1);
}
