/* Copyright 2001, Todd Lewis <tlewis@mindspring.com>
 * All rights reserved. */

#include "hank.h"
#include "hdt_blob.h"
#include "hdt_int32.h"
#include "packet.h"
#include "protocol_engine.h"
#include "protocols.h"
#include "support.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.h"

#include "icmp.hi"


static char rcsid[] = "$Id: icmp.xml,v 1.6 2001/12/04 05:47:11 tlewis Exp $";


extern struct smodule *icmp[];


extern struct protocol_engine icmp_pe;


static u_int32_t protocols[] = {

  ((HPN_IPV4 << 24) | 1),


  0
};

struct simple_prim
{
  prim_id prim;
  struct simple_prim *next;
};

static struct
{
  struct simple_prim *proto_self;


  struct simple_prim *present_Hdr;


  struct simple_prim *present_Unreachable;

  struct simple_prim *present_TTLexceeded;

  struct simple_prim *present_Parameter;


  struct simple_prim *present_SourceQuench;

  struct simple_prim *present_Redirect;

  struct simple_prim *present_Echo;


  struct simple_prim *present_Timestamp;


  struct simple_prim *present_Information;


}
present_sections;

enum ICMP_FIELD
{
  ICMP_SELF,


  ICMP_FIELD_Hdr,
  ICMP_FIELD_Hdr_type,
  ICMP_FIELD_Hdr_code,
  ICMP_FIELD_Hdr_checksum,
  ICMP_FIELD_Hdr_payload,
  ICMP_FIELD_Unreachable,
  ICMP_FIELD_Unreachable_unused,
  ICMP_FIELD_TTLexceeded,
  ICMP_FIELD_TTLexceeded_unused,
  ICMP_FIELD_Parameter,
  ICMP_FIELD_Parameter_pointer,
  ICMP_FIELD_Parameter_unused,
  ICMP_FIELD_SourceQuench,
  ICMP_FIELD_SourceQuench_unused,
  ICMP_FIELD_Redirect,
  ICMP_FIELD_Redirect_gateway,
  ICMP_FIELD_Echo,
  ICMP_FIELD_Echo_id,
  ICMP_FIELD_Echo_sequence_number,
  ICMP_FIELD_Echo_payload,
  ICMP_FIELD_Timestamp,
  ICMP_FIELD_Timestamp_id,
  ICMP_FIELD_Timestamp_sequence_number,
  ICMP_FIELD_Timestamp_originate_ts,
  ICMP_FIELD_Timestamp_receive_ts,
  ICMP_FIELD_Timestamp_transmit_ts,
  ICMP_FIELD_Information,
  ICMP_FIELD_Information_id,
  ICMP_FIELD_Information_sequence_number,


  ICMP_FIELD_INVALID
};

enum ICMP_SECTIONS
{


  ICMP_SECTION_Hdr,


  ICMP_SECTION_Unreachable,

  ICMP_SECTION_TTLexceeded,

  ICMP_SECTION_Parameter,


  ICMP_SECTION_SourceQuench,

  ICMP_SECTION_Redirect,

  ICMP_SECTION_Echo,


  ICMP_SECTION_Timestamp,


  ICMP_SECTION_Information,


  ICMP_SECTION_INVALID
};

struct ICMP_null
{
  int empty;


};
struct ICMP_Hdr
{
  u_int8_t type;
  u_int8_t code;
  u_int16_t checksum;
  unsigned char payload;
};
struct ICMP_Unreachable
{
  u_int32_t unused;
};
struct ICMP_TTLexceeded
{
  u_int32_t unused;
};
struct ICMP_Parameter
{
  u_int8_t pointer;
  u_int32_t unused;
};
struct ICMP_SourceQuench
{
  u_int32_t unused;
};
struct ICMP_Redirect
{
  ipv4_addr gateway;
};
struct ICMP_Echo
{
  u_int16_t id;
  u_int16_t sequence_number;
  unsigned char payload;
};
struct ICMP_Timestamp
{
  u_int16_t id;
  u_int16_t sequence_number;
  u_int32_t originate_ts;
  u_int32_t receive_ts;
  u_int32_t transmit_ts;
};
struct ICMP_Information
{
  u_int16_t id;
  u_int16_t sequence_number;


};


static char hexes[] = "0123456789abcdef";

static int gprid = 0;

static hdt_set hdt_sets[ICMP_FIELD_INVALID];


static int icmp_stock_reg_prim (int field, char *op, char *arg);
static int icmp_stock_unreg_prim (prim_id i);
static int icmp_stock_decompose (struct packet *p);
static int icmp_stock_find_matches (struct capsule *cp, struct prim_can *c);
static int icmp_stock_rewrite (struct capsule *c, void *new_payload,
			       u_int32_t len);
static char *icmp_stock_report_field (struct capsule *p, int field,
				      struct memcon_context *m);
static int icmp_field_lookup (char *section, char *field);

struct protocol_engine icmp_pe = {
  {
   "pe",
   "icmp",
   PEVMIN,
   PEVMAJ,
   &icmp_pe},
  &icmp_private_decompose,
  &icmp_stock_reg_prim,
  &icmp_stock_unreg_prim,
  &icmp_stock_find_matches,
  &icmp_stock_rewrite,
  &icmp_stock_report_field,
  &icmp_field_lookup,
  &protocols[0],
  0
};


struct smodule *icmp[] = {

  &(icmp_pe.s),
  NULL
};

static int
icmp_private_decompose (struct packet *p)
{
  struct capsule *e = p->encaps, *f;
  struct ICMP_Hdr *h;
  struct section *s, *t;


  while (e->next)
    e = e->next;

  h = (struct ICMP_Hdr *) e->payload;


  append_encap (p, protocols[0], e->payload_length, e->payload);
  f = e->next;
  f->pe = &icmp_pe;

  if ((s = p->c->eng->alloc (p->c, sizeof (*s))) == NULL)
    exit (-1);

  if ((t = p->c->eng->alloc (p->c, sizeof (*t))) == NULL)
    exit (-1);

  s->section_id = ICMP_SECTION_Hdr;
  s->content = h;
  s->next = t;


  switch (h->type)
    {

    case 0:
    case 8:
      t->section_id = ICMP_SECTION_Echo;
      break;

    case 3:
      t->section_id = ICMP_SECTION_Unreachable;
      break;

    case 4:
      t->section_id = ICMP_SECTION_SourceQuench;
      break;

    case 5:
      t->section_id = ICMP_SECTION_Redirect;
      break;

    case 11:
      t->section_id = ICMP_SECTION_TTLexceeded;
      break;

    case 12:
      t->section_id = ICMP_SECTION_Parameter;
      break;

    case 13:
    case 14:
      t->section_id = ICMP_SECTION_Timestamp;
      break;

    case 15:
    case 16:
      t->section_id = ICMP_SECTION_Information;
      break;

    default:
      printf ("AIEE!  Unrecognized icmp type: %i\n", h->type);

    }
  t->content = &(h->payload);
  t->next = NULL;

  f->sections = s;

  return (0);
}


static int
icmp_field_lookup (char *section, char *field)
{
  int p = -1;

  if ((section == NULL) && (field == NULL))
    {
      p = ICMP_SELF;
    }
  else if (section != NULL && !(strcmp (section, "INVALID")))
    {
      p = ICMP_FIELD_INVALID;


    }
  else if ((section == NULL) || (!(strcmp (section, "Hdr"))))
    {
      if (field == NULL)
	p = ICMP_FIELD_Hdr;
      else if (!(strcmp (field, "type")))
	p = ICMP_FIELD_Hdr_type;
      else if (!(strcmp (field, "code")))
	p = ICMP_FIELD_Hdr_code;
      else if (!(strcmp (field, "checksum")))
	p = ICMP_FIELD_Hdr_checksum;
      else if (!(strcmp (field, "payload")))
	p = ICMP_FIELD_Hdr_payload;
    }
  else if ((section == NULL) || (!(strcmp (section, "Unreachable"))))
    {
      if (field == NULL)
	p = ICMP_FIELD_Unreachable;
      else if (!(strcmp (field, "unused")))
	p = ICMP_FIELD_Unreachable_unused;
    }
  else if ((section == NULL) || (!(strcmp (section, "TTLexceeded"))))
    {
      if (field == NULL)
	p = ICMP_FIELD_TTLexceeded;
      else if (!(strcmp (field, "unused")))
	p = ICMP_FIELD_TTLexceeded_unused;
    }
  else if ((section == NULL) || (!(strcmp (section, "Parameter"))))
    {
      if (field == NULL)
	p = ICMP_FIELD_Parameter;
      else if (!(strcmp (field, "pointer")))
	p = ICMP_FIELD_Parameter_pointer;
      else if (!(strcmp (field, "unused")))
	p = ICMP_FIELD_Parameter_unused;
    }
  else if ((section == NULL) || (!(strcmp (section, "SourceQuench"))))
    {
      if (field == NULL)
	p = ICMP_FIELD_SourceQuench;
      else if (!(strcmp (field, "unused")))
	p = ICMP_FIELD_SourceQuench_unused;
    }
  else if ((section == NULL) || (!(strcmp (section, "Redirect"))))
    {
      if (field == NULL)
	p = ICMP_FIELD_Redirect;
      else if (!(strcmp (field, "gateway")))
	p = ICMP_FIELD_Redirect_gateway;
    }
  else if ((section == NULL) || (!(strcmp (section, "Echo"))))
    {
      if (field == NULL)
	p = ICMP_FIELD_Echo;
      else if (!(strcmp (field, "id")))
	p = ICMP_FIELD_Echo_id;
      else if (!(strcmp (field, "sequence_number")))
	p = ICMP_FIELD_Echo_sequence_number;
      else if (!(strcmp (field, "payload")))
	p = ICMP_FIELD_Echo_payload;
    }
  else if ((section == NULL) || (!(strcmp (section, "Timestamp"))))
    {
      if (field == NULL)
	p = ICMP_FIELD_Timestamp;
      else if (!(strcmp (field, "id")))
	p = ICMP_FIELD_Timestamp_id;
      else if (!(strcmp (field, "sequence_number")))
	p = ICMP_FIELD_Timestamp_sequence_number;
      else if (!(strcmp (field, "originate_ts")))
	p = ICMP_FIELD_Timestamp_originate_ts;
      else if (!(strcmp (field, "receive_ts")))
	p = ICMP_FIELD_Timestamp_receive_ts;
      else if (!(strcmp (field, "transmit_ts")))
	p = ICMP_FIELD_Timestamp_transmit_ts;
    }
  else if ((section == NULL) || (!(strcmp (section, "Information"))))
    {
      if (field == NULL)
	p = ICMP_FIELD_Information;
      else if (!(strcmp (field, "id")))
	p = ICMP_FIELD_Information_id;
      else if (!(strcmp (field, "sequence_number")))
	p = ICMP_FIELD_Information_sequence_number;


    }

  if (p == -1)
    {
      printf ("ICMP: Unrecognized field: \" %s \".  Returning error.\n",
	      field);
      return (ICMP_FIELD_INVALID);
    }

  ;
  return (p);
}


static int
icmp_stock_decompose (struct packet *p)
{
  struct capsule *e = p->encaps, *f;
  struct ICMP_Hdr *h;
  struct section *s;


  while (e->next)
    e = e->next;

  h = (struct ICMP_Hdr *) e->payload;


  append_encap (p, protocols[0], e->payload_length, e->payload);
  f = e->next;
  f->pe = &icmp_pe;

  if ((s = p->c->eng->alloc (p->c, sizeof (*s))) == NULL)
    exit (-1);
  s->section_id = ICMP_SECTION_Hdr;
  s->content = h;
  s->next = NULL;
  f->sections = s;
  f->length = e->payload_length;


  return (0);
}

static int
icmp_stock_reg_prim (int field, char *op, char *arg)
{
  int prid = ++gprid;


  ;

  if (field == ICMP_FIELD_INVALID)
    {
      printf
	("ICMP: Error.  Received invalid field in reg_prim(), op=%s, arg=%s.\n",
	 op, arg);
      return (-1);
    }


  if (!strcmp (op, "present"))
    {
      struct simple_prim *p;

      if (!(p = malloc (sizeof (*p))))
	exit (-1);
      p->prim = prid;

      if (field == ICMP_SELF)
	{

	  p->next = present_sections.proto_self;
	  present_sections.proto_self = p;
	  return (prid);


	}
      else if (field == ICMP_FIELD_Hdr)
	{
	  p->next = present_sections.present_Hdr;
	  present_sections.present_Hdr = p;
	  return (prid);


	}
      else if (field == ICMP_FIELD_Unreachable)
	{
	  p->next = present_sections.present_Unreachable;
	  present_sections.present_Unreachable = p;
	  return (prid);

	}
      else if (field == ICMP_FIELD_TTLexceeded)
	{
	  p->next = present_sections.present_TTLexceeded;
	  present_sections.present_TTLexceeded = p;
	  return (prid);

	}
      else if (field == ICMP_FIELD_Parameter)
	{
	  p->next = present_sections.present_Parameter;
	  present_sections.present_Parameter = p;
	  return (prid);


	}
      else if (field == ICMP_FIELD_SourceQuench)
	{
	  p->next = present_sections.present_SourceQuench;
	  present_sections.present_SourceQuench = p;
	  return (prid);

	}
      else if (field == ICMP_FIELD_Redirect)
	{
	  p->next = present_sections.present_Redirect;
	  present_sections.present_Redirect = p;
	  return (prid);

	}
      else if (field == ICMP_FIELD_Echo)
	{
	  p->next = present_sections.present_Echo;
	  present_sections.present_Echo = p;
	  return (prid);


	}
      else if (field == ICMP_FIELD_Timestamp)
	{
	  p->next = present_sections.present_Timestamp;
	  present_sections.present_Timestamp = p;
	  return (prid);


	}
      else if (field == ICMP_FIELD_Information)
	{
	  p->next = present_sections.present_Information;
	  present_sections.present_Information = p;
	  return (prid);


	}

      printf ("That shit (%i) ain't keepin' it real.\n", field);
      return (-1);
    }


  if (field == ICMP_FIELD_Hdr_type)
    {
      if (hdt_sets[ICMP_FIELD_Hdr_type] == 0)
	{
	  hdt_sets[ICMP_FIELD_Hdr_type] = hdt_int32_create ();
	  if (hdt_sets[ICMP_FIELD_Hdr_type] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "type");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[ICMP_FIELD_Hdr_type], prid, op, arg))
	{
	  printf ("Error adding primitive %s in icmp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ICMP_FIELD_Hdr_code)
    {
      if (hdt_sets[ICMP_FIELD_Hdr_code] == 0)
	{
	  hdt_sets[ICMP_FIELD_Hdr_code] = hdt_int32_create ();
	  if (hdt_sets[ICMP_FIELD_Hdr_code] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "code");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[ICMP_FIELD_Hdr_code], prid, op, arg))
	{
	  printf ("Error adding primitive %s in icmp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ICMP_FIELD_Hdr_checksum)
    {
      if (hdt_sets[ICMP_FIELD_Hdr_checksum] == 0)
	{
	  hdt_sets[ICMP_FIELD_Hdr_checksum] = hdt_int32_create ();
	  if (hdt_sets[ICMP_FIELD_Hdr_checksum] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "checksum");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[ICMP_FIELD_Hdr_checksum], prid, op, arg))
	{
	  printf ("Error adding primitive %s in icmp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ICMP_FIELD_Hdr_payload)
    {
      if (hdt_sets[ICMP_FIELD_Hdr_payload] == 0)
	{
	  hdt_sets[ICMP_FIELD_Hdr_payload] = hdt_blob_create ();;
	  if (hdt_sets[ICMP_FIELD_Hdr_payload] == 0)
	    {
	      printf ("AIEE!  Can't add match set for %s.%s!  Exiting.\n",
		      "Hdr", "payload");
	      exit (-1);
	    }
	}
      if (hdt_blob_add (hdt_sets[ICMP_FIELD_Hdr_payload], prid, op, arg))
	{
	  printf ("Error adding primitive %s in icmp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }

  if (field == ICMP_FIELD_Unreachable_unused)
    {
      if (hdt_sets[ICMP_FIELD_Unreachable_unused] == 0)
	{
	  hdt_sets[ICMP_FIELD_Unreachable_unused] = hdt_int32_create ();
	  if (hdt_sets[ICMP_FIELD_Unreachable_unused] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Unreachable", "unused");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ICMP_FIELD_Unreachable_unused], prid, op, arg))
	{
	  printf ("Error adding primitive %s in icmp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }

  if (field == ICMP_FIELD_TTLexceeded_unused)
    {
      if (hdt_sets[ICMP_FIELD_TTLexceeded_unused] == 0)
	{
	  hdt_sets[ICMP_FIELD_TTLexceeded_unused] = hdt_int32_create ();
	  if (hdt_sets[ICMP_FIELD_TTLexceeded_unused] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "TTLexceeded", "unused");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ICMP_FIELD_TTLexceeded_unused], prid, op, arg))
	{
	  printf ("Error adding primitive %s in icmp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }

  if (field == ICMP_FIELD_Parameter_pointer)
    {
      if (hdt_sets[ICMP_FIELD_Parameter_pointer] == 0)
	{
	  hdt_sets[ICMP_FIELD_Parameter_pointer] = hdt_int32_create ();
	  if (hdt_sets[ICMP_FIELD_Parameter_pointer] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Parameter", "pointer");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ICMP_FIELD_Parameter_pointer], prid, op, arg))
	{
	  printf ("Error adding primitive %s in icmp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ICMP_FIELD_Parameter_unused)
    {
      if (hdt_sets[ICMP_FIELD_Parameter_unused] == 0)
	{
	  hdt_sets[ICMP_FIELD_Parameter_unused] = hdt_int32_create ();
	  if (hdt_sets[ICMP_FIELD_Parameter_unused] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Parameter", "unused");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ICMP_FIELD_Parameter_unused], prid, op, arg))
	{
	  printf ("Error adding primitive %s in icmp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }

  if (field == ICMP_FIELD_SourceQuench_unused)
    {
      if (hdt_sets[ICMP_FIELD_SourceQuench_unused] == 0)
	{
	  hdt_sets[ICMP_FIELD_SourceQuench_unused] = hdt_int32_create ();
	  if (hdt_sets[ICMP_FIELD_SourceQuench_unused] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "SourceQuench", "unused");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ICMP_FIELD_SourceQuench_unused], prid, op, arg))
	{
	  printf ("Error adding primitive %s in icmp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }

  if (field == ICMP_FIELD_Redirect_gateway)
    {
      if (hdt_sets[ICMP_FIELD_Redirect_gateway] == 0)
	{
	  hdt_sets[ICMP_FIELD_Redirect_gateway] = hdt_ipv4_addr_create ();;
	  if (hdt_sets[ICMP_FIELD_Redirect_gateway] == 0)
	    {
	      printf ("AIEE!  Can't add %s match set for %s.%s!  Exiting.\n",
		      "ipv4_addr", "Redirect", "gateway");
	      exit (-1);
	    }
	}
      if (hdt_ipv4_addr_add
	  (hdt_sets[ICMP_FIELD_Redirect_gateway], prid, op, arg))
	{
	  printf ("Error adding primitive %s in icmp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }

  if (field == ICMP_FIELD_Echo_id)
    {
      if (hdt_sets[ICMP_FIELD_Echo_id] == 0)
	{
	  hdt_sets[ICMP_FIELD_Echo_id] = hdt_int32_create ();
	  if (hdt_sets[ICMP_FIELD_Echo_id] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Echo", "id");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[ICMP_FIELD_Echo_id], prid, op, arg))
	{
	  printf ("Error adding primitive %s in icmp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ICMP_FIELD_Echo_sequence_number)
    {
      if (hdt_sets[ICMP_FIELD_Echo_sequence_number] == 0)
	{
	  hdt_sets[ICMP_FIELD_Echo_sequence_number] = hdt_int32_create ();
	  if (hdt_sets[ICMP_FIELD_Echo_sequence_number] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Echo", "sequence_number");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ICMP_FIELD_Echo_sequence_number], prid, op, arg))
	{
	  printf ("Error adding primitive %s in icmp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ICMP_FIELD_Echo_payload)
    {
      if (hdt_sets[ICMP_FIELD_Echo_payload] == 0)
	{
	  hdt_sets[ICMP_FIELD_Echo_payload] = hdt_blob_create ();;
	  if (hdt_sets[ICMP_FIELD_Echo_payload] == 0)
	    {
	      printf ("AIEE!  Can't add match set for %s.%s!  Exiting.\n",
		      "Echo", "payload");
	      exit (-1);
	    }
	}
      if (hdt_blob_add (hdt_sets[ICMP_FIELD_Echo_payload], prid, op, arg))
	{
	  printf ("Error adding primitive %s in icmp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }

  if (field == ICMP_FIELD_Timestamp_id)
    {
      if (hdt_sets[ICMP_FIELD_Timestamp_id] == 0)
	{
	  hdt_sets[ICMP_FIELD_Timestamp_id] = hdt_int32_create ();
	  if (hdt_sets[ICMP_FIELD_Timestamp_id] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Timestamp", "id");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[ICMP_FIELD_Timestamp_id], prid, op, arg))
	{
	  printf ("Error adding primitive %s in icmp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ICMP_FIELD_Timestamp_sequence_number)
    {
      if (hdt_sets[ICMP_FIELD_Timestamp_sequence_number] == 0)
	{
	  hdt_sets[ICMP_FIELD_Timestamp_sequence_number] =
	    hdt_int32_create ();
	  if (hdt_sets[ICMP_FIELD_Timestamp_sequence_number] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Timestamp", "sequence_number");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ICMP_FIELD_Timestamp_sequence_number], prid, op, arg))
	{
	  printf ("Error adding primitive %s in icmp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ICMP_FIELD_Timestamp_originate_ts)
    {
      if (hdt_sets[ICMP_FIELD_Timestamp_originate_ts] == 0)
	{
	  hdt_sets[ICMP_FIELD_Timestamp_originate_ts] = hdt_int32_create ();
	  if (hdt_sets[ICMP_FIELD_Timestamp_originate_ts] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Timestamp", "originate_ts");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ICMP_FIELD_Timestamp_originate_ts], prid, op, arg))
	{
	  printf ("Error adding primitive %s in icmp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ICMP_FIELD_Timestamp_receive_ts)
    {
      if (hdt_sets[ICMP_FIELD_Timestamp_receive_ts] == 0)
	{
	  hdt_sets[ICMP_FIELD_Timestamp_receive_ts] = hdt_int32_create ();
	  if (hdt_sets[ICMP_FIELD_Timestamp_receive_ts] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Timestamp", "receive_ts");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ICMP_FIELD_Timestamp_receive_ts], prid, op, arg))
	{
	  printf ("Error adding primitive %s in icmp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ICMP_FIELD_Timestamp_transmit_ts)
    {
      if (hdt_sets[ICMP_FIELD_Timestamp_transmit_ts] == 0)
	{
	  hdt_sets[ICMP_FIELD_Timestamp_transmit_ts] = hdt_int32_create ();
	  if (hdt_sets[ICMP_FIELD_Timestamp_transmit_ts] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Timestamp", "transmit_ts");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ICMP_FIELD_Timestamp_transmit_ts], prid, op, arg))
	{
	  printf ("Error adding primitive %s in icmp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }

  if (field == ICMP_FIELD_Information_id)
    {
      if (hdt_sets[ICMP_FIELD_Information_id] == 0)
	{
	  hdt_sets[ICMP_FIELD_Information_id] = hdt_int32_create ();
	  if (hdt_sets[ICMP_FIELD_Information_id] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Information", "id");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[ICMP_FIELD_Information_id], prid, op, arg))
	{
	  printf ("Error adding primitive %s in icmp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ICMP_FIELD_Information_sequence_number)
    {
      if (hdt_sets[ICMP_FIELD_Information_sequence_number] == 0)
	{
	  hdt_sets[ICMP_FIELD_Information_sequence_number] =
	    hdt_int32_create ();
	  if (hdt_sets[ICMP_FIELD_Information_sequence_number] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Information", "sequence_number");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ICMP_FIELD_Information_sequence_number], prid, op, arg))
	{
	  printf ("Error adding primitive %s in icmp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }


  return (-1);
}

static int
icmp_stock_unreg_prim (prim_id i)
{

  return (-1);
}

static int
icmp_stock_rewrite (struct capsule *c, void *new_payload, u_int32_t len)
{
  if ((c != NULL) && (new_payload != NULL))
    len++;
  return (-1);
}

static char *
icmp_stock_report_field (struct capsule *cp, int field,
			 struct memcon_context *m)
{
  u_int32_t t;
  char *r, buf[1024];
  struct section *sec, *sec2;

  buf[0] = 0;

  switch (field)
    {


    case (ICMP_FIELD_Hdr_type):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ICMP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct ICMP_Hdr *) (sec->content))->type);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (ICMP_FIELD_Hdr_code):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ICMP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct ICMP_Hdr *) (sec->content))->code);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (ICMP_FIELD_Hdr_checksum):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ICMP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct ICMP_Hdr *) (sec->content))->checksum));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;


    case (ICMP_FIELD_Unreachable_unused):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ICMP_SECTION_Unreachable)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct ICMP_Unreachable *) (sec->content))->
			   unused));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "Unreachable");
	}
      break;

    case (ICMP_FIELD_TTLexceeded_unused):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ICMP_SECTION_TTLexceeded)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct ICMP_TTLexceeded *) (sec->content))->
			   unused));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "TTLexceeded");
	}
      break;

    case (ICMP_FIELD_Parameter_pointer):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ICMP_SECTION_Parameter)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct ICMP_Parameter *) (sec->content))->pointer);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "Parameter");
	}
      break;
    case (ICMP_FIELD_Parameter_unused):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ICMP_SECTION_Parameter)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct ICMP_Parameter *) (sec->content))->
			   unused));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "Parameter");
	}
      break;

    case (ICMP_FIELD_SourceQuench_unused):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ICMP_SECTION_SourceQuench)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct ICMP_SourceQuench *) (sec->content))->
			   unused));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "SourceQuench");
	}
      break;

    case (ICMP_FIELD_Redirect_gateway):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ICMP_SECTION_Redirect)
	    sec = sec2;
	}
      if (sec)
	{
	  return (hdt_ipv4_addr_report
		  (((struct ICMP_Redirect *) (sec->content))->gateway, m));
	}
      snprintf (buf, sizeof (buf), "(section not present)");
      break;

    case (ICMP_FIELD_Echo_id):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ICMP_SECTION_Echo)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct ICMP_Echo *) (sec->content))->id));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Echo");
	}
      break;
    case (ICMP_FIELD_Echo_sequence_number):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ICMP_SECTION_Echo)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct ICMP_Echo *) (sec->content))->
			   sequence_number));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Echo");
	}
      break;


    case (ICMP_FIELD_Timestamp_id):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ICMP_SECTION_Timestamp)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct ICMP_Timestamp *) (sec->content))->id));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "Timestamp");
	}
      break;
    case (ICMP_FIELD_Timestamp_sequence_number):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ICMP_SECTION_Timestamp)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct ICMP_Timestamp *) (sec->content))->
			   sequence_number));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "Timestamp");
	}
      break;
    case (ICMP_FIELD_Timestamp_originate_ts):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ICMP_SECTION_Timestamp)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct ICMP_Timestamp *) (sec->content))->
			   originate_ts));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "Timestamp");
	}
      break;
    case (ICMP_FIELD_Timestamp_receive_ts):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ICMP_SECTION_Timestamp)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct ICMP_Timestamp *) (sec->content))->
			   receive_ts));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "Timestamp");
	}
      break;
    case (ICMP_FIELD_Timestamp_transmit_ts):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ICMP_SECTION_Timestamp)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct ICMP_Timestamp *) (sec->content))->
			   transmit_ts));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "Timestamp");
	}
      break;

    case (ICMP_FIELD_Information_id):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ICMP_SECTION_Information)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct ICMP_Information *) (sec->content))->id));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "Information");
	}
      break;
    case (ICMP_FIELD_Information_sequence_number):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ICMP_SECTION_Information)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct ICMP_Information *) (sec->content))->
			   sequence_number));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "Information");
	}
      break;


    default:

      return (NULL);
    }
  t = (strlen (buf) + 1);
  r = m->eng->alloc (m, t);
  strncpy (r, buf, t);
  return (r);
}

static int
icmp_stock_find_matches (struct capsule *cp, struct prim_can *p)
{
  struct section *s;
  struct memcon_context *c = cp->parent->c;
  u_int32_t v;
  unsigned long beg, end;

  if (p)
    return (-1);

  if (cp->matches == NULL && gprid)
    {
      if (!(cp->matches = c->eng->alloc (c, gprid + 1)))
	exit (-1);
      memset (cp->matches, 0, gprid + 1);

      cp->matches[0] = 1;
    }


  for (s = cp->sections; s; s = s->next)
    {
      struct simple_prim *sp;

      for (sp = present_sections.proto_self; sp; sp = sp->next)
	{
	  ;
	  cp->matches[sp->prim] |= 1;
	}


      switch (s->section_id)
	{
	case ICMP_SECTION_INVALID:
	  return (-1);


	  break;
	case ICMP_SECTION_Hdr:;
	  for (sp = present_sections.present_Hdr; sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[ICMP_FIELD_Hdr_type])
	    {;
	      v = ((struct ICMP_Hdr *) (s->content))->type;;
	      if (hdt_int32_match
		  (hdt_sets[ICMP_FIELD_Hdr_type], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr", "type",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ICMP_FIELD_Hdr_code])
	    {;
	      v = ((struct ICMP_Hdr *) (s->content))->code;;
	      if (hdt_int32_match
		  (hdt_sets[ICMP_FIELD_Hdr_code], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr", "code",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ICMP_FIELD_Hdr_checksum])
	    {;
	      v = ntohs (((struct ICMP_Hdr *) (s->content))->checksum);;
	      if (hdt_int32_match
		  (hdt_sets[ICMP_FIELD_Hdr_checksum], NULL, v, cp->matches,
		   1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "checksum", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ICMP_FIELD_Hdr_payload])
	    {
	      struct blob bb;
	      bb.blob = &(((struct ICMP_Hdr *) (s->content))->payload);
	      bb.len = cp->length - offsetof (struct ICMP_Hdr, payload);;
	      if (hdt_blob_match
		  (hdt_sets[ICMP_FIELD_Hdr_payload], NULL, &bb, cp->matches,
		   1))
		{
		  printf ("AIEE!  Can't find blob match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "payload", (end - beg));
	  break;
	case ICMP_SECTION_Unreachable:;
	  for (sp = present_sections.present_Unreachable; sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[ICMP_FIELD_Unreachable_unused])
	    {;
	      v = ntohl (((struct ICMP_Unreachable *) (s->content))->unused);;
	      if (hdt_int32_match
		  (hdt_sets[ICMP_FIELD_Unreachable_unused], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Unreachable",
		  "unused", (end - beg));
	  break;
	case ICMP_SECTION_TTLexceeded:;
	  for (sp = present_sections.present_TTLexceeded; sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[ICMP_FIELD_TTLexceeded_unused])
	    {;
	      v = ntohl (((struct ICMP_TTLexceeded *) (s->content))->unused);;
	      if (hdt_int32_match
		  (hdt_sets[ICMP_FIELD_TTLexceeded_unused], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "TTLexceeded",
		  "unused", (end - beg));
	  break;
	case ICMP_SECTION_Parameter:;
	  for (sp = present_sections.present_Parameter; sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[ICMP_FIELD_Parameter_pointer])
	    {;
	      v = ((struct ICMP_Parameter *) (s->content))->pointer;;
	      if (hdt_int32_match
		  (hdt_sets[ICMP_FIELD_Parameter_pointer], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Parameter",
		  "pointer", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ICMP_FIELD_Parameter_unused])
	    {;
	      v = ntohl (((struct ICMP_Parameter *) (s->content))->unused);;
	      if (hdt_int32_match
		  (hdt_sets[ICMP_FIELD_Parameter_unused], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Parameter",
		  "unused", (end - beg));
	  break;
	case ICMP_SECTION_SourceQuench:;
	  for (sp = present_sections.present_SourceQuench; sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[ICMP_FIELD_SourceQuench_unused])
	    {;
	      v =
		ntohl (((struct ICMP_SourceQuench *) (s->content))->unused);;
	      if (hdt_int32_match
		  (hdt_sets[ICMP_FIELD_SourceQuench_unused], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "SourceQuench", "unused", (end - beg));
	  break;
	case ICMP_SECTION_Redirect:;
	  for (sp = present_sections.present_Redirect; sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[ICMP_FIELD_Redirect_gateway])
	    {;
	      if (hdt_ipv4_addr_match
		  (hdt_sets[ICMP_FIELD_Redirect_gateway], NULL,
		   ((struct ICMP_Redirect *) (s->content))->gateway,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find %s match.  Exiting.\n",
			  "ipv4_addr");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Redirect",
		  "gateway", (end - beg));
	  break;
	case ICMP_SECTION_Echo:;
	  for (sp = present_sections.present_Echo; sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[ICMP_FIELD_Echo_id])
	    {;
	      v = ntohs (((struct ICMP_Echo *) (s->content))->id);;
	      if (hdt_int32_match
		  (hdt_sets[ICMP_FIELD_Echo_id], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Echo", "id",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ICMP_FIELD_Echo_sequence_number])
	    {;
	      v =
		ntohs (((struct ICMP_Echo *) (s->content))->sequence_number);;
	      if (hdt_int32_match
		  (hdt_sets[ICMP_FIELD_Echo_sequence_number], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Echo",
		  "sequence_number", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ICMP_FIELD_Echo_payload])
	    {
	      struct blob bb;
	      bb.blob = &(((struct ICMP_Echo *) (s->content))->payload);
	      bb.len = cp->length - offsetof (struct ICMP_Echo, payload);;
	      if (hdt_blob_match
		  (hdt_sets[ICMP_FIELD_Echo_payload], NULL, &bb, cp->matches,
		   1))
		{
		  printf ("AIEE!  Can't find blob match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Echo",
		  "payload", (end - beg));
	  break;
	case ICMP_SECTION_Timestamp:;
	  for (sp = present_sections.present_Timestamp; sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[ICMP_FIELD_Timestamp_id])
	    {;
	      v = ntohs (((struct ICMP_Timestamp *) (s->content))->id);;
	      if (hdt_int32_match
		  (hdt_sets[ICMP_FIELD_Timestamp_id], NULL, v, cp->matches,
		   1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Timestamp",
		  "id", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ICMP_FIELD_Timestamp_sequence_number])
	    {;
	      v =
		ntohs (((struct ICMP_Timestamp *) (s->content))->
		       sequence_number);;
	      if (hdt_int32_match
		  (hdt_sets[ICMP_FIELD_Timestamp_sequence_number], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Timestamp",
		  "sequence_number", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ICMP_FIELD_Timestamp_originate_ts])
	    {;
	      v =
		ntohl (((struct ICMP_Timestamp *) (s->content))->
		       originate_ts);;
	      if (hdt_int32_match
		  (hdt_sets[ICMP_FIELD_Timestamp_originate_ts], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Timestamp",
		  "originate_ts", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ICMP_FIELD_Timestamp_receive_ts])
	    {;
	      v =
		ntohl (((struct ICMP_Timestamp *) (s->content))->receive_ts);;
	      if (hdt_int32_match
		  (hdt_sets[ICMP_FIELD_Timestamp_receive_ts], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Timestamp",
		  "receive_ts", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ICMP_FIELD_Timestamp_transmit_ts])
	    {;
	      v =
		ntohl (((struct ICMP_Timestamp *) (s->content))->
		       transmit_ts);;
	      if (hdt_int32_match
		  (hdt_sets[ICMP_FIELD_Timestamp_transmit_ts], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Timestamp",
		  "transmit_ts", (end - beg));
	  break;
	case ICMP_SECTION_Information:;
	  for (sp = present_sections.present_Information; sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[ICMP_FIELD_Information_id])
	    {;
	      v = ntohs (((struct ICMP_Information *) (s->content))->id);;
	      if (hdt_int32_match
		  (hdt_sets[ICMP_FIELD_Information_id], NULL, v, cp->matches,
		   1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Information",
		  "id", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ICMP_FIELD_Information_sequence_number])
	    {;
	      v =
		ntohs (((struct ICMP_Information *) (s->content))->
		       sequence_number);;
	      if (hdt_int32_match
		  (hdt_sets[ICMP_FIELD_Information_sequence_number], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Information",
		  "sequence_number", (end - beg));


	  break;
	default:
	  printf ("Unrecognized section: %i (not %i).\n", s->section_id,
		  ICMP_SECTION_Hdr);
	  break;
	}
    }

  return (-1);
}
