/* Copyright 2001, Todd Lewis <tlewis@mindspring.com>
 * All rights reserved. */

#include "hank.h"
#include "hdt_blob.h"
#include "hdt_int32.h"
#include "packet.h"
#include "protocol_engine.h"
#include "protocols.h"
#include "support.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.h"

#include "ipv4.hi"


static char rcsid[] = "$Id: ipv4.xml,v 1.5 2001/10/31 21:21:05 tlewis Exp $";


extern struct smodule *ipv4[];


extern struct protocol_engine ipv4_pe;


static u_int32_t protocols[] = {

  ((HPN_802 << 24) | 2048),
  ((HPN_PCAP << 24) | DLT_RAW),
  ((HPN_GRE << 24) | 2048),
  ((HPN_PPP << 24) | 0x21),
  ((HPN_IPV4 << 24) | 4),
  ((HPN_IPV4 << 24) | 94),


  0
};

struct simple_prim
{
  prim_id prim;
  struct simple_prim *next;
};

static struct
{
  struct simple_prim *proto_self;


  struct simple_prim *present_Hdr;


  struct simple_prim *present_flags_offset;

  struct simple_prim *present_class_one_option;

  struct simple_prim *present_security_option;


  struct simple_prim *present_loose_source_and_record_route_option;


  struct simple_prim *present_strict_source_and_record_route_option;


  struct simple_prim *present_record_route_option;


  struct simple_prim *present_internet_timestamp_option;


  struct simple_prim *present_stream_identifier_option;


}
present_sections;

enum IPV4_FIELD
{
  IPV4_SELF,


  IPV4_FIELD_Hdr,
  IPV4_FIELD_Hdr_header_length,
  IPV4_FIELD_Hdr_version,
  IPV4_FIELD_Hdr_tos,
  IPV4_FIELD_Hdr_length,
  IPV4_FIELD_Hdr_id,
  IPV4_FIELD_Hdr_flags_offset,
  IPV4_FIELD_Hdr_ttl,
  IPV4_FIELD_Hdr_protocol,
  IPV4_FIELD_Hdr_checksum,
  IPV4_FIELD_Hdr_source,
  IPV4_FIELD_Hdr_destination,
  IPV4_FIELD_Hdr_payload,
  IPV4_FIELD_flags_offset,
  IPV4_FIELD_flags_offset_flags_offset,
  IPV4_FIELD_class_one_option,
  IPV4_FIELD_class_one_option_option,
  IPV4_FIELD_security_option,
  IPV4_FIELD_security_option_type,
  IPV4_FIELD_security_option_length,
  IPV4_FIELD_security_option_security,
  IPV4_FIELD_security_option_compartments,
  IPV4_FIELD_security_option_handling_restrictions,
  IPV4_FIELD_security_option_unused,
  IPV4_FIELD_security_option_tcc_field,
  IPV4_FIELD_loose_source_and_record_route_option,
  IPV4_FIELD_loose_source_and_record_route_option_type,
  IPV4_FIELD_loose_source_and_record_route_option_length,
  IPV4_FIELD_loose_source_and_record_route_option_pointer,
  IPV4_FIELD_strict_source_and_record_route_option,
  IPV4_FIELD_strict_source_and_record_route_option_type,
  IPV4_FIELD_strict_source_and_record_route_option_length,
  IPV4_FIELD_strict_source_and_record_route_option_pointer,
  IPV4_FIELD_record_route_option,
  IPV4_FIELD_record_route_option_type,
  IPV4_FIELD_record_route_option_length,
  IPV4_FIELD_record_route_option_pointer,
  IPV4_FIELD_internet_timestamp_option,
  IPV4_FIELD_internet_timestamp_option_type,
  IPV4_FIELD_internet_timestamp_option_length,
  IPV4_FIELD_internet_timestamp_option_flag,
  IPV4_FIELD_internet_timestamp_option_overflow,
  IPV4_FIELD_internet_timestamp_option_pointer,
  IPV4_FIELD_stream_identifier_option,
  IPV4_FIELD_stream_identifier_option_type,
  IPV4_FIELD_stream_identifier_option_length,
  IPV4_FIELD_stream_identifier_option_stream_id,


  IPV4_FIELD_INVALID
};

enum IPV4_SECTIONS
{


  IPV4_SECTION_Hdr,


  IPV4_SECTION_flags_offset,

  IPV4_SECTION_class_one_option,

  IPV4_SECTION_security_option,


  IPV4_SECTION_loose_source_and_record_route_option,


  IPV4_SECTION_strict_source_and_record_route_option,


  IPV4_SECTION_record_route_option,


  IPV4_SECTION_internet_timestamp_option,


  IPV4_SECTION_stream_identifier_option,


  IPV4_SECTION_INVALID
};

struct IPV4_null
{
  int empty;


};
struct IPV4_Hdr
{
  u_int32_t header_length:4;
  u_int32_t version:4;
  u_int8_t tos;
  u_int16_t length;
  u_int16_t id;
  u_int16_t flags_offset;
  u_int8_t ttl;
  u_int8_t protocol;
  u_int16_t checksum;
  ipv4_addr source;
  ipv4_addr destination;
  unsigned char payload;
};
struct IPV4_flags_offset
{
  u_int16_t flags_offset;
};
struct IPV4_class_one_option
{
  u_int8_t option;
};
struct IPV4_security_option
{
  u_int8_t type;
  u_int8_t length;
  u_int16_t security;
  u_int16_t compartments;
  u_int16_t handling_restrictions;
  u_int8_t unused;
  u_int32_t tcc_field:24;
};
struct IPV4_loose_source_and_record_route_option
{
  u_int8_t type;
  u_int8_t length;
  u_int8_t pointer;
};
struct IPV4_strict_source_and_record_route_option
{
  u_int8_t type;
  u_int8_t length;
  u_int8_t pointer;
};
struct IPV4_record_route_option
{
  u_int8_t type;
  u_int8_t length;
  u_int8_t pointer;
};
struct IPV4_internet_timestamp_option
{
  u_int8_t type;
  u_int8_t length;
  u_int32_t flag:4;
  u_int32_t overflow:4;
  u_int8_t pointer;
};
struct IPV4_stream_identifier_option
{
  u_int8_t type;
  u_int8_t length;
  u_int16_t stream_id;


};


static char hexes[] = "0123456789abcdef";

static int gprid = 0;

static hdt_set hdt_sets[IPV4_FIELD_INVALID];


static int ipv4_stock_reg_prim (int field, char *op, char *arg);
static int ipv4_stock_unreg_prim (prim_id i);
static int ipv4_stock_decompose (struct packet *p);
static int ipv4_stock_find_matches (struct capsule *cp, struct prim_can *c);
static int ipv4_stock_rewrite (struct capsule *c, void *new_payload,
			       u_int32_t len);
static char *ipv4_stock_report_field (struct capsule *p, int field,
				      struct memcon_context *m);
static int ipv4_field_lookup (char *section, char *field);

struct protocol_engine ipv4_pe = {
  {
   "pe",
   "ipv4",
   PEVMIN,
   PEVMAJ,
   &ipv4_pe},
  &ipv4_private_decompose,
  &ipv4_private_reg_prim,
  &ipv4_stock_unreg_prim,
  &ipv4_private_find_matches,
  &ipv4_stock_rewrite,
  &ipv4_stock_report_field,
  &ipv4_field_lookup,
  &protocols[0],
  0
};


struct smodule *ipv4[] = {

  &(ipv4_pe.s),
  NULL
};

static int
ipv4_private_decompose (struct packet *p)
{
  struct capsule *e = p->encaps, *f;
  struct IPV4_Hdr *h;
  struct section *s, *t;
  int i;
  unsigned char *oc;

  while (e->next)
    e = e->next;

  h = (struct IPV4_Hdr *) e->payload;

  append_encap (p, protocols[0], e->payload_length, e->payload);
  f = e->next;
  f->pe = &ipv4_pe;

  if ((s = p->c->eng->alloc (p->c, sizeof (*s))) == NULL)
    exit (-1);
  s->section_id = IPV4_SECTION_Hdr;
  s->content = h;
  s->next = NULL;
  f->sections = s;
  f->length = e->payload_length;

  for (i = 20; i < (h->header_length * 4);)
    {
      oc = &e->payload[i];


      if (oc[0] == 0)
	{

	  break;

	}
      else if (oc[0] == 1)
	{

	  i++;

	}
      else if (oc[0] == 130)
	{

	  i += oc[1];
	  t = s;
	  if ((s = p->c->eng->alloc (p->c, sizeof (*s))) == NULL)
	    exit (-1);
	  s->section_id = IPV4_SECTION_security_option;
	  s->content = &oc[0];
	  t->next = s;
	  s->next = NULL;

	}
      else if (oc[0] == 131)
	{

	  i += oc[1];
	  t = s;
	  if ((s = p->c->eng->alloc (p->c, sizeof (*s))) == NULL)
	    exit (-1);
	  s->section_id = IPV4_SECTION_loose_source_and_record_route_option;
	  s->content = &oc[0];
	  t->next = s;
	  s->next = NULL;

	}
      else if (oc[0] == 137)
	{

	  i += oc[1];
	  t = s;
	  if ((s = p->c->eng->alloc (p->c, sizeof (*s))) == NULL)
	    exit (-1);
	  s->section_id = IPV4_SECTION_strict_source_and_record_route_option;
	  s->content = &oc[0];
	  t->next = s;
	  s->next = NULL;

	}
      else if (oc[0] == 7)
	{

	  i += oc[1];
	  t = s;
	  if ((s = p->c->eng->alloc (p->c, sizeof (*s))) == NULL)
	    exit (-1);
	  s->section_id = IPV4_SECTION_record_route_option;
	  s->content = &oc[0];
	  t->next = s;
	  s->next = NULL;

	}
      else if (oc[0] == 136)
	{

	  i += oc[1];
	  t = s;
	  if ((s = p->c->eng->alloc (p->c, sizeof (*s))) == NULL)
	    exit (-1);
	  s->section_id = IPV4_SECTION_stream_identifier_option;
	  s->content = &oc[0];
	  t->next = s;
	  s->next = NULL;

	}
      else if (oc[0] == 68)
	{

	  i += oc[1];
	  t = s;
	  if ((s = p->c->eng->alloc (p->c, sizeof (*s))) == NULL)
	    exit (-1);
	  s->section_id = IPV4_SECTION_internet_timestamp_option;
	  s->content = &oc[0];
	  t->next = s;
	  s->next = NULL;
	}
      else
	{

	}
    }

  f->payload_length = (f->length - (4 * h->header_length));
  f->payload_protocol = (u_int32_t) ((HPN_IPV4 << 24) | h->protocol);
  f->payload = &(e->payload[(4 * h->header_length)]);

  return (0);
}

static prim_id
ipv4_private_reg_prim (int field, char *op, char *arg)
{
  struct ipv4_sameip *s;

  if (!strcmp (op, "sameip"))
    {
      int prid = ++gprid;

      if ((s = malloc (sizeof (*s))) == NULL)
	{
	  fprintf (stderr, "Aiee!  Malloc failed; aborting.n");
	  exit (-1);
	}
      s->pid = prid;
      s->next = sameips;
      sameips = s;
      return (prid);
    }

  return (ipv4_stock_reg_prim (field, op, arg));
}

static int
ipv4_private_find_matches (struct capsule *e, struct prim_can *c)
{
  struct ipv4_sameip *s;
  struct section *t;
  struct IPV4_Hdr *h;
  int i;

  i = ipv4_stock_find_matches (e, c);


  for (t = e->sections; t; t = t->next)
    {
      if (t->section_id == IPV4_SECTION_Hdr)
	{
	  h = t->content;
	  if (h->source == h->destination)
	    {
	      for (s = sameips; s; s = s->next)
		e->matches[s->pid] = 1;
	      break;
	    }
	}
    }

  return (i);
}


static int
ipv4_field_lookup (char *section, char *field)
{
  int p = -1;

  if ((section == NULL) && (field == NULL))
    {
      p = IPV4_SELF;
    }
  else if (section != NULL && !(strcmp (section, "INVALID")))
    {
      p = IPV4_FIELD_INVALID;


    }
  else if ((section == NULL) || (!(strcmp (section, "Hdr"))))
    {
      if (field == NULL)
	p = IPV4_FIELD_Hdr;
      else if (!(strcmp (field, "header_length")))
	p = IPV4_FIELD_Hdr_header_length;
      else if (!(strcmp (field, "version")))
	p = IPV4_FIELD_Hdr_version;
      else if (!(strcmp (field, "tos")))
	p = IPV4_FIELD_Hdr_tos;
      else if (!(strcmp (field, "length")))
	p = IPV4_FIELD_Hdr_length;
      else if (!(strcmp (field, "id")))
	p = IPV4_FIELD_Hdr_id;
      else if (!(strcmp (field, "flags_offset")))
	p = IPV4_FIELD_Hdr_flags_offset;
      else if (!(strcmp (field, "ttl")))
	p = IPV4_FIELD_Hdr_ttl;
      else if (!(strcmp (field, "protocol")))
	p = IPV4_FIELD_Hdr_protocol;
      else if (!(strcmp (field, "checksum")))
	p = IPV4_FIELD_Hdr_checksum;
      else if (!(strcmp (field, "source")))
	p = IPV4_FIELD_Hdr_source;
      else if (!(strcmp (field, "destination")))
	p = IPV4_FIELD_Hdr_destination;
      else if (!(strcmp (field, "payload")))
	p = IPV4_FIELD_Hdr_payload;
    }
  else if ((section == NULL) || (!(strcmp (section, "flags_offset"))))
    {
      if (field == NULL)
	p = IPV4_FIELD_flags_offset;
      else if (!(strcmp (field, "flags_offset")))
	p = IPV4_FIELD_flags_offset_flags_offset;
    }
  else if ((section == NULL) || (!(strcmp (section, "class_one_option"))))
    {
      if (field == NULL)
	p = IPV4_FIELD_class_one_option;
      else if (!(strcmp (field, "option")))
	p = IPV4_FIELD_class_one_option_option;
    }
  else if ((section == NULL) || (!(strcmp (section, "security_option"))))
    {
      if (field == NULL)
	p = IPV4_FIELD_security_option;
      else if (!(strcmp (field, "type")))
	p = IPV4_FIELD_security_option_type;
      else if (!(strcmp (field, "length")))
	p = IPV4_FIELD_security_option_length;
      else if (!(strcmp (field, "security")))
	p = IPV4_FIELD_security_option_security;
      else if (!(strcmp (field, "compartments")))
	p = IPV4_FIELD_security_option_compartments;
      else if (!(strcmp (field, "handling_restrictions")))
	p = IPV4_FIELD_security_option_handling_restrictions;
      else if (!(strcmp (field, "unused")))
	p = IPV4_FIELD_security_option_unused;
      else if (!(strcmp (field, "tcc_field")))
	p = IPV4_FIELD_security_option_tcc_field;
    }
  else if ((section == NULL)
	   || (!(strcmp (section, "loose_source_and_record_route_option"))))
    {
      if (field == NULL)
	p = IPV4_FIELD_loose_source_and_record_route_option;
      else if (!(strcmp (field, "type")))
	p = IPV4_FIELD_loose_source_and_record_route_option_type;
      else if (!(strcmp (field, "length")))
	p = IPV4_FIELD_loose_source_and_record_route_option_length;
      else if (!(strcmp (field, "pointer")))
	p = IPV4_FIELD_loose_source_and_record_route_option_pointer;
    }
  else if ((section == NULL)
	   || (!(strcmp (section, "strict_source_and_record_route_option"))))
    {
      if (field == NULL)
	p = IPV4_FIELD_strict_source_and_record_route_option;
      else if (!(strcmp (field, "type")))
	p = IPV4_FIELD_strict_source_and_record_route_option_type;
      else if (!(strcmp (field, "length")))
	p = IPV4_FIELD_strict_source_and_record_route_option_length;
      else if (!(strcmp (field, "pointer")))
	p = IPV4_FIELD_strict_source_and_record_route_option_pointer;
    }
  else if ((section == NULL) || (!(strcmp (section, "record_route_option"))))
    {
      if (field == NULL)
	p = IPV4_FIELD_record_route_option;
      else if (!(strcmp (field, "type")))
	p = IPV4_FIELD_record_route_option_type;
      else if (!(strcmp (field, "length")))
	p = IPV4_FIELD_record_route_option_length;
      else if (!(strcmp (field, "pointer")))
	p = IPV4_FIELD_record_route_option_pointer;
    }
  else if ((section == NULL)
	   || (!(strcmp (section, "internet_timestamp_option"))))
    {
      if (field == NULL)
	p = IPV4_FIELD_internet_timestamp_option;
      else if (!(strcmp (field, "type")))
	p = IPV4_FIELD_internet_timestamp_option_type;
      else if (!(strcmp (field, "length")))
	p = IPV4_FIELD_internet_timestamp_option_length;
      else if (!(strcmp (field, "flag")))
	p = IPV4_FIELD_internet_timestamp_option_flag;
      else if (!(strcmp (field, "overflow")))
	p = IPV4_FIELD_internet_timestamp_option_overflow;
      else if (!(strcmp (field, "pointer")))
	p = IPV4_FIELD_internet_timestamp_option_pointer;
    }
  else if ((section == NULL)
	   || (!(strcmp (section, "stream_identifier_option"))))
    {
      if (field == NULL)
	p = IPV4_FIELD_stream_identifier_option;
      else if (!(strcmp (field, "type")))
	p = IPV4_FIELD_stream_identifier_option_type;
      else if (!(strcmp (field, "length")))
	p = IPV4_FIELD_stream_identifier_option_length;
      else if (!(strcmp (field, "stream_id")))
	p = IPV4_FIELD_stream_identifier_option_stream_id;


    }

  if (p == -1)
    {
      printf ("IPV4: Unrecognized field: \" %s \".  Returning error.\n",
	      field);
      return (IPV4_FIELD_INVALID);
    }

  ;
  return (p);
}


static int
ipv4_stock_decompose (struct packet *p)
{
  struct capsule *e = p->encaps, *f;
  struct IPV4_Hdr *h;
  struct section *s;


  while (e->next)
    e = e->next;

  h = (struct IPV4_Hdr *) e->payload;


  append_encap (p, protocols[0], e->payload_length, e->payload);
  f = e->next;
  f->pe = &ipv4_pe;

  if ((s = p->c->eng->alloc (p->c, sizeof (*s))) == NULL)
    exit (-1);
  s->section_id = IPV4_SECTION_Hdr;
  s->content = h;
  s->next = NULL;
  f->sections = s;
  f->length = e->payload_length;


  f->payload_length = (f->length - (4 * h->header_length));
  f->payload_protocol = (u_int32_t) ((HPN_IPV4 << 24) | h->protocol);
  f->payload = &(e->payload[(4 * h->header_length)]);


  return (0);
}

static int
ipv4_stock_reg_prim (int field, char *op, char *arg)
{
  int prid = ++gprid;


  ;

  if (field == IPV4_FIELD_INVALID)
    {
      printf
	("IPV4: Error.  Received invalid field in reg_prim(), op=%s, arg=%s.\n",
	 op, arg);
      return (-1);
    }


  if (!strcmp (op, "present"))
    {
      struct simple_prim *p;

      if (!(p = malloc (sizeof (*p))))
	exit (-1);
      p->prim = prid;

      if (field == IPV4_SELF)
	{

	  p->next = present_sections.proto_self;
	  present_sections.proto_self = p;
	  return (prid);


	}
      else if (field == IPV4_FIELD_Hdr)
	{
	  p->next = present_sections.present_Hdr;
	  present_sections.present_Hdr = p;
	  return (prid);


	}
      else if (field == IPV4_FIELD_flags_offset)
	{
	  p->next = present_sections.present_flags_offset;
	  present_sections.present_flags_offset = p;
	  return (prid);

	}
      else if (field == IPV4_FIELD_class_one_option)
	{
	  p->next = present_sections.present_class_one_option;
	  present_sections.present_class_one_option = p;
	  return (prid);

	}
      else if (field == IPV4_FIELD_security_option)
	{
	  p->next = present_sections.present_security_option;
	  present_sections.present_security_option = p;
	  return (prid);


	}
      else if (field == IPV4_FIELD_loose_source_and_record_route_option)
	{
	  p->next =
	    present_sections.present_loose_source_and_record_route_option;
	  present_sections.present_loose_source_and_record_route_option = p;
	  return (prid);


	}
      else if (field == IPV4_FIELD_strict_source_and_record_route_option)
	{
	  p->next =
	    present_sections.present_strict_source_and_record_route_option;
	  present_sections.present_strict_source_and_record_route_option = p;
	  return (prid);


	}
      else if (field == IPV4_FIELD_record_route_option)
	{
	  p->next = present_sections.present_record_route_option;
	  present_sections.present_record_route_option = p;
	  return (prid);


	}
      else if (field == IPV4_FIELD_internet_timestamp_option)
	{
	  p->next = present_sections.present_internet_timestamp_option;
	  present_sections.present_internet_timestamp_option = p;
	  return (prid);


	}
      else if (field == IPV4_FIELD_stream_identifier_option)
	{
	  p->next = present_sections.present_stream_identifier_option;
	  present_sections.present_stream_identifier_option = p;
	  return (prid);


	}

      printf ("That shit (%i) ain't keepin' it real.\n", field);
      return (-1);
    }


  if (field == IPV4_FIELD_Hdr_header_length)
    {
      if (hdt_sets[IPV4_FIELD_Hdr_header_length] == 0)
	{
	  hdt_sets[IPV4_FIELD_Hdr_header_length] = hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_Hdr_header_length] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "header_length");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_Hdr_header_length], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_Hdr_version)
    {
      if (hdt_sets[IPV4_FIELD_Hdr_version] == 0)
	{
	  hdt_sets[IPV4_FIELD_Hdr_version] = hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_Hdr_version] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "version");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[IPV4_FIELD_Hdr_version], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_Hdr_tos)
    {
      if (hdt_sets[IPV4_FIELD_Hdr_tos] == 0)
	{
	  hdt_sets[IPV4_FIELD_Hdr_tos] = hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_Hdr_tos] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "tos");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[IPV4_FIELD_Hdr_tos], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_Hdr_length)
    {
      if (hdt_sets[IPV4_FIELD_Hdr_length] == 0)
	{
	  hdt_sets[IPV4_FIELD_Hdr_length] = hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_Hdr_length] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "length");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[IPV4_FIELD_Hdr_length], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_Hdr_id)
    {
      if (hdt_sets[IPV4_FIELD_Hdr_id] == 0)
	{
	  hdt_sets[IPV4_FIELD_Hdr_id] = hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_Hdr_id] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "id");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[IPV4_FIELD_Hdr_id], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_Hdr_flags_offset)
    {
      if (hdt_sets[IPV4_FIELD_Hdr_flags_offset] == 0)
	{
	  hdt_sets[IPV4_FIELD_Hdr_flags_offset] = hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_Hdr_flags_offset] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "flags_offset");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_Hdr_flags_offset], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_Hdr_ttl)
    {
      if (hdt_sets[IPV4_FIELD_Hdr_ttl] == 0)
	{
	  hdt_sets[IPV4_FIELD_Hdr_ttl] = hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_Hdr_ttl] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "ttl");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[IPV4_FIELD_Hdr_ttl], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_Hdr_protocol)
    {
      if (hdt_sets[IPV4_FIELD_Hdr_protocol] == 0)
	{
	  hdt_sets[IPV4_FIELD_Hdr_protocol] = hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_Hdr_protocol] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "protocol");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[IPV4_FIELD_Hdr_protocol], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_Hdr_checksum)
    {
      if (hdt_sets[IPV4_FIELD_Hdr_checksum] == 0)
	{
	  hdt_sets[IPV4_FIELD_Hdr_checksum] = hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_Hdr_checksum] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "checksum");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[IPV4_FIELD_Hdr_checksum], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_Hdr_source)
    {
      if (hdt_sets[IPV4_FIELD_Hdr_source] == 0)
	{
	  hdt_sets[IPV4_FIELD_Hdr_source] = hdt_ipv4_addr_create ();;
	  if (hdt_sets[IPV4_FIELD_Hdr_source] == 0)
	    {
	      printf ("AIEE!  Can't add %s match set for %s.%s!  Exiting.\n",
		      "ipv4_addr", "Hdr", "source");
	      exit (-1);
	    }
	}
      if (hdt_ipv4_addr_add (hdt_sets[IPV4_FIELD_Hdr_source], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_Hdr_destination)
    {
      if (hdt_sets[IPV4_FIELD_Hdr_destination] == 0)
	{
	  hdt_sets[IPV4_FIELD_Hdr_destination] = hdt_ipv4_addr_create ();;
	  if (hdt_sets[IPV4_FIELD_Hdr_destination] == 0)
	    {
	      printf ("AIEE!  Can't add %s match set for %s.%s!  Exiting.\n",
		      "ipv4_addr", "Hdr", "destination");
	      exit (-1);
	    }
	}
      if (hdt_ipv4_addr_add
	  (hdt_sets[IPV4_FIELD_Hdr_destination], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_Hdr_payload)
    {
      if (hdt_sets[IPV4_FIELD_Hdr_payload] == 0)
	{
	  hdt_sets[IPV4_FIELD_Hdr_payload] = hdt_blob_create ();;
	  if (hdt_sets[IPV4_FIELD_Hdr_payload] == 0)
	    {
	      printf ("AIEE!  Can't add match set for %s.%s!  Exiting.\n",
		      "Hdr", "payload");
	      exit (-1);
	    }
	}
      if (hdt_blob_add (hdt_sets[IPV4_FIELD_Hdr_payload], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }

  if (field == IPV4_FIELD_flags_offset_flags_offset)
    {
      if (hdt_sets[IPV4_FIELD_flags_offset_flags_offset] == 0)
	{
	  hdt_sets[IPV4_FIELD_flags_offset_flags_offset] =
	    hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_flags_offset_flags_offset] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "flags_offset", "flags_offset");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_flags_offset_flags_offset], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }

  if (field == IPV4_FIELD_class_one_option_option)
    {
      if (hdt_sets[IPV4_FIELD_class_one_option_option] == 0)
	{
	  hdt_sets[IPV4_FIELD_class_one_option_option] = hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_class_one_option_option] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "class_one_option", "option");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_class_one_option_option], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }

  if (field == IPV4_FIELD_security_option_type)
    {
      if (hdt_sets[IPV4_FIELD_security_option_type] == 0)
	{
	  hdt_sets[IPV4_FIELD_security_option_type] = hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_security_option_type] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "security_option", "type");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_security_option_type], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_security_option_length)
    {
      if (hdt_sets[IPV4_FIELD_security_option_length] == 0)
	{
	  hdt_sets[IPV4_FIELD_security_option_length] = hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_security_option_length] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "security_option", "length");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_security_option_length], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_security_option_security)
    {
      if (hdt_sets[IPV4_FIELD_security_option_security] == 0)
	{
	  hdt_sets[IPV4_FIELD_security_option_security] = hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_security_option_security] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "security_option", "security");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_security_option_security], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_security_option_compartments)
    {
      if (hdt_sets[IPV4_FIELD_security_option_compartments] == 0)
	{
	  hdt_sets[IPV4_FIELD_security_option_compartments] =
	    hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_security_option_compartments] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "security_option", "compartments");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_security_option_compartments], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_security_option_handling_restrictions)
    {
      if (hdt_sets[IPV4_FIELD_security_option_handling_restrictions] == 0)
	{
	  hdt_sets[IPV4_FIELD_security_option_handling_restrictions] =
	    hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_security_option_handling_restrictions] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "security_option", "handling_restrictions");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_security_option_handling_restrictions], prid,
	   op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_security_option_unused)
    {
      if (hdt_sets[IPV4_FIELD_security_option_unused] == 0)
	{
	  hdt_sets[IPV4_FIELD_security_option_unused] = hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_security_option_unused] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "security_option", "unused");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_security_option_unused], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_security_option_tcc_field)
    {
      if (hdt_sets[IPV4_FIELD_security_option_tcc_field] == 0)
	{
	  hdt_sets[IPV4_FIELD_security_option_tcc_field] =
	    hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_security_option_tcc_field] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "security_option", "tcc_field");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_security_option_tcc_field], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }

  if (field == IPV4_FIELD_loose_source_and_record_route_option_type)
    {
      if (hdt_sets[IPV4_FIELD_loose_source_and_record_route_option_type] == 0)
	{
	  hdt_sets[IPV4_FIELD_loose_source_and_record_route_option_type] =
	    hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_loose_source_and_record_route_option_type]
	      == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "loose_source_and_record_route_option", "type");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_loose_source_and_record_route_option_type],
	   prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_loose_source_and_record_route_option_length)
    {
      if (hdt_sets[IPV4_FIELD_loose_source_and_record_route_option_length] ==
	  0)
	{
	  hdt_sets[IPV4_FIELD_loose_source_and_record_route_option_length] =
	    hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_loose_source_and_record_route_option_length]
	      == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "loose_source_and_record_route_option", "length");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_loose_source_and_record_route_option_length],
	   prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_loose_source_and_record_route_option_pointer)
    {
      if (hdt_sets[IPV4_FIELD_loose_source_and_record_route_option_pointer] ==
	  0)
	{
	  hdt_sets[IPV4_FIELD_loose_source_and_record_route_option_pointer] =
	    hdt_int32_create ();
	  if (hdt_sets
	      [IPV4_FIELD_loose_source_and_record_route_option_pointer] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "loose_source_and_record_route_option", "pointer");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_loose_source_and_record_route_option_pointer],
	   prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }

  if (field == IPV4_FIELD_strict_source_and_record_route_option_type)
    {
      if (hdt_sets[IPV4_FIELD_strict_source_and_record_route_option_type] ==
	  0)
	{
	  hdt_sets[IPV4_FIELD_strict_source_and_record_route_option_type] =
	    hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_strict_source_and_record_route_option_type]
	      == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "strict_source_and_record_route_option", "type");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_strict_source_and_record_route_option_type],
	   prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_strict_source_and_record_route_option_length)
    {
      if (hdt_sets[IPV4_FIELD_strict_source_and_record_route_option_length] ==
	  0)
	{
	  hdt_sets[IPV4_FIELD_strict_source_and_record_route_option_length] =
	    hdt_int32_create ();
	  if (hdt_sets
	      [IPV4_FIELD_strict_source_and_record_route_option_length] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "strict_source_and_record_route_option", "length");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_strict_source_and_record_route_option_length],
	   prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_strict_source_and_record_route_option_pointer)
    {
      if (hdt_sets[IPV4_FIELD_strict_source_and_record_route_option_pointer]
	  == 0)
	{
	  hdt_sets[IPV4_FIELD_strict_source_and_record_route_option_pointer] =
	    hdt_int32_create ();
	  if (hdt_sets
	      [IPV4_FIELD_strict_source_and_record_route_option_pointer] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "strict_source_and_record_route_option", "pointer");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_strict_source_and_record_route_option_pointer],
	   prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }

  if (field == IPV4_FIELD_record_route_option_type)
    {
      if (hdt_sets[IPV4_FIELD_record_route_option_type] == 0)
	{
	  hdt_sets[IPV4_FIELD_record_route_option_type] = hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_record_route_option_type] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "record_route_option", "type");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_record_route_option_type], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_record_route_option_length)
    {
      if (hdt_sets[IPV4_FIELD_record_route_option_length] == 0)
	{
	  hdt_sets[IPV4_FIELD_record_route_option_length] =
	    hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_record_route_option_length] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "record_route_option", "length");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_record_route_option_length], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_record_route_option_pointer)
    {
      if (hdt_sets[IPV4_FIELD_record_route_option_pointer] == 0)
	{
	  hdt_sets[IPV4_FIELD_record_route_option_pointer] =
	    hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_record_route_option_pointer] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "record_route_option", "pointer");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_record_route_option_pointer], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }

  if (field == IPV4_FIELD_internet_timestamp_option_type)
    {
      if (hdt_sets[IPV4_FIELD_internet_timestamp_option_type] == 0)
	{
	  hdt_sets[IPV4_FIELD_internet_timestamp_option_type] =
	    hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_internet_timestamp_option_type] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "internet_timestamp_option", "type");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_internet_timestamp_option_type], prid, op,
	   arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_internet_timestamp_option_length)
    {
      if (hdt_sets[IPV4_FIELD_internet_timestamp_option_length] == 0)
	{
	  hdt_sets[IPV4_FIELD_internet_timestamp_option_length] =
	    hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_internet_timestamp_option_length] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "internet_timestamp_option", "length");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_internet_timestamp_option_length], prid, op,
	   arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_internet_timestamp_option_flag)
    {
      if (hdt_sets[IPV4_FIELD_internet_timestamp_option_flag] == 0)
	{
	  hdt_sets[IPV4_FIELD_internet_timestamp_option_flag] =
	    hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_internet_timestamp_option_flag] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "internet_timestamp_option", "flag");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_internet_timestamp_option_flag], prid, op,
	   arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_internet_timestamp_option_overflow)
    {
      if (hdt_sets[IPV4_FIELD_internet_timestamp_option_overflow] == 0)
	{
	  hdt_sets[IPV4_FIELD_internet_timestamp_option_overflow] =
	    hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_internet_timestamp_option_overflow] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "internet_timestamp_option", "overflow");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_internet_timestamp_option_overflow], prid, op,
	   arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_internet_timestamp_option_pointer)
    {
      if (hdt_sets[IPV4_FIELD_internet_timestamp_option_pointer] == 0)
	{
	  hdt_sets[IPV4_FIELD_internet_timestamp_option_pointer] =
	    hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_internet_timestamp_option_pointer] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "internet_timestamp_option", "pointer");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_internet_timestamp_option_pointer], prid, op,
	   arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }

  if (field == IPV4_FIELD_stream_identifier_option_type)
    {
      if (hdt_sets[IPV4_FIELD_stream_identifier_option_type] == 0)
	{
	  hdt_sets[IPV4_FIELD_stream_identifier_option_type] =
	    hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_stream_identifier_option_type] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "stream_identifier_option", "type");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_stream_identifier_option_type], prid, op, arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_stream_identifier_option_length)
    {
      if (hdt_sets[IPV4_FIELD_stream_identifier_option_length] == 0)
	{
	  hdt_sets[IPV4_FIELD_stream_identifier_option_length] =
	    hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_stream_identifier_option_length] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "stream_identifier_option", "length");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_stream_identifier_option_length], prid, op,
	   arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == IPV4_FIELD_stream_identifier_option_stream_id)
    {
      if (hdt_sets[IPV4_FIELD_stream_identifier_option_stream_id] == 0)
	{
	  hdt_sets[IPV4_FIELD_stream_identifier_option_stream_id] =
	    hdt_int32_create ();
	  if (hdt_sets[IPV4_FIELD_stream_identifier_option_stream_id] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "stream_identifier_option", "stream_id");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[IPV4_FIELD_stream_identifier_option_stream_id], prid, op,
	   arg))
	{
	  printf ("Error adding primitive %s in ipv4!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }


  return (-1);
}

static int
ipv4_stock_unreg_prim (prim_id i)
{

  return (-1);
}

static int
ipv4_stock_rewrite (struct capsule *c, void *new_payload, u_int32_t len)
{
  if ((c != NULL) && (new_payload != NULL))
    len++;
  return (-1);
}

static char *
ipv4_stock_report_field (struct capsule *cp, int field,
			 struct memcon_context *m)
{
  u_int32_t t;
  char *r, buf[1024];
  struct section *sec, *sec2;

  buf[0] = 0;

  switch (field)
    {


    case (IPV4_FIELD_Hdr_header_length):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_Hdr *) (sec->content))->header_length);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (IPV4_FIELD_Hdr_version):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_Hdr *) (sec->content))->version);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (IPV4_FIELD_Hdr_tos):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_Hdr *) (sec->content))->tos);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (IPV4_FIELD_Hdr_length):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct IPV4_Hdr *) (sec->content))->length));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (IPV4_FIELD_Hdr_id):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct IPV4_Hdr *) (sec->content))->id));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (IPV4_FIELD_Hdr_flags_offset):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct IPV4_Hdr *) (sec->content))->
			   flags_offset));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (IPV4_FIELD_Hdr_ttl):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_Hdr *) (sec->content))->ttl);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (IPV4_FIELD_Hdr_protocol):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_Hdr *) (sec->content))->protocol);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (IPV4_FIELD_Hdr_checksum):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct IPV4_Hdr *) (sec->content))->checksum));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (IPV4_FIELD_Hdr_source):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  return (hdt_ipv4_addr_report
		  (((struct IPV4_Hdr *) (sec->content))->source, m));
	}
      snprintf (buf, sizeof (buf), "(section not present)");
      break;
    case (IPV4_FIELD_Hdr_destination):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  return (hdt_ipv4_addr_report
		  (((struct IPV4_Hdr *) (sec->content))->destination, m));
	}
      snprintf (buf, sizeof (buf), "(section not present)");
      break;


    case (IPV4_FIELD_flags_offset_flags_offset):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_flags_offset)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct IPV4_flags_offset *) (sec->content))->
			   flags_offset));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "flags_offset");
	}
      break;

    case (IPV4_FIELD_class_one_option_option):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_class_one_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_class_one_option *) (sec->content))->
		    option);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "class_one_option");
	}
      break;

    case (IPV4_FIELD_security_option_type):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_security_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_security_option *) (sec->content))->type);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "security_option");
	}
      break;
    case (IPV4_FIELD_security_option_length):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_security_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_security_option *) (sec->content))->length);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "security_option");
	}
      break;
    case (IPV4_FIELD_security_option_security):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_security_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct IPV4_security_option *) (sec->content))->
			   security));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "security_option");
	}
      break;
    case (IPV4_FIELD_security_option_compartments):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_security_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct IPV4_security_option *) (sec->content))->
			   compartments));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "security_option");
	}
      break;
    case (IPV4_FIELD_security_option_handling_restrictions):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_security_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct IPV4_security_option *) (sec->content))->
			   handling_restrictions));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "security_option");
	}
      break;
    case (IPV4_FIELD_security_option_unused):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_security_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_security_option *) (sec->content))->unused);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "security_option");
	}
      break;
    case (IPV4_FIELD_security_option_tcc_field):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_security_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct IPV4_security_option *) (sec->content))->
			   tcc_field));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "security_option");
	}
      break;

    case (IPV4_FIELD_loose_source_and_record_route_option_type):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id ==
	      IPV4_SECTION_loose_source_and_record_route_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_loose_source_and_record_route_option
		      *) (sec->content))->type);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "loose_source_and_record_route_option");
	}
      break;
    case (IPV4_FIELD_loose_source_and_record_route_option_length):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id ==
	      IPV4_SECTION_loose_source_and_record_route_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_loose_source_and_record_route_option
		      *) (sec->content))->length);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "loose_source_and_record_route_option");
	}
      break;
    case (IPV4_FIELD_loose_source_and_record_route_option_pointer):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id ==
	      IPV4_SECTION_loose_source_and_record_route_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_loose_source_and_record_route_option
		      *) (sec->content))->pointer);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "loose_source_and_record_route_option");
	}
      break;

    case (IPV4_FIELD_strict_source_and_record_route_option_type):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id ==
	      IPV4_SECTION_strict_source_and_record_route_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_strict_source_and_record_route_option
		      *) (sec->content))->type);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "strict_source_and_record_route_option");
	}
      break;
    case (IPV4_FIELD_strict_source_and_record_route_option_length):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id ==
	      IPV4_SECTION_strict_source_and_record_route_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_strict_source_and_record_route_option
		      *) (sec->content))->length);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "strict_source_and_record_route_option");
	}
      break;
    case (IPV4_FIELD_strict_source_and_record_route_option_pointer):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id ==
	      IPV4_SECTION_strict_source_and_record_route_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_strict_source_and_record_route_option
		      *) (sec->content))->pointer);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "strict_source_and_record_route_option");
	}
      break;

    case (IPV4_FIELD_record_route_option_type):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_record_route_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_record_route_option *) (sec->content))->
		    type);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "record_route_option");
	}
      break;
    case (IPV4_FIELD_record_route_option_length):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_record_route_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_record_route_option *) (sec->content))->
		    length);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "record_route_option");
	}
      break;
    case (IPV4_FIELD_record_route_option_pointer):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_record_route_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_record_route_option *) (sec->content))->
		    pointer);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "record_route_option");
	}
      break;

    case (IPV4_FIELD_internet_timestamp_option_type):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_internet_timestamp_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_internet_timestamp_option *) (sec->
								content))->
		    type);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "internet_timestamp_option");
	}
      break;
    case (IPV4_FIELD_internet_timestamp_option_length):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_internet_timestamp_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_internet_timestamp_option *) (sec->
								content))->
		    length);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "internet_timestamp_option");
	}
      break;
    case (IPV4_FIELD_internet_timestamp_option_flag):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_internet_timestamp_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_internet_timestamp_option *) (sec->
								content))->
		    flag);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "internet_timestamp_option");
	}
      break;
    case (IPV4_FIELD_internet_timestamp_option_overflow):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_internet_timestamp_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_internet_timestamp_option *) (sec->
								content))->
		    overflow);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "internet_timestamp_option");
	}
      break;
    case (IPV4_FIELD_internet_timestamp_option_pointer):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_internet_timestamp_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_internet_timestamp_option *) (sec->
								content))->
		    pointer);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "internet_timestamp_option");
	}
      break;

    case (IPV4_FIELD_stream_identifier_option_type):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_stream_identifier_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_stream_identifier_option *) (sec->
							       content))->
		    type);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "stream_identifier_option");
	}
      break;
    case (IPV4_FIELD_stream_identifier_option_length):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_stream_identifier_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ((struct IPV4_stream_identifier_option *) (sec->
							       content))->
		    length);
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "stream_identifier_option");
	}
      break;
    case (IPV4_FIELD_stream_identifier_option_stream_id):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == IPV4_SECTION_stream_identifier_option)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct IPV4_stream_identifier_option *) (sec->
								      content))->
			   stream_id));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "stream_identifier_option");
	}
      break;


    default:

      return (NULL);
    }
  t = (strlen (buf) + 1);
  r = m->eng->alloc (m, t);
  strncpy (r, buf, t);
  return (r);
}

static int
ipv4_stock_find_matches (struct capsule *cp, struct prim_can *p)
{
  struct section *s;
  struct memcon_context *c = cp->parent->c;
  u_int32_t v;
  unsigned long beg, end;

  if (p)
    return (-1);

  if (cp->matches == NULL && gprid)
    {
      if (!(cp->matches = c->eng->alloc (c, gprid + 1)))
	exit (-1);
      memset (cp->matches, 0, gprid + 1);

      cp->matches[0] = 1;
    }


  for (s = cp->sections; s; s = s->next)
    {
      struct simple_prim *sp;

      for (sp = present_sections.proto_self; sp; sp = sp->next)
	{
	  ;
	  cp->matches[sp->prim] |= 1;
	}


      switch (s->section_id)
	{
	case IPV4_SECTION_INVALID:
	  return (-1);


	  break;
	case IPV4_SECTION_Hdr:;
	  for (sp = present_sections.present_Hdr; sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_Hdr_header_length])
	    {;
	      v = ((struct IPV4_Hdr *) (s->content))->header_length;;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_Hdr_header_length], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "header_length", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_Hdr_version])
	    {;
	      v = ((struct IPV4_Hdr *) (s->content))->version;;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_Hdr_version], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "version", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_Hdr_tos])
	    {;
	      v = ((struct IPV4_Hdr *) (s->content))->tos;;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_Hdr_tos], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr", "tos",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_Hdr_length])
	    {;
	      v = ntohs (((struct IPV4_Hdr *) (s->content))->length);;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_Hdr_length], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "length", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_Hdr_id])
	    {;
	      v = ntohs (((struct IPV4_Hdr *) (s->content))->id);;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_Hdr_id], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr", "id",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_Hdr_flags_offset])
	    {;
	      v = ntohs (((struct IPV4_Hdr *) (s->content))->flags_offset);;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_Hdr_flags_offset], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "flags_offset", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_Hdr_ttl])
	    {;
	      v = ((struct IPV4_Hdr *) (s->content))->ttl;;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_Hdr_ttl], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr", "ttl",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_Hdr_protocol])
	    {;
	      v = ((struct IPV4_Hdr *) (s->content))->protocol;;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_Hdr_protocol], NULL, v, cp->matches,
		   1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "protocol", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_Hdr_checksum])
	    {;
	      v = ntohs (((struct IPV4_Hdr *) (s->content))->checksum);;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_Hdr_checksum], NULL, v, cp->matches,
		   1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "checksum", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_Hdr_source])
	    {;
	      if (hdt_ipv4_addr_match
		  (hdt_sets[IPV4_FIELD_Hdr_source], NULL,
		   ((struct IPV4_Hdr *) (s->content))->source, cp->matches,
		   1))
		{
		  printf ("AIEE!  Can't find %s match.  Exiting.\n",
			  "ipv4_addr");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "source", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_Hdr_destination])
	    {;
	      if (hdt_ipv4_addr_match
		  (hdt_sets[IPV4_FIELD_Hdr_destination], NULL,
		   ((struct IPV4_Hdr *) (s->content))->destination,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find %s match.  Exiting.\n",
			  "ipv4_addr");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "destination", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_Hdr_payload])
	    {
	      struct blob bb;
	      bb.blob = &(((struct IPV4_Hdr *) (s->content))->payload);
	      bb.len = cp->length - offsetof (struct IPV4_Hdr, payload);;
	      if (hdt_blob_match
		  (hdt_sets[IPV4_FIELD_Hdr_payload], NULL, &bb, cp->matches,
		   1))
		{
		  printf ("AIEE!  Can't find blob match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "payload", (end - beg));
	  break;
	case IPV4_SECTION_flags_offset:;
	  for (sp = present_sections.present_flags_offset; sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_flags_offset_flags_offset])
	    {;
	      v =
		ntohs (((struct IPV4_flags_offset *) (s->content))->
		       flags_offset);;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_flags_offset_flags_offset], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "flags_offset", "flags_offset", (end - beg));
	  break;
	case IPV4_SECTION_class_one_option:;
	  for (sp = present_sections.present_class_one_option; sp;
	       sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_class_one_option_option])
	    {;
	      v = ((struct IPV4_class_one_option *) (s->content))->option;;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_class_one_option_option], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "class_one_option", "option", (end - beg));
	  break;
	case IPV4_SECTION_security_option:;
	  for (sp = present_sections.present_security_option; sp;
	       sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_security_option_type])
	    {;
	      v = ((struct IPV4_security_option *) (s->content))->type;;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_security_option_type], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "security_option", "type", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_security_option_length])
	    {;
	      v = ((struct IPV4_security_option *) (s->content))->length;;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_security_option_length], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "security_option", "length", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_security_option_security])
	    {;
	      v =
		ntohs (((struct IPV4_security_option *) (s->content))->
		       security);;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_security_option_security], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "security_option", "security", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_security_option_compartments])
	    {;
	      v =
		ntohs (((struct IPV4_security_option *) (s->content))->
		       compartments);;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_security_option_compartments], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "security_option", "compartments", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_security_option_handling_restrictions])
	    {;
	      v =
		ntohs (((struct IPV4_security_option *) (s->content))->
		       handling_restrictions);;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_security_option_handling_restrictions],
		   NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "security_option", "handling_restrictions", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_security_option_unused])
	    {;
	      v = ((struct IPV4_security_option *) (s->content))->unused;;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_security_option_unused], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "security_option", "unused", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_security_option_tcc_field])
	    {;
	      v =
		ntohl (((struct IPV4_security_option *) (s->content))->
		       tcc_field);;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_security_option_tcc_field], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "security_option", "tcc_field", (end - beg));
	  break;
	case IPV4_SECTION_loose_source_and_record_route_option:;
	  for (sp =
	       present_sections.present_loose_source_and_record_route_option;
	       sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_loose_source_and_record_route_option_type])
	    {;
	      v =
		((struct IPV4_loose_source_and_record_route_option *) (s->
								       content))->
		type;;
	      if (hdt_int32_match
		  (hdt_sets
		   [IPV4_FIELD_loose_source_and_record_route_option_type],
		   NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "loose_source_and_record_route_option", "type",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets
	      [IPV4_FIELD_loose_source_and_record_route_option_length])
	    {;
	      v =
		((struct IPV4_loose_source_and_record_route_option *) (s->
								       content))->
		length;;
	      if (hdt_int32_match
		  (hdt_sets
		   [IPV4_FIELD_loose_source_and_record_route_option_length],
		   NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "loose_source_and_record_route_option", "length",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets
	      [IPV4_FIELD_loose_source_and_record_route_option_pointer])
	    {;
	      v =
		((struct IPV4_loose_source_and_record_route_option *) (s->
								       content))->
		pointer;;
	      if (hdt_int32_match
		  (hdt_sets
		   [IPV4_FIELD_loose_source_and_record_route_option_pointer],
		   NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "loose_source_and_record_route_option", "pointer",
		  (end - beg));
	  break;
	case IPV4_SECTION_strict_source_and_record_route_option:;
	  for (sp =
	       present_sections.present_strict_source_and_record_route_option;
	       sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_strict_source_and_record_route_option_type])
	    {;
	      v =
		((struct IPV4_strict_source_and_record_route_option *) (s->
									content))->
		type;;
	      if (hdt_int32_match
		  (hdt_sets
		   [IPV4_FIELD_strict_source_and_record_route_option_type],
		   NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "strict_source_and_record_route_option", "type",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets
	      [IPV4_FIELD_strict_source_and_record_route_option_length])
	    {;
	      v =
		((struct IPV4_strict_source_and_record_route_option *) (s->
									content))->
		length;;
	      if (hdt_int32_match
		  (hdt_sets
		   [IPV4_FIELD_strict_source_and_record_route_option_length],
		   NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "strict_source_and_record_route_option", "length",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets
	      [IPV4_FIELD_strict_source_and_record_route_option_pointer])
	    {;
	      v =
		((struct IPV4_strict_source_and_record_route_option *) (s->
									content))->
		pointer;;
	      if (hdt_int32_match
		  (hdt_sets
		   [IPV4_FIELD_strict_source_and_record_route_option_pointer],
		   NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "strict_source_and_record_route_option", "pointer",
		  (end - beg));
	  break;
	case IPV4_SECTION_record_route_option:;
	  for (sp = present_sections.present_record_route_option; sp;
	       sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_record_route_option_type])
	    {;
	      v = ((struct IPV4_record_route_option *) (s->content))->type;;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_record_route_option_type], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "record_route_option", "type", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_record_route_option_length])
	    {;
	      v = ((struct IPV4_record_route_option *) (s->content))->length;;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_record_route_option_length], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "record_route_option", "length", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_record_route_option_pointer])
	    {;
	      v =
		((struct IPV4_record_route_option *) (s->content))->pointer;;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_record_route_option_pointer], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "record_route_option", "pointer", (end - beg));
	  break;
	case IPV4_SECTION_internet_timestamp_option:;
	  for (sp = present_sections.present_internet_timestamp_option; sp;
	       sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_internet_timestamp_option_type])
	    {;
	      v =
		((struct IPV4_internet_timestamp_option *) (s->content))->
		type;;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_internet_timestamp_option_type], NULL,
		   v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "internet_timestamp_option", "type", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_internet_timestamp_option_length])
	    {;
	      v =
		((struct IPV4_internet_timestamp_option *) (s->content))->
		length;;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_internet_timestamp_option_length],
		   NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "internet_timestamp_option", "length", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_internet_timestamp_option_flag])
	    {;
	      v =
		((struct IPV4_internet_timestamp_option *) (s->content))->
		flag;;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_internet_timestamp_option_flag], NULL,
		   v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "internet_timestamp_option", "flag", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_internet_timestamp_option_overflow])
	    {;
	      v =
		((struct IPV4_internet_timestamp_option *) (s->content))->
		overflow;;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_internet_timestamp_option_overflow],
		   NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "internet_timestamp_option", "overflow", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_internet_timestamp_option_pointer])
	    {;
	      v =
		((struct IPV4_internet_timestamp_option *) (s->content))->
		pointer;;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_internet_timestamp_option_pointer],
		   NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "internet_timestamp_option", "pointer", (end - beg));
	  break;
	case IPV4_SECTION_stream_identifier_option:;
	  for (sp = present_sections.present_stream_identifier_option; sp;
	       sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_stream_identifier_option_type])
	    {;
	      v =
		((struct IPV4_stream_identifier_option *) (s->content))->
		type;;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_stream_identifier_option_type], NULL,
		   v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "stream_identifier_option", "type", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_stream_identifier_option_length])
	    {;
	      v =
		((struct IPV4_stream_identifier_option *) (s->content))->
		length;;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_stream_identifier_option_length], NULL,
		   v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "stream_identifier_option", "length", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[IPV4_FIELD_stream_identifier_option_stream_id])
	    {;
	      v =
		ntohs (((struct IPV4_stream_identifier_option *) (s->
								  content))->
		       stream_id);;
	      if (hdt_int32_match
		  (hdt_sets[IPV4_FIELD_stream_identifier_option_stream_id],
		   NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n",
		  "stream_identifier_option", "stream_id", (end - beg));


	  break;
	default:
	  printf ("Unrecognized section: %i (not %i).\n", s->section_id,
		  IPV4_SECTION_Hdr);
	  break;
	}
    }

  return (-1);
}
