/* Copyright 2001, Todd Lewis <tlewis@mindspring.com>
 * All rights reserved. */

#include "hank.h"
#include "hdt_blob.h"
#include "hdt_int32.h"
#include "packet.h"
#include "protocol_engine.h"
#include "protocols.h"
#include "support.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.h"

#include "oncrpc.hi"


static char rcsid[] =
  "$Id: oncrpc.xml,v 1.2 2001/10/29 06:23:30 tlewis Exp $";


extern struct smodule *oncrpc[];


extern struct protocol_engine oncrpc_pe;


static u_int32_t protocols[] = {

  ((HPN_UDP << 24) | 111),


  0
};

struct simple_prim
{
  prim_id prim;
  struct simple_prim *next;
};

static struct
{
  struct simple_prim *proto_self;


  struct simple_prim *present_Hdr;


  struct simple_prim *present_call_body;


  struct simple_prim *present_reply_body;


}
present_sections;

enum ONCRPC_FIELD
{
  ONCRPC_SELF,


  ONCRPC_FIELD_Hdr,
  ONCRPC_FIELD_Hdr_xid,
  ONCRPC_FIELD_Hdr_direction,
  ONCRPC_FIELD_Hdr_payload,
  ONCRPC_FIELD_call_body,
  ONCRPC_FIELD_call_body_rpcvers,
  ONCRPC_FIELD_call_body_prog,
  ONCRPC_FIELD_call_body_vers,
  ONCRPC_FIELD_call_body_proc,
  ONCRPC_FIELD_reply_body,
  ONCRPC_FIELD_reply_body_rp_stat,


  ONCRPC_FIELD_INVALID
};

enum ONCRPC_SECTIONS
{


  ONCRPC_SECTION_Hdr,


  ONCRPC_SECTION_call_body,


  ONCRPC_SECTION_reply_body,


  ONCRPC_SECTION_INVALID
};

struct ONCRPC_null
{
  int empty;


};
struct ONCRPC_Hdr
{
  u_int32_t xid;
  u_int32_t direction;
  unsigned char payload;
};
struct ONCRPC_call_body
{
  u_int32_t rpcvers;
  u_int32_t prog;
  u_int32_t vers;
  u_int32_t proc;
};
struct ONCRPC_reply_body
{
  u_int32_t rp_stat;


};


static char hexes[] = "0123456789abcdef";

static int gprid = 0;

static hdt_set hdt_sets[ONCRPC_FIELD_INVALID];


static int oncrpc_stock_reg_prim (int field, char *op, char *arg);
static int oncrpc_stock_unreg_prim (prim_id i);
static int oncrpc_stock_decompose (struct packet *p);
static int oncrpc_stock_find_matches (struct capsule *cp, struct prim_can *c);
static int oncrpc_stock_rewrite (struct capsule *c, void *new_payload,
				 u_int32_t len);
static char *oncrpc_stock_report_field (struct capsule *p, int field,
					struct memcon_context *m);
static int oncrpc_field_lookup (char *section, char *field);

struct protocol_engine oncrpc_pe = {
  {
   "pe",
   "oncrpc",
   PEVMIN,
   PEVMAJ,
   &oncrpc_pe},
  &oncrpc_private_decompose,
  &oncrpc_stock_reg_prim,
  &oncrpc_stock_unreg_prim,
  &oncrpc_stock_find_matches,
  &oncrpc_stock_rewrite,
  &oncrpc_stock_report_field,
  &oncrpc_field_lookup,
  &protocols[0],
  0
};


struct smodule *oncrpc[] = {

  &(oncrpc_pe.s),
  NULL
};

static int
oncrpc_private_decompose (struct packet *p)
{
  struct capsule *e = p->encaps, *f;
  struct ONCRPC_Hdr *h;
  struct section *s, *t;
  int tcpudp = 0;

  while (e->next)
    {
      if (e->protocol == ((HPN_IPV4 << 24) | 6))
	tcpudp = 1;
      if (e->protocol == ((HPN_IPV4 << 24) | 17))
	tcpudp = 2;

      e = e->next;
    }


  if (tcpudp == 1)
    h = (struct ONCRPC_Hdr *) &e->payload[4];
  else
    h = (struct ONCRPC_Hdr *) e->payload;

  printf ("ONCRPC: xid %i, direction %i.\n",
	  htonl (h->xid), htonl (h->direction));

  append_encap (p, protocols[0], e->payload_length, e->payload);
  f = e->next;
  f->pe = &oncrpc_pe;

  if ((s = p->c->eng->alloc (p->c, sizeof (*s))) == NULL)
    exit (-1);

  if ((t = p->c->eng->alloc (p->c, sizeof (*t))) == NULL)
    exit (-1);

  s->section_id = ONCRPC_SECTION_Hdr;
  s->content = h;
  s->next = t;

  switch (htonl (h->direction))
    {

    case 0:
      t->section_id = ONCRPC_SECTION_call_body;
      break;

    case 1:
      t->section_id = ONCRPC_SECTION_reply_body;
      break;

    default:
      printf ("AIEE!  Unrecognized oncrpc direction: %i\n",
	      htonl (h->direction));
    }

  t->content = &(h->payload);
  t->next = NULL;

  f->sections = s;

  return (0);
}


static int
oncrpc_field_lookup (char *section, char *field)
{
  int p = -1;

  if ((section == NULL) && (field == NULL))
    {
      p = ONCRPC_SELF;
    }
  else if (section != NULL && !(strcmp (section, "INVALID")))
    {
      p = ONCRPC_FIELD_INVALID;


    }
  else if ((section == NULL) || (!(strcmp (section, "Hdr"))))
    {
      if (field == NULL)
	p = ONCRPC_FIELD_Hdr;
      else if (!(strcmp (field, "xid")))
	p = ONCRPC_FIELD_Hdr_xid;
      else if (!(strcmp (field, "direction")))
	p = ONCRPC_FIELD_Hdr_direction;
      else if (!(strcmp (field, "payload")))
	p = ONCRPC_FIELD_Hdr_payload;
    }
  else if ((section == NULL) || (!(strcmp (section, "call_body"))))
    {
      if (field == NULL)
	p = ONCRPC_FIELD_call_body;
      else if (!(strcmp (field, "rpcvers")))
	p = ONCRPC_FIELD_call_body_rpcvers;
      else if (!(strcmp (field, "prog")))
	p = ONCRPC_FIELD_call_body_prog;
      else if (!(strcmp (field, "vers")))
	p = ONCRPC_FIELD_call_body_vers;
      else if (!(strcmp (field, "proc")))
	p = ONCRPC_FIELD_call_body_proc;
    }
  else if ((section == NULL) || (!(strcmp (section, "reply_body"))))
    {
      if (field == NULL)
	p = ONCRPC_FIELD_reply_body;
      else if (!(strcmp (field, "rp_stat")))
	p = ONCRPC_FIELD_reply_body_rp_stat;


    }

  if (p == -1)
    {
      printf ("ONCRPC: Unrecognized field: \" %s \".  Returning error.\n",
	      field);
      return (ONCRPC_FIELD_INVALID);
    }

  ;
  return (p);
}


static int
oncrpc_stock_decompose (struct packet *p)
{
  struct capsule *e = p->encaps, *f;
  struct ONCRPC_Hdr *h;
  struct section *s;


  while (e->next)
    e = e->next;

  h = (struct ONCRPC_Hdr *) e->payload;


  append_encap (p, protocols[0], e->payload_length, e->payload);
  f = e->next;
  f->pe = &oncrpc_pe;

  if ((s = p->c->eng->alloc (p->c, sizeof (*s))) == NULL)
    exit (-1);
  s->section_id = ONCRPC_SECTION_Hdr;
  s->content = h;
  s->next = NULL;
  f->sections = s;
  f->length = e->payload_length;


  return (0);
}

static int
oncrpc_stock_reg_prim (int field, char *op, char *arg)
{
  int prid = ++gprid;


  ;

  if (field == ONCRPC_FIELD_INVALID)
    {
      printf
	("ONCRPC: Error.  Received invalid field in reg_prim(), op=%s, arg=%s.\n",
	 op, arg);
      return (-1);
    }


  if (!strcmp (op, "present"))
    {
      struct simple_prim *p;

      if (!(p = malloc (sizeof (*p))))
	exit (-1);
      p->prim = prid;

      if (field == ONCRPC_SELF)
	{

	  p->next = present_sections.proto_self;
	  present_sections.proto_self = p;
	  return (prid);


	}
      else if (field == ONCRPC_FIELD_Hdr)
	{
	  p->next = present_sections.present_Hdr;
	  present_sections.present_Hdr = p;
	  return (prid);


	}
      else if (field == ONCRPC_FIELD_call_body)
	{
	  p->next = present_sections.present_call_body;
	  present_sections.present_call_body = p;
	  return (prid);


	}
      else if (field == ONCRPC_FIELD_reply_body)
	{
	  p->next = present_sections.present_reply_body;
	  present_sections.present_reply_body = p;
	  return (prid);


	}

      printf ("That shit (%i) ain't keepin' it real.\n", field);
      return (-1);
    }


  if (field == ONCRPC_FIELD_Hdr_xid)
    {
      if (hdt_sets[ONCRPC_FIELD_Hdr_xid] == 0)
	{
	  hdt_sets[ONCRPC_FIELD_Hdr_xid] = hdt_int32_create ();
	  if (hdt_sets[ONCRPC_FIELD_Hdr_xid] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "xid");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[ONCRPC_FIELD_Hdr_xid], prid, op, arg))
	{
	  printf ("Error adding primitive %s in oncrpc!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ONCRPC_FIELD_Hdr_direction)
    {
      if (hdt_sets[ONCRPC_FIELD_Hdr_direction] == 0)
	{
	  hdt_sets[ONCRPC_FIELD_Hdr_direction] = hdt_int32_create ();
	  if (hdt_sets[ONCRPC_FIELD_Hdr_direction] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "direction");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[ONCRPC_FIELD_Hdr_direction], prid, op, arg))
	{
	  printf ("Error adding primitive %s in oncrpc!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ONCRPC_FIELD_Hdr_payload)
    {
      if (hdt_sets[ONCRPC_FIELD_Hdr_payload] == 0)
	{
	  hdt_sets[ONCRPC_FIELD_Hdr_payload] = hdt_blob_create ();;
	  if (hdt_sets[ONCRPC_FIELD_Hdr_payload] == 0)
	    {
	      printf ("AIEE!  Can't add match set for %s.%s!  Exiting.\n",
		      "Hdr", "payload");
	      exit (-1);
	    }
	}
      if (hdt_blob_add (hdt_sets[ONCRPC_FIELD_Hdr_payload], prid, op, arg))
	{
	  printf ("Error adding primitive %s in oncrpc!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }

  if (field == ONCRPC_FIELD_call_body_rpcvers)
    {
      if (hdt_sets[ONCRPC_FIELD_call_body_rpcvers] == 0)
	{
	  hdt_sets[ONCRPC_FIELD_call_body_rpcvers] = hdt_int32_create ();
	  if (hdt_sets[ONCRPC_FIELD_call_body_rpcvers] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "call_body", "rpcvers");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ONCRPC_FIELD_call_body_rpcvers], prid, op, arg))
	{
	  printf ("Error adding primitive %s in oncrpc!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ONCRPC_FIELD_call_body_prog)
    {
      if (hdt_sets[ONCRPC_FIELD_call_body_prog] == 0)
	{
	  hdt_sets[ONCRPC_FIELD_call_body_prog] = hdt_int32_create ();
	  if (hdt_sets[ONCRPC_FIELD_call_body_prog] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "call_body", "prog");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ONCRPC_FIELD_call_body_prog], prid, op, arg))
	{
	  printf ("Error adding primitive %s in oncrpc!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ONCRPC_FIELD_call_body_vers)
    {
      if (hdt_sets[ONCRPC_FIELD_call_body_vers] == 0)
	{
	  hdt_sets[ONCRPC_FIELD_call_body_vers] = hdt_int32_create ();
	  if (hdt_sets[ONCRPC_FIELD_call_body_vers] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "call_body", "vers");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ONCRPC_FIELD_call_body_vers], prid, op, arg))
	{
	  printf ("Error adding primitive %s in oncrpc!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == ONCRPC_FIELD_call_body_proc)
    {
      if (hdt_sets[ONCRPC_FIELD_call_body_proc] == 0)
	{
	  hdt_sets[ONCRPC_FIELD_call_body_proc] = hdt_int32_create ();
	  if (hdt_sets[ONCRPC_FIELD_call_body_proc] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "call_body", "proc");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ONCRPC_FIELD_call_body_proc], prid, op, arg))
	{
	  printf ("Error adding primitive %s in oncrpc!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }

  if (field == ONCRPC_FIELD_reply_body_rp_stat)
    {
      if (hdt_sets[ONCRPC_FIELD_reply_body_rp_stat] == 0)
	{
	  hdt_sets[ONCRPC_FIELD_reply_body_rp_stat] = hdt_int32_create ();
	  if (hdt_sets[ONCRPC_FIELD_reply_body_rp_stat] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "reply_body", "rp_stat");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[ONCRPC_FIELD_reply_body_rp_stat], prid, op, arg))
	{
	  printf ("Error adding primitive %s in oncrpc!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }


  return (-1);
}

static int
oncrpc_stock_unreg_prim (prim_id i)
{

  return (-1);
}

static int
oncrpc_stock_rewrite (struct capsule *c, void *new_payload, u_int32_t len)
{
  if ((c != NULL) && (new_payload != NULL))
    len++;
  return (-1);
}

static char *
oncrpc_stock_report_field (struct capsule *cp, int field,
			   struct memcon_context *m)
{
  u_int32_t t;
  char *r, buf[1024];
  struct section *sec, *sec2;

  buf[0] = 0;

  switch (field)
    {


    case (ONCRPC_FIELD_Hdr_xid):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ONCRPC_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct ONCRPC_Hdr *) (sec->content))->xid));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (ONCRPC_FIELD_Hdr_direction):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ONCRPC_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct ONCRPC_Hdr *) (sec->content))->
			   direction));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;


    case (ONCRPC_FIELD_call_body_rpcvers):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ONCRPC_SECTION_call_body)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct ONCRPC_call_body *) (sec->content))->
			   rpcvers));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "call_body");
	}
      break;
    case (ONCRPC_FIELD_call_body_prog):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ONCRPC_SECTION_call_body)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct ONCRPC_call_body *) (sec->content))->
			   prog));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "call_body");
	}
      break;
    case (ONCRPC_FIELD_call_body_vers):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ONCRPC_SECTION_call_body)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct ONCRPC_call_body *) (sec->content))->
			   vers));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "call_body");
	}
      break;
    case (ONCRPC_FIELD_call_body_proc):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ONCRPC_SECTION_call_body)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct ONCRPC_call_body *) (sec->content))->
			   proc));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "call_body");
	}
      break;

    case (ONCRPC_FIELD_reply_body_rp_stat):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == ONCRPC_SECTION_reply_body)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohl (((struct ONCRPC_reply_body *) (sec->content))->
			   rp_stat));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)",
		    "reply_body");
	}
      break;


    default:

      return (NULL);
    }
  t = (strlen (buf) + 1);
  r = m->eng->alloc (m, t);
  strncpy (r, buf, t);
  return (r);
}

static int
oncrpc_stock_find_matches (struct capsule *cp, struct prim_can *p)
{
  struct section *s;
  struct memcon_context *c = cp->parent->c;
  u_int32_t v;
  unsigned long beg, end;

  if (p)
    return (-1);

  if (cp->matches == NULL && gprid)
    {
      if (!(cp->matches = c->eng->alloc (c, gprid + 1)))
	exit (-1);
      memset (cp->matches, 0, gprid + 1);

      cp->matches[0] = 1;
    }


  for (s = cp->sections; s; s = s->next)
    {
      struct simple_prim *sp;

      for (sp = present_sections.proto_self; sp; sp = sp->next)
	{
	  ;
	  cp->matches[sp->prim] |= 1;
	}


      switch (s->section_id)
	{
	case ONCRPC_SECTION_INVALID:
	  return (-1);


	  break;
	case ONCRPC_SECTION_Hdr:;
	  for (sp = present_sections.present_Hdr; sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[ONCRPC_FIELD_Hdr_xid])
	    {;
	      v = ntohl (((struct ONCRPC_Hdr *) (s->content))->xid);;
	      if (hdt_int32_match
		  (hdt_sets[ONCRPC_FIELD_Hdr_xid], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr", "xid",
		  (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ONCRPC_FIELD_Hdr_direction])
	    {;
	      v = ntohl (((struct ONCRPC_Hdr *) (s->content))->direction);;
	      if (hdt_int32_match
		  (hdt_sets[ONCRPC_FIELD_Hdr_direction], NULL, v, cp->matches,
		   1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "direction", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ONCRPC_FIELD_Hdr_payload])
	    {
	      struct blob bb;
	      bb.blob = &(((struct ONCRPC_Hdr *) (s->content))->payload);
	      bb.len = cp->length - offsetof (struct ONCRPC_Hdr, payload);;
	      if (hdt_blob_match
		  (hdt_sets[ONCRPC_FIELD_Hdr_payload], NULL, &bb, cp->matches,
		   1))
		{
		  printf ("AIEE!  Can't find blob match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "payload", (end - beg));
	  break;
	case ONCRPC_SECTION_call_body:;
	  for (sp = present_sections.present_call_body; sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[ONCRPC_FIELD_call_body_rpcvers])
	    {;
	      v =
		ntohl (((struct ONCRPC_call_body *) (s->content))->rpcvers);;
	      if (hdt_int32_match
		  (hdt_sets[ONCRPC_FIELD_call_body_rpcvers], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "call_body",
		  "rpcvers", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ONCRPC_FIELD_call_body_prog])
	    {;
	      v = ntohl (((struct ONCRPC_call_body *) (s->content))->prog);;
	      if (hdt_int32_match
		  (hdt_sets[ONCRPC_FIELD_call_body_prog], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "call_body",
		  "prog", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ONCRPC_FIELD_call_body_vers])
	    {;
	      v = ntohl (((struct ONCRPC_call_body *) (s->content))->vers);;
	      if (hdt_int32_match
		  (hdt_sets[ONCRPC_FIELD_call_body_vers], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "call_body",
		  "vers", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[ONCRPC_FIELD_call_body_proc])
	    {;
	      v = ntohl (((struct ONCRPC_call_body *) (s->content))->proc);;
	      if (hdt_int32_match
		  (hdt_sets[ONCRPC_FIELD_call_body_proc], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "call_body",
		  "proc", (end - beg));
	  break;
	case ONCRPC_SECTION_reply_body:;
	  for (sp = present_sections.present_reply_body; sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[ONCRPC_FIELD_reply_body_rp_stat])
	    {;
	      v =
		ntohl (((struct ONCRPC_reply_body *) (s->content))->rp_stat);;
	      if (hdt_int32_match
		  (hdt_sets[ONCRPC_FIELD_reply_body_rp_stat], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "reply_body",
		  "rp_stat", (end - beg));


	  break;
	default:
	  printf ("Unrecognized section: %i (not %i).\n", s->section_id,
		  ONCRPC_SECTION_Hdr);
	  break;
	}
    }

  return (-1);
}
