/* Copyright 2001, Todd Lewis <tlewis@mindspring.com>
 * All rights reserved. */

#include "hank.h"
#include "hdt_blob.h"
#include "hdt_int32.h"
#include "packet.h"
#include "protocol_engine.h"
#include "protocols.h"
#include "support.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.h"

#include "udp.hi"


static char rcsid[] = "$Id: udp.xml,v 1.7 2001/10/29 06:23:30 tlewis Exp $";


extern struct smodule *udp[];


extern struct protocol_engine udp_pe;


static u_int32_t protocols[] = {

  ((HPN_IPV4 << 24) | 17),
  ((HPN_IPV6 << 24) | 17),


  0
};

struct simple_prim
{
  prim_id prim;
  struct simple_prim *next;
};

static struct
{
  struct simple_prim *proto_self;


  struct simple_prim *present_Hdr;


}
present_sections;

enum UDP_FIELD
{
  UDP_SELF,


  UDP_FIELD_Hdr,
  UDP_FIELD_Hdr_source_port,
  UDP_FIELD_Hdr_destination_port,
  UDP_FIELD_Hdr_length,
  UDP_FIELD_Hdr_checksum,
  UDP_FIELD_Hdr_payload,


  UDP_FIELD_INVALID
};

enum UDP_SECTIONS
{


  UDP_SECTION_Hdr,


  UDP_SECTION_INVALID
};

struct UDP_null
{
  int empty;


};
struct UDP_Hdr
{
  u_int16_t source_port;
  u_int16_t destination_port;
  u_int16_t length;
  u_int16_t checksum;
  unsigned char payload;


};


static char hexes[] = "0123456789abcdef";

static int gprid = 0;

static hdt_set hdt_sets[UDP_FIELD_INVALID];


static int udp_stock_reg_prim (int field, char *op, char *arg);
static int udp_stock_unreg_prim (prim_id i);
static int udp_stock_decompose (struct packet *p);
static int udp_stock_find_matches (struct capsule *cp, struct prim_can *c);
static int udp_stock_rewrite (struct capsule *c, void *new_payload,
			      u_int32_t len);
static char *udp_stock_report_field (struct capsule *p, int field,
				     struct memcon_context *m);
static int udp_field_lookup (char *section, char *field);

struct protocol_engine udp_pe = {
  {
   "pe",
   "udp",
   PEVMIN,
   PEVMAJ,
   &udp_pe},
  &udp_stock_decompose,
  &udp_stock_reg_prim,
  &udp_stock_unreg_prim,
  &udp_stock_find_matches,
  &udp_stock_rewrite,
  &udp_stock_report_field,
  &udp_field_lookup,
  &protocols[0],
  0
};


struct smodule *udp[] = {

  &(udp_pe.s),
  NULL
};


static int
udp_field_lookup (char *section, char *field)
{
  int p = -1;

  if ((section == NULL) && (field == NULL))
    {
      p = UDP_SELF;
    }
  else if (section != NULL && !(strcmp (section, "INVALID")))
    {
      p = UDP_FIELD_INVALID;


    }
  else if ((section == NULL) || (!(strcmp (section, "Hdr"))))
    {
      if (field == NULL)
	p = UDP_FIELD_Hdr;
      else if (!(strcmp (field, "source_port")))
	p = UDP_FIELD_Hdr_source_port;
      else if (!(strcmp (field, "destination_port")))
	p = UDP_FIELD_Hdr_destination_port;
      else if (!(strcmp (field, "length")))
	p = UDP_FIELD_Hdr_length;
      else if (!(strcmp (field, "checksum")))
	p = UDP_FIELD_Hdr_checksum;
      else if (!(strcmp (field, "payload")))
	p = UDP_FIELD_Hdr_payload;


    }

  if (p == -1)
    {
      printf ("UDP: Unrecognized field: \" %s \".  Returning error.\n",
	      field);
      return (UDP_FIELD_INVALID);
    }

  ;
  return (p);
}


static int
udp_stock_decompose (struct packet *p)
{
  struct capsule *e = p->encaps, *f;
  struct UDP_Hdr *h;
  struct section *s;


  while (e->next)
    e = e->next;

  h = (struct UDP_Hdr *) e->payload;


  append_encap (p, protocols[0], e->payload_length, e->payload);
  f = e->next;
  f->pe = &udp_pe;

  if ((s = p->c->eng->alloc (p->c, sizeof (*s))) == NULL)
    exit (-1);
  s->section_id = UDP_SECTION_Hdr;
  s->content = h;
  s->next = NULL;
  f->sections = s;
  f->length = e->payload_length;


  f->payload_length = (f->length - (offsetof (struct UDP_Hdr, payload)));
  f->payload_protocol =
    (u_int32_t) ((HPN_UDP << 24) |
		 (ntohs (h->destination_port) <
		  1024 ? ntohs (h->destination_port) : ntohs (h->
							      source_port)));
  f->payload = &(e->payload[(offsetof (struct UDP_Hdr, payload))]);


  return (0);
}

static int
udp_stock_reg_prim (int field, char *op, char *arg)
{
  int prid = ++gprid;


  ;

  if (field == UDP_FIELD_INVALID)
    {
      printf
	("UDP: Error.  Received invalid field in reg_prim(), op=%s, arg=%s.\n",
	 op, arg);
      return (-1);
    }


  if (!strcmp (op, "present"))
    {
      struct simple_prim *p;

      if (!(p = malloc (sizeof (*p))))
	exit (-1);
      p->prim = prid;

      if (field == UDP_SELF)
	{

	  p->next = present_sections.proto_self;
	  present_sections.proto_self = p;
	  return (prid);


	}
      else if (field == UDP_FIELD_Hdr)
	{
	  p->next = present_sections.present_Hdr;
	  present_sections.present_Hdr = p;
	  return (prid);


	}

      printf ("That shit (%i) ain't keepin' it real.\n", field);
      return (-1);
    }


  if (field == UDP_FIELD_Hdr_source_port)
    {
      if (hdt_sets[UDP_FIELD_Hdr_source_port] == 0)
	{
	  hdt_sets[UDP_FIELD_Hdr_source_port] = hdt_int32_create ();
	  if (hdt_sets[UDP_FIELD_Hdr_source_port] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "source_port");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[UDP_FIELD_Hdr_source_port], prid, op, arg))
	{
	  printf ("Error adding primitive %s in udp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == UDP_FIELD_Hdr_destination_port)
    {
      if (hdt_sets[UDP_FIELD_Hdr_destination_port] == 0)
	{
	  hdt_sets[UDP_FIELD_Hdr_destination_port] = hdt_int32_create ();
	  if (hdt_sets[UDP_FIELD_Hdr_destination_port] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "destination_port");
	      exit (-1);
	    }
	}
      if (hdt_int32_add
	  (hdt_sets[UDP_FIELD_Hdr_destination_port], prid, op, arg))
	{
	  printf ("Error adding primitive %s in udp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == UDP_FIELD_Hdr_length)
    {
      if (hdt_sets[UDP_FIELD_Hdr_length] == 0)
	{
	  hdt_sets[UDP_FIELD_Hdr_length] = hdt_int32_create ();
	  if (hdt_sets[UDP_FIELD_Hdr_length] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "length");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[UDP_FIELD_Hdr_length], prid, op, arg))
	{
	  printf ("Error adding primitive %s in udp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == UDP_FIELD_Hdr_checksum)
    {
      if (hdt_sets[UDP_FIELD_Hdr_checksum] == 0)
	{
	  hdt_sets[UDP_FIELD_Hdr_checksum] = hdt_int32_create ();
	  if (hdt_sets[UDP_FIELD_Hdr_checksum] == 0)
	    {
	      printf ("AIEE!  Can't add int match set for %s.%s!  Exiting.\n",
		      "Hdr", "checksum");
	      exit (-1);
	    }
	}
      if (hdt_int32_add (hdt_sets[UDP_FIELD_Hdr_checksum], prid, op, arg))
	{
	  printf ("Error adding primitive %s in udp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }
  if (field == UDP_FIELD_Hdr_payload)
    {
      if (hdt_sets[UDP_FIELD_Hdr_payload] == 0)
	{
	  hdt_sets[UDP_FIELD_Hdr_payload] = hdt_blob_create ();;
	  if (hdt_sets[UDP_FIELD_Hdr_payload] == 0)
	    {
	      printf ("AIEE!  Can't add match set for %s.%s!  Exiting.\n",
		      "Hdr", "payload");
	      exit (-1);
	    }
	}
      if (hdt_blob_add (hdt_sets[UDP_FIELD_Hdr_payload], prid, op, arg))
	{
	  printf ("Error adding primitive %s in udp!\n", op);
	  return (-1);
	}
      else
	{;
	}
      return (prid);
    }


  return (-1);
}

static int
udp_stock_unreg_prim (prim_id i)
{

  return (-1);
}

static int
udp_stock_rewrite (struct capsule *c, void *new_payload, u_int32_t len)
{
  if ((c != NULL) && (new_payload != NULL))
    len++;
  return (-1);
}

static char *
udp_stock_report_field (struct capsule *cp, int field,
			struct memcon_context *m)
{
  u_int32_t t;
  char *r, buf[1024];
  struct section *sec, *sec2;

  buf[0] = 0;

  switch (field)
    {


    case (UDP_FIELD_Hdr_source_port):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == UDP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct UDP_Hdr *) (sec->content))->source_port));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (UDP_FIELD_Hdr_destination_port):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == UDP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct UDP_Hdr *) (sec->content))->
			   destination_port));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (UDP_FIELD_Hdr_length):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == UDP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct UDP_Hdr *) (sec->content))->length));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;
    case (UDP_FIELD_Hdr_checksum):
      sec = NULL;
      for (sec2 = cp->sections; sec2; sec2 = sec2->next)
	{
	  if (sec2->section_id == UDP_SECTION_Hdr)
	    sec = sec2;
	}
      if (sec)
	{
	  snprintf (buf, sizeof (buf), "%u",
		    ntohs (((struct UDP_Hdr *) (sec->content))->checksum));
	}
      else
	{
	  snprintf (buf, sizeof (buf), "(section %s not present)", "Hdr");
	}
      break;


    default:

      return (NULL);
    }
  t = (strlen (buf) + 1);
  r = m->eng->alloc (m, t);
  strncpy (r, buf, t);
  return (r);
}

static int
udp_stock_find_matches (struct capsule *cp, struct prim_can *p)
{
  struct section *s;
  struct memcon_context *c = cp->parent->c;
  u_int32_t v;
  unsigned long beg, end;

  if (p)
    return (-1);

  if (cp->matches == NULL && gprid)
    {
      if (!(cp->matches = c->eng->alloc (c, gprid + 1)))
	exit (-1);
      memset (cp->matches, 0, gprid + 1);

      cp->matches[0] = 1;
    }


  for (s = cp->sections; s; s = s->next)
    {
      struct simple_prim *sp;

      for (sp = present_sections.proto_self; sp; sp = sp->next)
	{
	  ;
	  cp->matches[sp->prim] |= 1;
	}


      switch (s->section_id)
	{
	case UDP_SECTION_INVALID:
	  return (-1);


	  break;
	case UDP_SECTION_Hdr:;
	  for (sp = present_sections.present_Hdr; sp; sp = sp->next)
	    {
	      cp->matches[sp->prim] |= 1;
	    }
	  rdtscl (beg);
	  if (hdt_sets[UDP_FIELD_Hdr_source_port])
	    {;
	      v = ntohs (((struct UDP_Hdr *) (s->content))->source_port);;
	      if (hdt_int32_match
		  (hdt_sets[UDP_FIELD_Hdr_source_port], NULL, v, cp->matches,
		   1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "source_port", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[UDP_FIELD_Hdr_destination_port])
	    {;
	      v =
		ntohs (((struct UDP_Hdr *) (s->content))->destination_port);;
	      if (hdt_int32_match
		  (hdt_sets[UDP_FIELD_Hdr_destination_port], NULL, v,
		   cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "destination_port", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[UDP_FIELD_Hdr_length])
	    {;
	      v = ntohs (((struct UDP_Hdr *) (s->content))->length);;
	      if (hdt_int32_match
		  (hdt_sets[UDP_FIELD_Hdr_length], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "length", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[UDP_FIELD_Hdr_checksum])
	    {;
	      v = ntohs (((struct UDP_Hdr *) (s->content))->checksum);;
	      if (hdt_int32_match
		  (hdt_sets[UDP_FIELD_Hdr_checksum], NULL, v, cp->matches, 1))
		{
		  printf ("AIEE!  Can't find int match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "checksum", (end - beg));
	  rdtscl (beg);
	  if (hdt_sets[UDP_FIELD_Hdr_payload])
	    {
	      struct blob bb;
	      bb.blob = &(((struct UDP_Hdr *) (s->content))->payload);
	      bb.len = cp->length - offsetof (struct UDP_Hdr, payload);;
	      if (hdt_blob_match
		  (hdt_sets[UDP_FIELD_Hdr_payload], NULL, &bb, cp->matches,
		   1))
		{
		  printf ("AIEE!  Can't find blob match.  Exiting.\n");
		  exit (-1);
		}
	    }
	  rdtscl (end);
	  printf ("Matching on field %s:%s took %lu cycles.\n", "Hdr",
		  "payload", (end - beg));


	  break;
	default:
	  printf ("Unrecognized section: %i (not %i).\n", s->section_id,
		  UDP_SECTION_Hdr);
	  break;
	}
    }

  return (-1);
}
