#ifndef   HANK_PROTOCOLS_H
#define   HANK_PROTOCOLS_H

/* Copyright 2001, Todd Lewis <tlewis@mindspring.com>
 * Released under GNU GPL v2. */

/* protocols.h: protocol family numbering */

/* A hank protocol number is a 32-bit quantity.  The first
 * eight bits are the major protocol number, so there are a
 * maximum of 256 protocols that can be defined in HANK_PROT_MAJOR.
 * The last 24 bits are the minor protocol number.
 *
 * N.b., these are completely different assignments than the
 * 16-bit protocol assignments in rfc-1700, which aren't very
 * recent.  These assignments are completely private to hank,
 * although of course anyone else who cares to do so is free to
 * use them.
 *
 * Here are some examples:
 *
 * Ethernet (viz. bpf.h)
 *   protocol->num = (HPN_PCAP << 24) | DLT_EN10MB;
 * IP (vis. RFC-1700)
 *   protocol->num = (HPN_802 << 24) | 2048; (search for "EtherTypes" in rfc1700)
 *   protocol->num = (HPN_PPP << 24) | 0x21; (search for "PPP DLL" in rfc1700)
 * UDP
 *   protocol->num = (HPN_IP << 24) | 17;
 * DNS
 *   protocol->num = (HPN_UDP << 24) | 53;
 *
 * The private protocol numbers are there for people to experiment
 * with.  Please do not feel bad about using them to develop
 * your neat gadget/doohicky thingy.
 *
 * If, when your doohicky experiment turns out well,  you need
 * a protocol added to the public list, then that is no big deal.
 * I am very happy to add protocols for you.  Just send
 * mail, for now, to tlewis@mindspring.com, and I'll add
 * it.  However, if you add protocols on your own and let
 * them escape into the wild, or, God forbid, if you renumber
 * these protocols, then Hank Himself will hunt you down
 * and inflict a Sisyphisean punishment on you.  We will not
 * have a reenactment of the pcap experience of every
 * vendor redefining their data link type numbers and causing
 * a huge, unnecessary compatibility mess!
 */

#define X(a,b) HPN_ ## a = b,
enum HANK_PROT_MAJ {
#include "protocols.X"
	EMPTY = 0
};
#undef X

#ifndef  DLT_RAW
#define  DLT_RAW      12 /* raw IP */
#endif /*DLT_RAW*/

enum PAE_PSEUDO_PROTOCOLS {
	PAE_PCAP = 1,
	PAE_NETFILTER = 2,
	PAE_DIVERT = 3,
	PAE_EMPTY = 0
};

#endif /* HANK_PROTOCOLS_H */
