/* Copyright 2001, Todd Lewis <tlewis@mindspring.com>
 * Released under GNU GPL v2. */

/* reporter_simple.c: spits stuff out */

#include "hank.h"
#include "event.h"
#include "packet.h"
#include "protocol_engine.h"
#include "protocol_lookup.h"
#include "support.h"
#include <stdio.h>
#include <string.h>

static char rcsid[] = "$Id: reporter_simple.c,v 1.12 2001/11/15 07:35:25 tlewis Exp $";

static event_id create_event(void);
static handler_id register_handler(event_id incoming, event_id target, xmlNodePtr m);
static int unregister_handler(handler_id handler);
static int set_handler_target(event_id incoming, handler_id handler, event_id target);
static void handle_event(struct packet *p);

struct event_handler reporter = {
	{
		"event_handler",
		"reporter",
		EHMAJ,
		EHMIN,
		&reporter
	},
	&create_event,
	&register_handler,
	&unregister_handler,
	&set_handler_target,
	&handle_event
};

struct report_frag {
	char *text;
	struct protocol_engine *pe;
	int field;
	struct report_frag *next;
};

struct report_frag *reports[1024];
static int reports_used=0;

extern struct protocol_lookup pls;
extern lt proto_table;

static event_id
create_event(void)
{
	int i=reports_used++;

	/* printf("Creating report event %i.\n", i); */
	return((EH_REPORTER<<24)|i);
}

static int
unregister_handler(handler_id handler)
{
	return(-1);
}

static int
set_handler_target(event_id incoming, handler_id handler, event_id target)
{
	/* we don't produce events */
	return(-1);
}

static void
handle_event(struct packet *p)
{
	struct event *e;
	struct capsule *c;
	struct report_frag *rf;


	for(e=p->q;e;e=e->next){
		if((e->id>>24)==EH_REPORTER){
			if(!e->consumed){
				/* printf("Handling event %i.%i.\n", e->id>>24, e->id&0xffffff); */
				printf("REPORT: ");
				for(rf=reports[(e->id&0xffffff)];rf;rf=rf->next){
					if(rf->text){
						printf("%s", rf->text);
					} else {
						for(c=p->encaps;c;c=c->next)
							if(c->protocol==rf->pe->protocols[0])
								printf("%s", rf->pe->report_field(c, rf->field, p->c));
					}
				}
				printf("\n");
				e->consumed=1;
			}
		}
	}
}

static handler_id 
register_handler(event_id incoming, event_id target, xmlNodePtr m)
{
	struct report_frag *rf, *prev=NULL;
	char *proto, *section, *field;
	xmlNodePtr node;


	rf=malloc(sizeof(*rf));
	if(!rf)
		exit(-1);
	rf->text=NULL;
	rf->next=NULL;

	reports[(incoming&0xffffff)]=rf;

	/* printf("Registering report as %i.\n", incoming&0xffffff); */

	if(target)
		return(-1);

	for(node=m->children;node;node=node->next){

		if(!rf){
			if(!(rf=malloc(sizeof(*rf))))
				exit(-1);
			rf->text=NULL;
			rf->next=NULL;
			if(prev)
				prev->next=rf;
		}

		if (!strcmp(node->name, "text")) {
			/* printf("REGISTERING REPORT TEXT \"%s\" in %i.\n", \
				node->content, incoming&0xffffff); */
			rf->text=strdup(node->content);
		} else if (!strcmp(node->name, "field")) {

      	proto   = (char *)xmlGetProp(node, (unsigned const char *)"proto");
			section = (char *)xmlGetProp(node, (unsigned const char *)"section");
			field   = (char *)xmlGetProp(node, (unsigned const char *)"field");

			/* printf("REGISTERING PROTO FIELD %s:%s:%s in %i.\n", \
				proto, section, field, incoming&0xffffff); */

			if(!(rf->pe=pls.find_pe_name(proto_table, proto))){
				printf("AIEE!  Can't find proto \"%s\".\n", proto);
				return(-1);
			}
			rf->field=rf->pe->field_lookup(section, field);
			/* printf("Registered field %i with pe for %s.\n", rf->field, rf->pe->s.name); */
		} else {

			fprintf(stderr, "PARSE ERROR: unrecognized report element: %s.\n", node->name);
			return(-1);
		}

		prev=rf;
		rf=NULL;
	}

	return(0);
}

