#ifndef HANK_SUPPORT_H
#define HANK_SUPPORT_H

#include "memcon.h"
#include "types.h"
#include <netinet/in.h>
#include <arpa/inet.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

struct packet;

struct int_field_test {
	u_int32_t field, a, b; /* gcc doesn't check enums anyway... */
	char op;
};

/* arrays of these are terminated by 'num==0', so don't use 0! */
struct name_table {
	const char *name;
	u_int32_t num;
};

extern struct name_table protocol_names[];

struct plist {
	void *pointers;
	unsigned int array_size;
	unsigned int in_use;
};

/*
#define format_inet_addr(buf,addr) \
   snprintf(buf, sizeof(buf), "%i.%i.%i.%i", \  
      (unsigned char)addr>>24, (unsigned char)addr>>16,\
      (unsigned char)addr>>8, (unsigned char)addr)
*/

#ifndef rdtscl
#define rdtscl(low) \
     __asm__ __volatile__ ("rdtsc" : "=a" (low) : : "edx")
#endif /*rdtscl*/

/* Returns a u_int32_t[2] with min and max     */
/* Syntax: "min(-max)", e.g., "0-1023" or "23" */
/*         (if a single int, then min=max=int) */
extern u_int32_t *hs_convert_ranges(char *arg, struct memcon_context *m);

extern void bit_dump(u_int32_t n, char *buf, size_t size);

extern void split_kv(char *string, char *kv[2], char delim);

extern const char * find_proto_name(u_int32_t protocol);
extern const u_int32_t find_proto_num(const char *name);
extern const char* find_by_num (struct name_table *t, u_int32_t num   );
extern u_int32_t   find_by_name(struct name_table *t, const char *name);

extern u_int16_t bs16(u_int16_t);
extern u_int32_t bs32(u_int32_t);

/* This is really boolean */
extern char test_int_field(struct int_field_test *t, u_int32_t val);

extern void append_encap(struct packet *p, u_int32_t protocol, size_t length, unsigned char *payload);

extern xmlNodePtr get_real_xml_node(xmlNodePtr);

/* stream */

typedef u_int32_t streams_col_t;

/* size here refers to the bitsize of the collection.  E.g., "16"
   will allow you to handle 65535 streams.  Max is 32. */
extern streams_col_t create_streams(int size);
extern int destroy_streams(streams_col_t);

/* returns 0 on success, other on failure */
extern int register_stream(streams_col_t col, void *stream,
	u_int32_t saddr, u_int32_t daddr, u_int16_t sport, u_int16_t dport);

/* returns null on no match */
extern void *get_stream(streams_col_t col, 
	u_int32_t saddr, u_int32_t daddr, u_int16_t sport, u_int16_t dport);

/* returns 0 on success, other on failure */
extern int unregister_stream(streams_col_t col, 
	u_int32_t saddr, u_int32_t daddr, u_int16_t sport, u_int16_t dport);

extern unsigned int plist_add(struct plist *list, void* pointer);

#endif /*HANK_SUPPORT_H*/

